// changes log:
// 02.02.2000 01:18 added relink feature (added CAdrenalin::Relink method, changed 
//                  CAdrenalin::ParseCmdLine and CAdrenalin::PrintHelp method)

#include "pch.h"

#include "Mem.hpp"
#include "Common.hpp"
#include "CfgParse.hpp"
#include "Misc.hpp"
#include "Links.hpp"
#include "Areas.hpp"
#include "Msg.hpp"
#include "Tic.hpp"
#include "crc32.hpp"
#include "Temp.hpp"

#include "CfgLexems.hpp"

#include "Adrenalin.hpp"

// main class instance

CAdrenalin*	pAdrenalin;

// common constants

const char	szForwardsFileName[] = "forwards.lst";

int main(int argc, char **argv) {
	pAdrenalin = new CAdrenalin( argc, argv );
	int	nResult = pAdrenalin->Execute();
	delete	pAdrenalin;
	return (nResult);
}

CAdrenalin::CAdrenalin(int argc, char **argv) : Links(), Areas(), Language() {
    int i, j;

    // check for previous version
    hBusyMutex = CreateMutex( NULL, TRUE, "FILECHO-PROCESSEROR-ADRENALIN" );
    if (GetLastError() == ERROR_ALREADY_EXISTS) {
        cout << "Program is already working\n";
        ErrorCode = ERROR_ALREADY_WORK;
        Action = ACTION_ERROR;
        return;
    }

    // prepare forward flags
    m_fForwardsRead    = false;
    m_fForwardsChanged = false;

    // nothing to purge yet
    PurgeAreasList = LastPurgeArea = NULL;
    
    // AKA weren't defined yet
    GlobalVars.AKAAddrNumber = 0;
    
    // aliases weren't defined yet
    GlobalVars.AliasesNumber = 0;

    // keep startup path
    strcpy( StartupPath, argv[0] );
    for (i = 0, j = -1; StartupPath[i] != '\0'; i++) {
        if (StartupPath[i] == '\\')
            j = i;
    }
    StartupPath[++j] = '\0';

    // read configuration

    ParseAppCfg();

    // no comments ;)

    if (ErrorCode == ERROR_NONE)
    	checkDiskLimits();
    else {
	    char	szErrorMessage[256];
	    GetErrorMessage( ErrorCode, szErrorMessage );
	    cerr << "Error: " << szErrorMessage << endl;
    }

    // can we continue?

    if (ErrorCode == ERROR_NONE)
        ParseAreasCfg();

    // prepare language
    if (ErrorCode == ERROR_NONE) {
        if (access( LngFileName, 0) == 0) {
            if (!Language.Read(LngFileName)) {
                GlobalVars.Log << "! Error reading language file " << LngFileName
                    << endl;
                cout << "Error reading language file " << LngFileName << endl;
                ErrorCode = ERROR_LNG_READ;
                Action    = ACTION_ERROR;
            }
            else if (Language.Count() != LNG_COUNT) {
                GlobalVars.Log << "! Wrong number of messages in language file:\n  "
                    << Language.Count() << " instead of " << LNG_COUNT << endl;
                cout << "Wrong number of messages in language file:\n  "
                    << Language.Count() << " instead of " << LNG_COUNT << endl;
                ErrorCode = ERROR_LNG_NUMBER;
                Action    = ACTION_ERROR;
            }
        }
        else {
            GlobalVars.Log << "! Language file " << LngFileName << " not found.\n";
            cout << "Language file " << LngFileName << " not found.\n";
            ErrorCode = ERROR_LNG_FILE_NOT_FOUND;
            Action    = ACTION_ERROR;
        }
    }
    if (ErrorCode == ERROR_NONE)
        ParseCmdLine( argc, argv );
    else
        Action = ACTION_ERROR;
    if (ErrorCode == ERROR_NONE) {
        if (!AllocGlobalMsgBuf()) {
            ErrorCode = ERROR_MEMORY;
            Action    = ACTION_ERROR;
        }
    }
}

CAdrenalin::~CAdrenalin() {
    int    i;
    char   AreasCfgFullName[MAX_PATH];
    time_t timestamp;

    if (!hBusyMutex)
        return;

    if (m_fForwardsChanged)
	    SaveForwards();
    
    timestamp = time(NULL);
    GlobalVars.Log << "Stop : " << asctime(localtime(&timestamp));
    GlobalVars.Log.close();
    
    if (Areas.LinksChanged()) {
        Areas.Sort();
        strcpy( AreasCfgFullName, StartupPath );
        strcat( AreasCfgFullName, "Areas.cfg" );
        Areas.SaveToFile( AreasCfgFullName );
    }
    
    if (GlobalVars.AliasesNumber > 0) {
        for (i = 0; i < GlobalVars.AliasesNumber; i++)
            delete [] GlobalVars.Aliases[i];
        delete [] GlobalVars.Aliases;
    }

    if (GlobalVars.AKAAddrNumber > 0)
        delete [] GlobalVars.AKAAddr;

    // free purge data
    while (PurgeAreasList != NULL) {
        LastPurgeArea = PurgeAreasList->Next;
        delete [] PurgeAreasList->String;
        delete PurgeAreasList;
        PurgeAreasList = LastPurgeArea;
    }

    while (diskLimits != NULL) {
    	DiskLimitsList*	nextListItem = diskLimits->next;
    	delete diskLimits;
    	diskLimits = nextListItem;
    }
    
    FreeGlobalMsgBuf();
    CloseHandle(hBusyMutex);
}


void CAdrenalin::ParseAppCfg() {
    // config file names list
    struct SCfgList {
        char      Name[MAX_PATH];  // config file name
        int       SkipLines;       // lines to skip
        SCfgList *Prev, *Next;
    };
    
    int         i, j, k, line;
    char       *AppCfgName = "Adrenalin.cfg";
    char        AppCfgFullName[MAX_PATH];
    char        LogFileName[MAX_PATH];
    char        FileName[MAX_PATH];
    SLink       NewLink;
    int         ParsingFailed;
    time_t      timestamp;
    SAddrList  *AKAAddrList, *LastAKAAddr, *CurAKAAddr;
    SFTNAddress NewAKAAddr;
    int         AKAAddrNumber;
    
    int         WaitingUseAKA;
    BOOL        waitingForDefaultGroup;

    bool	fWaitingForLinkProps;	// if true, we've already read Link keyword
    
    BOOL        IncludeFound;
    CCfgParser  CfgParser;
    SCfgList   *CfgList, *CurCfg;

    strcpy( AppCfgFullName, StartupPath );
    strcat( AppCfgFullName, AppCfgName );

    // prepare values
    
    GlobalVars.MainAddr.Domain[0]  = '\0';
    GlobalVars.SysOp[0]            = '\0';
    GlobalVars.InboundPath[0]      = '\0';
    GlobalVars.OutboundPath[0]     = '\0';
    GlobalVars.BinkOutboundPath[0] = '\0';
    GlobalVars.NetmailPath[0]      = '\0';
    GlobalVars.BadTicsPath[0]      = '\0';
    GlobalVars.NewTicsPath[0]      = '\0';
    GlobalVars.AutoCreatedPath[0]  = '\0';
    GlobalVars.LongDescPrefix[0]   = '\0';
    GlobalVars.AutoDescFileName[0] = '\0';
    LngFileName[0]                 = '\0';
    GlobalVars.LongPaths           = -1;
    GlobalVars.Aliases             = NULL;
    GlobalVars.AKAAddr             = NULL;
    GlobalVars.HandleReplaces      = -1;
    GlobalVars.HatchOriginMain     = -1;
    GlobalVars.AllAKASeen          = -1;
    GlobalVars.KillProcessedMail   = -1;
    GlobalVars.ShowMustGoOn        = -1;
    GlobalVars.LogMsgNames         = -1;
    
    GlobalVars.FixFlagName[0]   = '\0';
    GlobalVars.TossFlagName[0]  = '\0';
    GlobalVars.HatchFlagName[0] = '\0';

    GlobalVars.fCheckDomainInBSO        = true;
    GlobalVars.m_nForwardExpirationTime = 0;

    AKAAddrList = LastAKAAddr = NULL;
    AKAAddrNumber = 0;
    
    WaitingUseAKA = 0;
    waitingForDefaultGroup = FALSE;

    fWaitingForLinkProps = false;
    
    LogFileName[0] = '\0';

    diskLimits = NULL;
    
    ErrorCode = ERROR_NONE;
    line = 0;

    // add first item AppCfgFullName into config names list
    CfgList = new SCfgList;
    CfgList->Prev = CfgList->Next = NULL;
    strcpy( CfgList->Name, AppCfgFullName );
    CfgList->SkipLines = 0;

    while (CfgList != NULL && ErrorCode == ERROR_NONE) {
        IncludeFound = FALSE;
        if (CfgParser.Open( CfgList->Name )) {
            // skip desired number of lines
            i = 0;
            while (!CfgParser.Eof() && !CfgParser.Failed() &&
                   CfgParser.Line < CfgList->SkipLines)
                CfgParser.ReadLine();
            // read lines
            line = CfgList->SkipLines;
            while (ErrorCode == ERROR_NONE && !CfgParser.Eof() && IncludeFound == FALSE) {
                CfgParser.ReadLine();
                if (!CfgParser.Failed()) {
                    line++;
                    if (CfgParser.Count() == 0)
                        continue;
		    // cout << getCfgLexemID( CfgParser[0] ) << endl;
                    switch (getCfgLexemID( CfgParser[0] )) {
                        /*********
                          Address
                         *********/
                        case CFG_LEXEM_ADDRESS: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.MainAddr.Domain[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (!ParseFTNAddress( CfgParser[1], GlobalVars.MainAddr )) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (*(GlobalVars.MainAddr.Domain) == '\0')
                                strcpy( GlobalVars.MainAddr.Domain, "Fidonet" );
                            break;
                        }
                        /*********
                          AKA
                         *********/
                        case CFG_LEXEM_AKA: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            // if main address undefined, error occured
                            if (GlobalVars.MainAddr.Domain[0] == '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_FORWARD_DEF;
                                break;
                            }
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (!ParseLinkAddr( CfgParser[1], NewAKAAddr )) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            // check if already defined;
                            if (AddrCmp( GlobalVars.MainAddr, NewAKAAddr ) == 0) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            CurAKAAddr = AKAAddrList;
                            while (CurAKAAddr != NULL) {
                                if (AddrCmp( CurAKAAddr->Addr, NewAKAAddr ) == 0)
                                    break;
                                CurAKAAddr = CurAKAAddr->Next;
                            }
                            if (CurAKAAddr != NULL) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            // add new AKA into AKA list
                            if (AKAAddrList == NULL) {
                                AKAAddrList = new SAddrList;
                                LastAKAAddr = AKAAddrList;
                            }
                            else {
                                LastAKAAddr->Next = new SAddrList;
                                LastAKAAddr = LastAKAAddr->Next;
                            }
                            LastAKAAddr->Next = NULL;
                            memcpy( &(LastAKAAddr->Addr), &NewAKAAddr,
                                    sizeof(SFTNAddress) );
                            AKAAddrNumber++;
                            break;
                        }
                        /*******
                          SysOp
                         *******/
                        case CFG_LEXEM_SYSOP: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() == 1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.SysOp[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strncpy( GlobalVars.SysOp, CfgParser[1], 35 );
                            k = strlen(GlobalVars.SysOp); // copy to pos k
                            j = 35 - k; // how many symbols we can still copy
                            for (i = 2; i < CfgParser.Count() && j > 1; i++) {
                                strcpy( GlobalVars.SysOp + k, " " );
                                k++;
                                j--;
                                strncpy( GlobalVars.SysOp + k, CfgParser[i], j );
                                k = strlen(GlobalVars.SysOp);
                                j = 35 - k;       
                            }
                            break;
                        }
                        /*********
                          Inbound
                         *********/
                        case CFG_LEXEM_INBOUND:  {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.InboundPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.InboundPath, CfgParser[1] );
                            if (*GlobalVars.InboundPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.InboundPath );
                            if (*GlobalVars.InboundPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.InboundPath[strlen(GlobalVars.InboundPath)-1]
                                != '\\')
                                strcat( GlobalVars.InboundPath, "\\");
                            AddStartupPath( GlobalVars.InboundPath );
                            break;
                        }
                        /**********
                          Outbound
                         **********/
                        case CFG_LEXEM_OUTBOUND: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.OutboundPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.OutboundPath, CfgParser[1] );
                            if (*GlobalVars.OutboundPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.OutboundPath );
                            if (*GlobalVars.OutboundPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.OutboundPath[strlen(GlobalVars.OutboundPath)-1]
                                != '\\')
                                strcat( GlobalVars.OutboundPath, "\\");
                            AddStartupPath( GlobalVars.OutboundPath );
                            break;
                        }
                        /**************
                          BinkOutbound
                         **************/
                        case CFG_LEXEM_BINK_OUTBOUND: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.BinkOutboundPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.BinkOutboundPath, CfgParser[1] );
                            if (*GlobalVars.BinkOutboundPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.BinkOutboundPath );
                            if (*GlobalVars.BinkOutboundPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.BinkOutboundPath[strlen(GlobalVars.BinkOutboundPath)-1] != '\\')
                                strcat( GlobalVars.BinkOutboundPath, "\\");
                            AddStartupPath( GlobalVars.BinkOutboundPath );
                            break;
                        }
                        /*********
                          Netmail
                         *********/
                        case CFG_LEXEM_NETMAIL: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.NetmailPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.NetmailPath, CfgParser[1] );
                            if (*GlobalVars.NetmailPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.NetmailPath );
                            if (*GlobalVars.NetmailPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.NetmailPath[strlen(GlobalVars.NetmailPath)-1]
                                != '\\')
                                strcat( GlobalVars.NetmailPath, "\\");
                            AddStartupPath( GlobalVars.NetmailPath );
                            break;
                        }
                        /*********
                          BadTics
                         *********/
                        case CFG_LEXEM_BAD_TICS: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.BadTicsPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.BadTicsPath, CfgParser[1] );
                            if (*GlobalVars.BadTicsPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.BadTicsPath );
                            if (*GlobalVars.BadTicsPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.BadTicsPath[strlen(GlobalVars.BadTicsPath)-1]
                                != '\\')
                                strcat( GlobalVars.BadTicsPath, "\\");
                            AddStartupPath( GlobalVars.BadTicsPath );
                            break;
                        }
                        /*********
                          NewTics
                         *********/
                        case CFG_LEXEM_NEW_TICS: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.NewTicsPath[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.NewTicsPath, CfgParser[1] );
                            if (*GlobalVars.NewTicsPath == '"')
                                RemoveBoundaryQuotes( GlobalVars.NewTicsPath );
                            if (*GlobalVars.NewTicsPath == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.NewTicsPath[strlen(GlobalVars.NewTicsPath)-1]
                                != '\\')
                                strcat( GlobalVars.NewTicsPath, "\\");
                            AddStartupPath( GlobalVars.NewTicsPath );
                            break;
                        }
                        /*************
                          AutoCreated
                         *************/
                        case CFG_LEXEM_AUTOCREATED: {
				if (fWaitingForLinkProps) {
					if (Links.Selected != NULL && Links.Selected->szAutoCreatedPathName[0] != '\0') {
						printf( "Error in file %s, line %ld\n", CfgList->Name, line );
						ErrorCode = ERROR_DUPLICATE;
						break;
					}
					if (CfgParser.Count() != 2) {
						printf( "Error in file %s, line %ld\n", CfgList->Name, line );
						ErrorCode = ERROR_NUMBER;
						break;
					}
					strcpy( Links.Selected->szAutoCreatedPathName, CfgParser[1] );

					if (Links.Selected->szAutoCreatedPathName[0]  == '"')
						RemoveBoundaryQuotes( Links.Selected->szAutoCreatedPathName );
					if (Links.Selected->szAutoCreatedPathName[0] == '\0') {
						printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
						ErrorCode = ERROR_SYNTAX;
						break;
					}
					if (Links.Selected->szAutoCreatedPathName[strlen(Links.Selected->szAutoCreatedPathName)-1] != '\\') {
						strcat(Links.Selected->szAutoCreatedPathName, "\\" );
					}
					AddStartupPath( Links.Selected->szAutoCreatedPathName );
				}
				else {
				    WaitingUseAKA = 0;
				    waitingForDefaultGroup = FALSE;
				    if (CfgParser.Count() != 2) {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_NUMBER;
					break;
				    }
				    if (GlobalVars.AutoCreatedPath[0] != '\0') {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_DUPLICATE;
					break;
				    }
				    strcpy( GlobalVars.AutoCreatedPath, CfgParser[1] );
				    if (*GlobalVars.AutoCreatedPath == '"')
					RemoveBoundaryQuotes( GlobalVars.AutoCreatedPath );
				    if (*GlobalVars.AutoCreatedPath == '\0') {
					printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_SYNTAX;
					break;
				    }
				    if (GlobalVars.AutoCreatedPath[strlen(GlobalVars.AutoCreatedPath)-1]
					!= '\\')
					strcat( GlobalVars.AutoCreatedPath, "\\");
				    AddStartupPath( GlobalVars.AutoCreatedPath );
				}
				break;
                        }
                        /****************
                          LongDescPrefix
                         ****************/
                        case CFG_LEXEM_LONG_DESC_PREFIX: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            i = strlen(CfgParser[1]);
                            if (i > LONG_DESC_PREFIX_SIZE ||
                                (CfgParser[1][0] == '"' && i > LONG_DESC_PREFIX_SIZE+2))
                            {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (GlobalVars.LongDescPrefix[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.LongDescPrefix, CfgParser[1] );
                            if (*GlobalVars.LongDescPrefix == '"')
                                RemoveBoundaryQuotes( GlobalVars.LongDescPrefix );
                            if (*GlobalVars.LongDescPrefix == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /*************
                          LogFile
                         *************/
                        case CFG_LEXEM_LOG_FILE: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (LogFileName[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( LogFileName, CfgParser[1] );
                            if (*LogFileName == '"')
                                RemoveBoundaryQuotes( LogFileName );
                            if (*LogFileName == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            AddStartupPath( LogFileName );
                            break;
                        }
                        /**************
                          AutoDescFile
                         **************/
                        case CFG_LEXEM_AUTO_DESC_FILE: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.AutoDescFileName[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( GlobalVars.AutoDescFileName, CfgParser[1] );
                            if (*GlobalVars.AutoDescFileName == '"')
                                RemoveBoundaryQuotes( GlobalVars.AutoDescFileName );
                            if (*GlobalVars.AutoDescFileName == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            AddStartupPath( GlobalVars.AutoDescFileName );
                            break;
                        }
                        /*********
                          LngFile
                         *********/
                        case CFG_LEXEM_LNG_FILE: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (LngFileName[0] != '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            strcpy( LngFileName, CfgParser[1] );
                            if (*LngFileName == '"')
                                RemoveBoundaryQuotes( LngFileName );
                            if (*LngFileName == '\0') {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            AddStartupPath( LngFileName );
                            break;
                        }
                        /*********
                          Flags
                         *********/
                        case CFG_LEXEM_CREATE_FLAG_AFTER: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 3) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (stricmp( CfgParser[1], "Fix" ) == 0) {
                                if (*GlobalVars.FixFlagName != '\0') {
                                    printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_DUPLICATE;
                                    break;
                                }
                                strcpy( GlobalVars.FixFlagName, CfgParser[2] );
                                if (*GlobalVars.FixFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.FixFlagName );
                                if (*GlobalVars.FixFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.FixFlagName );
                            }
                            else if (stricmp( CfgParser[1], "Toss" ) == 0) {
                                if (*GlobalVars.TossFlagName != '\0') {
                                    printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_DUPLICATE;
                                    break;
                                }
                                strcpy( GlobalVars.TossFlagName, CfgParser[2] );
                                if (*GlobalVars.TossFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.TossFlagName );
                                if (*GlobalVars.TossFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.TossFlagName );
                            }
                            else if (stricmp( CfgParser[1], "Hatch" ) == 0) {
                                if (*GlobalVars.HatchFlagName != '\0') {
                                    printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_DUPLICATE;
                                    break;
                                }
                                strcpy( GlobalVars.HatchFlagName, CfgParser[2] );
                                if (*GlobalVars.HatchFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.HatchFlagName );
                                if (*GlobalVars.HatchFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.HatchFlagName );
                            }
                            else if (stricmp( CfgParser[1], "All" ) == 0) {
                                if (*GlobalVars.FixFlagName != '\0'  ||
                                    *GlobalVars.TossFlagName != '\0' ||
                                    *GlobalVars.HatchFlagName != '\0')
                                {
                                    printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_DUPLICATE;
                                    break;
                                }
                                strcpy( GlobalVars.FixFlagName, CfgParser[2] );
                                if (*GlobalVars.FixFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.FixFlagName );
                                if (*GlobalVars.FixFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.FixFlagName );
                                strcpy( GlobalVars.TossFlagName, CfgParser[2] );
                                if (*GlobalVars.TossFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.TossFlagName );
                                if (*GlobalVars.TossFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.TossFlagName );
                                strcpy( GlobalVars.HatchFlagName, CfgParser[2] );
                                if (*GlobalVars.HatchFlagName == '"')
                                    RemoveBoundaryQuotes( GlobalVars.HatchFlagName );
                                if (*GlobalVars.HatchFlagName == '\0') {
                                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_SYNTAX;
                                    break;
                                }
                                AddStartupPath( GlobalVars.HatchFlagName );
                            }
                            break;
                        }
                        /***********
                          LongPaths
                         ***********/
                        case CFG_LEXEM_LONG_PATHS: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.LongPaths != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.LongPaths = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.LongPaths = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /******************
                          CheckDomainInBSO
                         ******************/
                        case CFG_LEXEM_CHECK_DOMAIN_IN_BSO: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.fCheckDomainInBSO = true;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.fCheckDomainInBSO = false;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /****************
                          HandleReplaces
                         ****************/
                        case CFG_LEXEM_HANDLE_REPLACES: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.HandleReplaces != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.HandleReplaces = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.HandleReplaces = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /************
                          AllAKASeen
                         ************/
                        case CFG_LEXEM_ALL_AKA_SEEN: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.AllAKASeen != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.AllAKASeen = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.AllAKASeen = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /*******************
                          KillProcessedMail
                         *******************/
                        case CFG_LEXEM_KILL_PROCESSED_MAIL: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.KillProcessedMail != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.KillProcessedMail = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.KillProcessedMail = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /**************
                          ShowMustGoOn
                         **************/
                        case CFG_LEXEM_SHOW_MUST_GO_ON: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.ShowMustGoOn != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.ShowMustGoOn = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.ShowMustGoOn = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /*************
                          LogMsgNames
                         *************/
                        case CFG_LEXEM_LOG_MSG_NAMES: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.LogMsgNames != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Yes") == 0) {
                                GlobalVars.LogMsgNames = 1;
                            }
                            else if (stricmp(CfgParser[1], "No") == 0) {
                                GlobalVars.LogMsgNames = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /********************
                          HatchOriginAddress
                         ********************/
                        case CFG_LEXEM_HATCH_ORIGIN_ADDRESS: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.HatchOriginMain != -1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            if (stricmp(CfgParser[1], "Main") == 0) {
                                GlobalVars.HatchOriginMain = 1;
                            }
                            else if (stricmp(CfgParser[1], "AKA") == 0) {
                                GlobalVars.HatchOriginMain = 0;
                            }
                            else {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            break;
                        }
                        /*********
                          Aliases
                         *********/
                        case CFG_LEXEM_ALIASES: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            if (CfgParser.Count() == 1) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.Aliases != NULL) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            // check dupe aliases
                            for (i = 1; i < CfgParser.Count(); i++) {
                                if (stricmp( CfgParser[i], "Adrenalin" ) == 0)
                                    break;
                                for (j = 0; j < i; j++) {
                                    if (stricmp( CfgParser[i], CfgParser[j] ) == 0)
                                        break;
                                }
                                if (j != i)
                                    break;
                            }
                            // loop was broken if i != CfgParser.Count()
                            if (i != CfgParser.Count()) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            GlobalVars.AliasesNumber = CfgParser.Count();
                            GlobalVars.Aliases = new char*[GlobalVars.AliasesNumber];
                            GlobalVars.Aliases[0] = new char[strlen("Adrenalin")+1];
                            strcpy( GlobalVars.Aliases[0], "Adrenalin" );
                            for (i = 1; i < GlobalVars.AliasesNumber; i++) {
                                GlobalVars.Aliases[i] = new char[strlen(CfgParser[i])+1];
                                strcpy( GlobalVars.Aliases[i], CfgParser[i] );
                            }
                            break;
                        }
                        /******
                          Link
                         ******/
                        case CFG_LEXEM_LINK: {
                            WaitingUseAKA = 1;
                            waitingForDefaultGroup = TRUE;
			    fWaitingForLinkProps = true;
                            if (CfgParser.Count() < 3) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (GlobalVars.MainAddr.Domain[0] == '\0') {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_FORWARD_DEF;
                                break;
                            }
			    memset( &NewLink, 0, sizeof(NewLink) );
                            if (!ParseLinkAddr( CfgParser[1], NewLink.Address )) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (strlen(CfgParser[2]) > PASSWORD_SIZE) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }			    
                            strcpy( NewLink.Password, CfgParser[2] );
                            strcpy( NewLink.cfgFileName, CfgList->Name );
                            NewLink.Line              = CfgParser.Line;
                            // set fields into default values
                            NewLink.RouteType         = -1;
                            NewLink.BSOType           = -1;
                            NewLink.AllowAreaCreation = -1;
                            NewLink.defaultGroup      = 0;
                            NewLink.Activity          = ACTIVITY_USUAL;
                            NewLink.Priority          = 0;
                            NewLink.Groups            = 0;
                            i = 3;
                            ParsingFailed = 0;
                            while (i < CfgParser.Count() && !ParsingFailed) {
                                if (strlen(CfgParser[i]) > 1) {
                                    if (NewLink.Groups != 0) {
                                        ParsingFailed = 1;
                                    }
                                    else if (!ParseLinkAccess( CfgParser[i], NewLink ))
                                        ParsingFailed = 1;
                                }
                                else {
                                    switch (toupper(CfgParser[i][0])) {
                                    // routing
                                    case 'S':
                                        if (NewLink.RouteType == -1)
                                            NewLink.RouteType = ROUTE_SEND;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'R':
                                        if (NewLink.RouteType == -1)
                                            NewLink.RouteType = ROUTE_RECEIVE;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'B':
                                        if (NewLink.RouteType == -1)
                                            NewLink.RouteType = ROUTE_BIDIRECT;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    // outbound
                                    case 'N':
                                        if (NewLink.BSOType == -1)
                                            NewLink.BSOType = BSO_NORMAL;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'D':
                                        if (NewLink.BSOType == -1)
                                            NewLink.BSOType = BSO_DIRECT;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'H':
                                        if (NewLink.BSOType == -1)
                                            NewLink.BSOType = BSO_HOLD;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'I':
                                        if (NewLink.BSOType == -1)
                                            NewLink.BSOType = BSO_IMMEDIATE;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'C':
                                        if (NewLink.BSOType == -1)
                                            NewLink.BSOType = BSO_CRASH;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    // allow autocreating
                                    case 'A':
                                        if (NewLink.AllowAreaCreation == -1)
                                            NewLink.AllowAreaCreation = 1;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    // activity state
                                    case 'P':
                                        if (NewLink.Activity == ACTIVITY_USUAL)
                                            NewLink.Activity = ACTIVITY_PASSIVE;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    case 'U':
                                        if (NewLink.Activity == ACTIVITY_USUAL)
                                            NewLink.Activity = ACTIVITY_UNAVAILABLE;
                                        else
                                            ParsingFailed = 1;
                                        break;
                                    default:
                                        ParsingFailed = 1;
                                    }
                                }
                                i++;
                            }
                            if (ParsingFailed) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name,
                                        line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (NewLink.AllowAreaCreation == -1)
                                NewLink.AllowAreaCreation = 0;
                            if (NewLink.RouteType == -1)
                                NewLink.RouteType = ROUTE_RECEIVE;
                            if (NewLink.BSOType == -1)
                                NewLink.BSOType = BSO_HOLD;
                            if (NewLink.Groups == 0)
                                NewLink.Groups = 0x3FFFFFF;
                            NewLink.OurAddr = NULL;
                            if (Links.Select(NewLink.Address)) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_DUPLICATE;
                                break;
                            }
                            Links.Add(NewLink);
                            Links.Select( NewLink.Address );
                            break;
                        }
                        /********
                          UseAKA
                         ********/
                        case CFG_LEXEM_USE_AKA: {
                            if (!WaitingUseAKA) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_FORWARD_DEF;
                                break;
                            }
                            WaitingUseAKA = 0;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (!ParseLinkAddr( CfgParser[1], NewAKAAddr )) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            if (AddrCmp( GlobalVars.MainAddr, NewAKAAddr ) == 0)
                                Links.Selected->OurAddr = &(GlobalVars.MainAddr);
                            else {
                                CurAKAAddr = AKAAddrList;
                                while (CurAKAAddr != NULL) {
                                    if (AddrCmp( CurAKAAddr->Addr, NewAKAAddr ) == 0)
                                        break;
                                    CurAKAAddr = CurAKAAddr->Next;
                                }
                                if (CurAKAAddr != NULL)
                                    Links.Selected->OurAddr = &(CurAKAAddr->Addr);
                                else {
                                    printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                    ErrorCode = ERROR_UNKNOWN_AKA;
                                    break;
                                }
                            }
                            break;
                        }
                        /**************
                          DefaultGroup
                         **************/
                        case CFG_LEXEM_DEFAULT_GROUP: {
                            if (!waitingForDefaultGroup) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_FORWARD_DEF;
                                break;
                            }
                            waitingForDefaultGroup = FALSE;
                            if (CfgParser.Count() != 2) {
                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_NUMBER;
                                break;
                            }
                            if (strlen(CfgParser[1]) != 1 || !isalpha(CfgParser[1][0])) {
                                printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                                ErrorCode = ERROR_SYNTAX;
                                break;
                            }
                            Links.Selected->defaultGroup = 1 << (toupper(CfgParser[1][0]) - 'A');
                            break;
                        }
			/************
			  AvailAreas
			 ************/
			case CFG_LEXEM_AVAIL_AREAS: {
				if (!fWaitingForLinkProps) {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_FORWARD_DEF;
					break;
				}
				if (Links.Selected != NULL && Links.Selected->szfnAvailAreas[0] != '\0') {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_DUPLICATE;
					break;
				}
				if (CfgParser.Count() != 2) {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_NUMBER;
					break;
				}
				strcpy( Links.Selected->szfnAvailAreas, CfgParser[1] );

                                if (Links.Selected->szfnAvailAreas[0]  == '"')
					RemoveBoundaryQuotes( Links.Selected->szfnAvailAreas );
				if (Links.Selected->szfnAvailAreas[0] == '\0') {
					printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_SYNTAX;
					break;
				}
				AddStartupPath( Links.Selected->szfnAvailAreas );
				break;
			}
                        /***********
                          RobotName
                         ***********/
                        case CFG_LEXEM_ROBOT_NAME: {
				if (!fWaitingForLinkProps) {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_FORWARD_DEF;
					break;
				}
				if (CfgParser.Count() != 2) {
					printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_NUMBER;
					break;
				}
				if (Links.Selected->m_szRobotName[0] != '\0') {
	                                printf( "Error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_DUPLICATE;
					break;
				}
				char	szTmp[LINE_SIZE];
				strcpy( szTmp, CfgParser[1] );
				if (szTmp[0] == '\"') {
					RemoveBoundaryQuotes( szTmp );
				}
				if (szTmp[0] == '\0') {
					printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
					ErrorCode = ERROR_SYNTAX;
					break;
				}
				if (strlen(szTmp) > 35) {
					cerr << "Warning: robot name " << szTmp << " truncated to 35 symbols\n";
				}
				strncpy( Links.Selected->m_szRobotName, szTmp, 35 );
				break;
                        }
                        /*
                         * FreeDiskSpaceLimit <DRIVE ROOT> <LIMIT IN KBYTES>
                         *
                         */
                        case CFG_LEXEM_FREE_DISK_SPACE_LIMIT: {
                        	WaitingUseAKA = 0;
                        	waitingForDefaultGroup = FALSE;
				fWaitingForLinkProps = false;
                        	if (CfgParser.Count() != 3) {
                            		printf( "Error in file %s, line %ld:\nwrong number of arguments\n", CfgList->Name, line );
                            		ErrorCode = ERROR_NUMBER;
                            		break;
                        	}
                        	// check for duplicates
                        	DiskLimitsList* curLimit = diskLimits;
                        	while (curLimit != NULL) {
                        		if (stricmp(curLimit->root, CfgParser[1]) == 0)
                        			break;
                        	        curLimit = curLimit->next;
                        	}
                        	if (curLimit != NULL) {
                            	printf( "Error in file %s, line %ld:\nduplicate drive name\n", CfgList->Name, line );
                            	ErrorCode = ERROR_DUPLICATE;
                            	break;
                        	}
                        	DWORD	limit;
                        	if (parseIntWithCommas( CfgParser[2], limit ) == 1) {
                        		DiskLimitsList*	newLimit;
                        		if (diskLimits == NULL) {
                        			newLimit = new DiskLimitsList;
                        		}
                        		else {
                        			newLimit = diskLimits;
                        			while (newLimit->next != NULL)
                        				newLimit = newLimit->next;
                        		        newLimit->next = new DiskLimitsList;
                        		        newLimit = newLimit->next;
                        		}
                        		newLimit->next = NULL;
                        		strcpy( newLimit->root, CfgParser[1] );
                        		newLimit->limit = limit;
                        		if (diskLimits == NULL)
                        			diskLimits = newLimit;
                        	}
                        	else {
                            	printf( "Error in file %s, line %ld:\n2nd argument is not a valid number\n", CfgList->Name, line );
                            	ErrorCode = ERROR_NUMBER;
                            	break;
                        	}
                        	break;
                        }
			/***********************
			  ForwardExpirationTime
			 ***********************/
			case CFG_LEXEM_FORWARD_EXPIRATION_TIME: {
				WaitingUseAKA = 0;
				waitingForDefaultGroup = FALSE;
				fWaitingForLinkProps = false;
				if (CfgParser.Count() != 2) {
                            		printf( "Error in file %s, line %ld:\nwrong number of arguments\n", CfgList->Name, line );
                            		ErrorCode = ERROR_NUMBER;
                            		break;
				}
				if (GlobalVars.m_nForwardExpirationTime != 0) { // already defined
                            		printf( "Error in file %s, line %ld\nAlready defined parameter\n", CfgList->Name, line );
                            		ErrorCode = ERROR_DUPLICATE;
                            		break;
				}
				GlobalVars.m_nForwardExpirationTime = atoi(CfgParser[1]);
				if (GlobalVars.m_nForwardExpirationTime == 0) {
                            		printf( "Error in file %s, line %ld:\nincorrect number\n", CfgList->Name, line );
                            		ErrorCode = ERROR_SYNTAX;
                            		break;
				}
				break;
			}
                        /*********
                          Include
                         *********/
                        case CFG_LEXEM_INCLUDE: {
                            WaitingUseAKA = 0;
                            waitingForDefaultGroup = FALSE;
			    fWaitingForLinkProps = false;
                            strcpy( FileName, CfgParser[1] );
                            if (FileName[0] == '"')
                                RemoveBoundaryQuotes( FileName );
                            // check if full name or not
                            if (FileName[0] != '\0' && FileName[1] == ':')
                                strcpy( AppCfgFullName, FileName );
                            else {
                                strcpy( AppCfgFullName, StartupPath );
                                strcat( AppCfgFullName, FileName );
                            }
                            // check if we already have this config
                            CurCfg = CfgList;
                            while (CurCfg != NULL) {
                                if (stricmp( CurCfg->Name, AppCfgFullName ) == 0)
                                    break;
                                CurCfg = CurCfg->Prev;
                            }
                            if (CurCfg == NULL) {
                                CfgList->SkipLines = CfgParser.Line + 1;
                                CfgList->Next = new SCfgList;
                                CfgList->Next->Prev = CfgList;
                                CfgList = CfgList->Next;
                                CfgList->Next = NULL;
                                strcpy( CfgList->Name, AppCfgFullName );
                                CfgList->SkipLines = 0;
                                IncludeFound = TRUE;
                                break;
                            }
                            break;
                        }
                        default: {
                            printf( "Error in file %s, line %ld\n", CfgList->Name, line );
                            ErrorCode = ERROR_UNKNOWN_KEYWORD;
                            break;
                        }
                    }
                }
                else if (!CfgParser.Eof()) {
                    line++;
                    printf( "Syntax error in file %s, line %ld\n", CfgList->Name, line );
                    ErrorCode = ERROR_SYNTAX;
                    break;
                }
            }
            
        }
        else {
            printf( "Error: configuration file %s not found\n",
                    CfgList->Name );
            ErrorCode = ERROR_FILE_ABSENT;
        }
        if (ErrorCode == ERROR_NONE && IncludeFound == FALSE) {
            // go back one list item
            CurCfg = CfgList->Prev;
            delete CfgList;
            CfgList = CurCfg;
        }
    }
    
    if (ErrorCode != ERROR_NONE) {
        // free AKA addresses if needed
        while (AKAAddrList != NULL) {
            LastAKAAddr = AKAAddrList->Next;
            delete AKAAddrList;
            AKAAddrList = LastAKAAddr;
        }
        // free config names list if have to
        while (CfgList != NULL) {
            CurCfg = CfgList->Prev;
            delete CfgList;
            CfgList = CurCfg;
        }
        return;
    }
    if (GlobalVars.MainAddr.Domain[0] == '\0') {
        printf( "Error: system address wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.InboundPath[0] == '\0') {
        printf( "Error: Inbound path wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.BinkOutboundPath[0] == '\0') {
        printf( "Error: BinkOutbound path wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.NetmailPath[0] == '\0') {
        printf( "Error: Netmail path wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.BadTicsPath[0] == '\0') {
        printf( "Error: BadTics path wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.NewTicsPath[0] == '\0') {
        printf( "Error: NewTics path wasn't defined in file %s\n",
                AppCfgName );
        ErrorCode = ERROR_ABSENT_DEF;
        return;
    }
    if (GlobalVars.Aliases == NULL) {
        GlobalVars.AliasesNumber = 1;
        GlobalVars.Aliases = new char*[1];
        GlobalVars.Aliases[0] = new char[strlen("Adrenalin")+1];
        strcpy( GlobalVars.Aliases[0], "Adrenalin" );
    }
    if (!Links.SelectFirst())
        puts( "Warning: no links defined" );
    // set defaults for
    // AutoCreatedPath
    if (GlobalVars.AutoCreatedPath[0] == '\0')
        strcpy( GlobalVars.AutoCreatedPath, StartupPath );
    // LongPaths
    if (GlobalVars.LongPaths == -1)
        GlobalVars.LongPaths = 1;
    // HandleReplaces
    if (GlobalVars.HandleReplaces == -1)
        GlobalVars.HandleReplaces = 0;
    // AllAKASeen
    if (GlobalVars.AllAKASeen == -1)
        GlobalVars.AllAKASeen = 0;
    // KillProcessedMail
    if (GlobalVars.KillProcessedMail == -1)
        GlobalVars.KillProcessedMail = 0;
    // ShowMustGoOn
    if (GlobalVars.ShowMustGoOn == -1)
        GlobalVars.ShowMustGoOn = 1;
    if (GlobalVars.LogMsgNames == -1)
        GlobalVars.LogMsgNames = 0;
    // HatchOriginMain
    if (GlobalVars.HatchOriginMain == -1)
        GlobalVars.HatchOriginMain = 0;
    // LongDescPrefix
    if (GlobalVars.LongDescPrefix[0] == '\0')
        strcpy( GlobalVars.LongDescPrefix, " " );
    // init AKA addresses
    if (AKAAddrNumber > 0) {
        GlobalVars.AKAAddr       = new SFTNAddress[AKAAddrNumber];
        GlobalVars.AKAAddrNumber = AKAAddrNumber;
        i = 0;
        while (AKAAddrList != NULL) {
            memcpy( GlobalVars.AKAAddr + i, &(AKAAddrList->Addr),
                    sizeof(SFTNAddress) );
            // correct corresponding AKAs
            Links.SelectFirst();
            while (Links.Selected != NULL) {
                if (Links.Selected->OurAddr == &(AKAAddrList->Addr))
                    Links.Selected->OurAddr = GlobalVars.AKAAddr + i;
                Links.SelectNext();
            }
            i++;
            LastAKAAddr = AKAAddrList->Next;
            delete AKAAddrList;
            AKAAddrList = LastAKAAddr;
        }
    }
    // choose AKA addresses for links
    Links.SelectFirst();
    while (Links.Selected != NULL) {
        if (Links.Selected->OurAddr == NULL)
            ChooseAKA( Links.Selected );
        Links.SelectNext();
    }
    // LngFile
    if (*LngFileName == '\0') {
        strcpy( LngFileName, StartupPath );
        strcat( LngFileName, "Adrenalin.lng" );
    }
    // init Log
    if (*LogFileName == '\0') {
        strcpy( LogFileName, StartupPath );
        strcat( LogFileName, "adrenalin.log" );
    }
    else {
        // get dir from name
        for (i = 0, j = -1; LogFileName[i] != '\0'; i++) {
            if (LogFileName[i] == '\\')
                j = i;
        }
        if (j >= 0) {
            LogFileName[j] = '\0';
            MakeDirStructure( LogFileName );
            LogFileName[j] = '\\';
        }
    }
    GlobalVars.Log.open( LogFileName, ios::app );
    if (!GlobalVars.Log)
        GlobalVars.Log.open( LogFileName );
    if (!GlobalVars.Log)
        cerr << "Failed to open log file " << LogFileName << endl;
    else {
        timestamp = time(NULL);
        GlobalVars.Log << endl << "Start: " << asctime(localtime(&timestamp));
    }
}

void formatSize(DWORD size, char* str) {
	// 1234 -> 1,234

	int	i, j, count;

	count = 0;
	j = 0;
	while (TRUE) {
		str[j] = (char)(size % 10) + '0';
		size /= 10;
		j++;
		count++;
		if (count == 3 && size > 0) {
			str[j] = ',';
			j++;
			count = 0;
		}
		if (size == 0)
			break;
	}
	str[j] = '\0';

	for (i = 0; i < (j >> 1); i++) {
		char	c  = str[i];
		str[i]     = str[j - 1 - i];
		str[j - 1 - i] = c;
	}
}

void CAdrenalin::checkDiskLimits() {
	if (diskLimits == NULL)
		return;

	// lets get OS version info to decide what function to use:
	//   GetDiskFreeSpace or GetDiskFreeSpaceEx
	BOOL		useExFunction = FALSE;
	OSVERSIONINFO	versionInfo;

	if (GetVersionEx(&versionInfo)) {
		switch (versionInfo.dwPlatformId) {
			case VER_PLATFORM_WIN32_WINDOWS:
				if (LOWORD(versionInfo.dwBuildNumber) > 1000)
					useExFunction = TRUE;
				break;
			case VER_PLATFORM_WIN32_NT:
				useExFunction = TRUE;
				break;
                }
	}

	DiskLimitsList*	currentLimit;
	BOOL		success;
	DWORD		sectorsPerCluster;
	DWORD		bytesPerSector;
	DWORD		numberOfFreeClusters;
	DWORD		totalNumberOfClusters;
	ULARGE_INTEGER	freeBytesAvail;
	ULARGE_INTEGER	totalBytes;
	ULARGE_INTEGER	totalFreeBytes;

	struct SizesList {
		DWORD		freeKBytes;
		DiskLimitsList*	where;
		SizesList*	next;
	};

	SizesList*	failedSizes;
	SizesList*	lastFailedSize;

	char		str[80];
	
	failedSizes = NULL;

	currentLimit = diskLimits;
	while (currentLimit != NULL) {
		if (useExFunction) {
			success = GetDiskFreeSpaceEx( currentLimit->root,
                        			      &freeBytesAvail,
                                                      &totalBytes,
                                                      &totalFreeBytes );
		}
		else {
			success = GetDiskFreeSpace( currentLimit->root,
                        			    &sectorsPerCluster,
                                                    &bytesPerSector,
                                                    &numberOfFreeClusters,
                                                    &totalNumberOfClusters );
                        if (success) {
                        	freeBytesAvail.LowPart  = (sectorsPerCluster*bytesPerSector*numberOfFreeClusters) & 0xFFFFFFFF;
                        	freeBytesAvail.HighPart = (sectorsPerCluster*bytesPerSector*numberOfFreeClusters) >> 32;
                        }
		}
		if (success) {
			DWORD	h = freeBytesAvail.HighPart;
			DWORD	l = freeBytesAvail.LowPart;
			DWORD	freeKBytes = ((__int64)l >> 10) + ((__int64)h << 22);
			if (freeKBytes < currentLimit->limit) {
				if (failedSizes == NULL) {
                                	failedSizes    = new SizesList;	
                                	lastFailedSize = failedSizes;
                                }
                                else {
                                	lastFailedSize->next = new SizesList;
                                	lastFailedSize       = lastFailedSize->next;
                                }
                                lastFailedSize->next       = NULL;
                                lastFailedSize->freeKBytes = freeKBytes;
                                lastFailedSize->where      = currentLimit;
			}
		}
		else {
                        char* lpMsgBuf;
                        
                        FormatMessage(
                            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                            NULL,
                            GetLastError(),
                            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                            (LPTSTR) &lpMsgBuf,
                            0,
                            NULL
                        );
                        
                        CharToOem( lpMsgBuf, lpMsgBuf );

                        // truncate tail CRs and LFs
                        int	len = strlen(lpMsgBuf);
                        for (int i = len-1; i > 0; i--) {
                        	if (lpMsgBuf[i] == 0x0D || lpMsgBuf[i] == 0x0A)
                        		lpMsgBuf[i] = '\0';
                        	else
                        		break;
                        }

                        // Display the string.
                        GlobalVars.Log << "! Error getting drive " << currentLimit->root << " free space:\n" << lpMsgBuf << endl;
                        cout           << "! Error getting drive " << currentLimit->root << " free space:\n" << lpMsgBuf << endl;
                        
                        // Free the buffer.
                        LocalFree( lpMsgBuf );
		}
		currentLimit = currentLimit->next;
	}

	if (failedSizes != NULL) {
		GlobalVars.Log << "! Low disk space\n";
		cout           << "! Low disk space\n";

		ErrorCode = ERROR_LOW_DISK_SPACE;
		char	errorMsg[256];
		SizesList*	startSizes = failedSizes;
		while (failedSizes != NULL) {
			formatSize( failedSizes->freeKBytes, str );
			sprintf( errorMsg, "  %s -  %s kByte(s) with the limit of ",failedSizes->where->root, str );
			formatSize( failedSizes->where->limit, str );
			sprintf( errorMsg + strlen(errorMsg), "%s KByte(s)\n", str );

			GlobalVars.Log << errorMsg;
			cerr           << errorMsg;

			failedSizes = failedSizes->next;
		}
		failedSizes = startSizes;

		// prepare data for sysop
		char	lowSpaceFile[MAX_PATH];
		char	curTime[256];
		BOOL	isAppend = FALSE;

                time_t		timestamp = time(NULL);
                struct tm*	tmStruct = localtime(&timestamp);

                sprintf( curTime, "%02d-%02d-%04d", tmStruct->tm_mday, tmStruct->tm_mon + 1, (tmStruct->tm_year < 80 ? tmStruct->tm_year + 2000 : tmStruct->tm_year + 1900) );

		strcpy( lowSpaceFile, StartupPath );
		strcat( lowSpaceFile, "LowSpace.dat" );

		ifstream	in(lowSpaceFile);
		if (in) {
			char	buf[256];
			
			// read date
                        in.getline( buf, 256 );


			if (in && strcmp( buf, curTime ) == 0) {
				isAppend = TRUE;
				// skip all low space data already found
				while (!in.eof() && failedSizes != NULL) {
					in.getline( buf, 256 );
					if (!in)
						break;
				        SizesList*	prevSize    = NULL;
				        SizesList*	currentSize = failedSizes;
				        while (currentSize != NULL && stricmp( currentSize->where->root, buf ) != 0) {
				        	if (prevSize == NULL)
				        		prevSize = failedSizes;
				        	else
				        		prevSize = prevSize->next;
				        	currentSize = currentSize->next;
				        }
				        if (currentSize != NULL) {
				        	if (prevSize == NULL)
				        		failedSizes = failedSizes->next;
				        	else
				        		prevSize->next = currentSize->next;
				        	delete currentSize;
				        }
				}
			}
			in.close();
		}
		if (failedSizes != NULL) {
        		ofstream	out;
        		if (isAppend) {
        			out.open( lowSpaceFile, ios::app );
        		}
        		if (isAppend == FALSE || !out) {
        			out.open( lowSpaceFile );
        			out << curTime << endl;
        		}

        		CStringList	bodyLines;

        		bodyLines.Add("");
        		bodyLines.Add("Low disk space on drive(s):");

        		char	line[256];

        		SizesList*	currentSize = failedSizes;
        		while (currentSize != NULL) {
        			formatSize( currentSize->freeKBytes, str );
        			sprintf( line, "  %s - %s kbyte(s) with the limit of ", currentSize->where->root, str );
        			formatSize( currentSize->where->limit, str );
        			sprintf( line + strlen(line), "%s kbyte(s)", str );
        			bodyLines.Add( line );
        			out << currentSize->where->root << endl;
        			SizesList*	nextSize = currentSize->next;
        			delete currentSize;
        			currentSize = nextSize;
        		}
        		out.close();
                        bodyLines.Add( "" );
                        sprintf( line, "--- %s v%lu.%lu%s", PROG_NAME, PROG_VER_MAJOR, PROG_VER_MINOR, PROG_VER_SUFFIX );
                        bodyLines.Add( line );
        		MakeDirStructure( GlobalVars.NetmailPath );
        		CMsg	msg;
        		msg.From = GlobalVars.MainAddr;
        		msg.To   = GlobalVars.MainAddr;
        		strcpy( msg.Header.FromUserName, GlobalVars.SysOp );
        		strcpy( msg.Header.ToUserName,   "Adrenalin" );
                        msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                   "Low disk space!", bodyLines.Strings );
		}
	}
}


void CAdrenalin::ParseAreasCfg() {
    char        *AreasCfgName = "Areas.cfg";
    char         AreasCfgFullName[MAX_PATH];
    int          line;
    int          i;
    int          DoRead;
    SArea        NewArea;
    int          ParsingFailed;
    SLinked     *LinkedList, *LastLinked;
    int          RouteType;
    SFTNAddress  LinkAddr;
    

    ErrorCode = ERROR_NONE;
    
    strcpy( AreasCfgFullName, StartupPath );
    strcat( AreasCfgFullName, AreasCfgName );
    
    CCfgParser CfgParser(AreasCfgFullName);
    if (CfgParser.Failed()) {
        printf( "Warning: configuration file %s not found\n", AreasCfgName );
        return;
    }
    
    line   = 0;
    DoRead = 1;
    while (!CfgParser.Eof()) {
        if (DoRead)
            CfgParser.ReadLine();
        if (!CfgParser.Failed()) {
            if (DoRead)
                line++;
            if (CfgParser.Count() == 0) {
                DoRead = 1;
                continue;
            }
            /******
              Area
             ******/
            if (stricmp( CfgParser[0], "Area" ) == 0) {
                if (CfgParser.Count() < 3) {
                    printf( "Error in file %s, line %ld\n", AreasCfgName,
                            line );
                    ErrorCode = ERROR_NUMBER;
                    break;
                }
                if (strlen(CfgParser[1]) > AREA_NAME_SIZE) {
                    printf( "Error in file %s, line %ld\n", AreasCfgName,
                            line );
                    ErrorCode = ERROR_SYNTAX;
                    break;
                }
                strcpy( NewArea.Name, CfgParser[1] );
                strcpy( NewArea.Path, CfgParser[2] );
                if (*NewArea.Path == '"')
                    RemoveBoundaryQuotes( NewArea.Path );
                if (*NewArea.Path == '\0') {
                    printf( "Syntax error in file %s, line %ld\n",
                            AreasCfgName, line );
                    ErrorCode = ERROR_SYNTAX;
                    break;
                }
                if (NewArea.Path[strlen(NewArea.Path)-1] != '\\')
                    strcat( NewArea.Path, "\\" );
                NewArea.Line            = line;
                NewArea.Desc[0]         = '\0';
                NewArea.RouteType       = -1;
                NewArea.Priority        = -1;
                NewArea.Group           = 0;
                NewArea.PassThrough     = -1;
                NewArea.VetoManualPurge = -1;
                i = 3;
                ParsingFailed = 0;
                while (i < CfgParser.Count() && !ParsingFailed) {
                    if (strlen( CfgParser[i] ) == 1) {
                        // routing
                        switch (toupper(CfgParser[i][0])) {
                        case 'S':
                            if (NewArea.RouteType == -1)
                                NewArea.RouteType = ROUTE_SEND;
                            else
                                ParsingFailed = 1;
                            break;
                        case 'R':
                            if (NewArea.RouteType == -1)
                                NewArea.RouteType = ROUTE_RECEIVE;
                            else
                                ParsingFailed = 1;
                            break;
                        case 'B':
                            if (NewArea.RouteType == -1)
                                NewArea.RouteType = ROUTE_BIDIRECT;
                            else
                                ParsingFailed = 1;
                            break;
                        case 'P':
                            if (NewArea.PassThrough == -1)
                                NewArea.PassThrough = 1;
                            else
                                ParsingFailed = 1;
                            break;
                        case 'V':
                            if (NewArea.VetoManualPurge== -1)
                                NewArea.VetoManualPurge = 1;
                            else
                                ParsingFailed = 1;
                            break;
                        default:
                            ParsingFailed = 1;
                        }
                        
                    }
                    else if (CfgParser[i][0] == '"'){
                        // description
                        if (NewArea.Desc[0] == '\0')
                            strcpy( NewArea.Desc, CfgParser[i] );
                        else
                            ParsingFailed = 1;
                    }
                    else {
                        // access
                        if (NewArea.Priority != -1)
                            ParsingFailed = 1;
                        else if (!ParseAreaAccess( CfgParser[i], NewArea ))
                            ParsingFailed = 1;
                    }
                    i++;
                }
                if (ParsingFailed) {
                    printf( "Syntax error in file %s, line %ld\n",
                            AreasCfgName, line );
                    ErrorCode = ERROR_SYNTAX;
                    break;
                }
                if (NewArea.RouteType == -1)
                    NewArea.RouteType = ROUTE_NONE;
                if (NewArea.Priority == -1)
                    NewArea.Priority = 0;
                if (NewArea.PassThrough == -1)
                    NewArea.PassThrough = 0;
                if (NewArea.VetoManualPurge == -1)
                    NewArea.VetoManualPurge = 0;
                if (NewArea.Desc[0] == '"')
                    RemoveBoundaryQuotes( NewArea.Desc );
                LinkedList = LastLinked = NULL;
                ParsingFailed = 0;
                while (!CfgParser.Eof() && !ParsingFailed) {
                    CfgParser.ReadLine();
                    if (CfgParser.Failed())
                        break;
                    line++;
                    if (CfgParser.Count() == 0)
                        continue;
                    /********
                      Linked
                     ********/
                    if (stricmp( CfgParser[0], "Linked" ) == 0) {
                        if (CfgParser.Count() == 1) {
                            ParsingFailed = 1;
                            break;
                        }
                        for (i = 1; i < CfgParser.Count(); i++) {
                            RouteType = -1;
                            switch (toupper(CfgParser[i][0])) {
                            case 'S':
                                RouteType = ROUTE_SEND;
                                break;
                            case 'R':
                                RouteType = ROUTE_RECEIVE;
                                break;
                            case 'B':
                                RouteType = ROUTE_BIDIRECT;
                                break;
                            }
                            if (RouteType != -1)
                                ParsingFailed = !ParseLinkAddr( CfgParser[i] +
                                                                1, LinkAddr );
                            else
                                ParsingFailed = !ParseLinkAddr( CfgParser[i],
                                                                LinkAddr );
                            if (ParsingFailed)
                                break;
                            if (!Links.Select(LinkAddr)) {
                                ParsingFailed = 1;
                                break;
                            }
                            if (LinkedList == NULL) {
                                LinkedList = new SLinked;
                                LastLinked = LinkedList;
                            }
                            else {
                                LastLinked->Next = new SLinked;
                                LastLinked       = LastLinked->Next;
                            }
                            LastLinked->Next = NULL;
                            if (RouteType == -1)
                                LastLinked->RouteType =
                                    Links.Selected->RouteType;
                            else
                                LastLinked->RouteType = RouteType;
                            LastLinked->LinkPtr = Links.Selected;
                        }
                    }
                    else
                        break;
                }
                if (ParsingFailed) {
                    while (LinkedList != NULL) {
                        LastLinked = LinkedList->Next;
                        delete LinkedList;
                        LinkedList = LastLinked;
                    }
                    printf( "Error in file %s, line %ld\n", AreasCfgName,
                            line );
                    ErrorCode = ERROR_SYNTAX;
                    break;
                }
                NewArea.LinkedList = LinkedList;
                NewArea.LastLinked = LastLinked;
                NewArea.LinksChanged = 0;
                Areas.Add( NewArea );
                DoRead = 0;
            }
            else {
                printf( "Error in file %s, line %ld\n", AreasCfgName, line );
                ErrorCode = ERROR_UNKNOWN_KEYWORD;
                break;
            }
        }
        else if (!CfgParser.Eof()) {
            if (DoRead)
                line++;
            printf( "Syntax error in file %s, line %ld\n", AreasCfgName, line );
            ErrorCode = ERROR_SYNTAX;
            break;
        }
    }
    Areas.Sort();
}

void CAdrenalin::ParseCmdLine(int argc, char **argv) {
    int i;
    
    // parse actions
    Action = ACTION_NONE;
    if (argc <= 1) { // no parameters
        Action = ACTION_HELP;
    }
    else {
        for (i = 1; i < argc; i++) {
            if (stricmp( argv[i], "HELP" ) == 0) {
                Action |= ACTION_HELP;
            }
            else if (stricmp( argv[i], "FIX" ) == 0) {
                Action |= ACTION_FIX;
            }
            else if (stricmp( argv[i], "TOSS" ) == 0) {
                Action |= ACTION_TOSS;
            }
            else if (stricmp( argv[i], "REPORT" ) == 0) {
                Action |= ACTION_REPORT;
            }
            else if (stricmp( argv[i], "DEBUG" ) == 0) {
		Action |= ACTION_DEBUG;
            }
	    else if (stricmp( argv[i], "RELINK" ) == 0) {
		Action |= ACTION_RELINK;
		if (i == argc - 1) {
			cerr << "RELINK address absent\n";
			Action |= ACTION_ERROR;
			ErrorCode = ERROR_UNDEFINED_ADDRESS;
			break;
		}
		i++;
		// get relink address
		if (!ParseFTNAddress( argv[i], m_addrRelink )) {
			cerr << "Failed to parse RELINK address\n";
			Action |= ACTION_ERROR;
			ErrorCode = ERROR_WRONG_ADDRESS;
			break;
		}
		i++;
		// get relink robot name
		if (i < argc) {
			if (strlen(argv[i]) > 35) {
				cerr << "Warning: too long robot name, truncated to 35 symbols.\n";
			}
			strncpy( m_szRelinkRobotName, argv[i], 35 );
		}
		else {
			m_szRelinkRobotName[0] = '\0';
		}
	    }
            else if (stricmp( argv[i], "PURGE" ) == 0) {
                Action |= ACTION_PURGE;
                PurgeAreasList = LastPurgeArea = NULL;
                // check if we have areas to purge defined
                if (i < argc-1) {
                    for (++i; i < argc; i++) {
                        if (PurgeAreasList == NULL) {
                            PurgeAreasList = new SStringList;
                            LastPurgeArea = PurgeAreasList;
                        }
                        else {
                            LastPurgeArea->Next = new SStringList;
                            LastPurgeArea = LastPurgeArea->Next;
                        }
                        LastPurgeArea->Next = NULL;
                        LastPurgeArea->String = new char[strlen(argv[i]) + 1];
                        strcpy( LastPurgeArea->String, argv[i] );
                    }
                }
            }
            else if (stricmp( argv[i], "HATCH" ) == 0) {
                Action |= ACTION_HATCH;
                i++;
                if (i == argc) {
                    puts("Error: file to hatch not defined");
                    Action |= ACTION_ERROR;
                    ErrorCode = ERROR_HATCH_NO_FILE_DEF;
                    break;
                }
                strcpy( HatchFile, argv[i] );
                i++;
                if (i == argc) {
                    puts("Error: area to hatch not defined");
                    Action |= ACTION_ERROR;
                    ErrorCode = ERROR_HATCH_NO_AREA_DEF;
                    break;
                }
                strcpy( HatchArea, argv[i] );
                i++;
                if (i == argc) {
                    puts("Error: description not defined");
                    Action |= ACTION_ERROR;
                    ErrorCode = ERROR_HATCH_NO_DESC_DEF;
                    break;
                }
                strcpy( HatchDesc, argv[i] );
                CharToOem( HatchDesc, HatchDesc );
                HatchReplaces[0] = '\0';
                leaveHatchedFile = FALSE;
                i++;
                while (i < argc && argv[i][0] == '-') {
                    if (toupper(argv[i][1]) == 'R') {
                        // replaces
                        i++;
                        if (i == argc) {
                            puts("Error: Replaces not defined");
                            Action |= ACTION_ERROR;
                            ErrorCode = ERROR_HATCH_NO_REPLACES;
                            break;
                        }
                        strcpy( HatchReplaces, argv[i] );
                        i++;
                    }
                    else if (toupper(argv[i][1]) == 'L') {
                    	leaveHatchedFile = TRUE;
                    	i++;
                    }
                    else {
                        printf( "Error: unknown command line option %s\n", argv[i]+1 );
                        Action |= ACTION_ERROR;
                        ErrorCode = ERROR_UNKNOWN_OPT;
                        break;
                    }
                }
                if (i < argc && ((Action & ACTION_ERROR) == 0)) {
                    puts("Error: extra parameter(s) for hatch");
                    Action |= ACTION_ERROR;
                    ErrorCode = ERROR_HATCH_EXTRA_ARG;
                }
                break;
            }
            else {
                printf( "Error: unknown command line parameter %s\n", argv[i] );
                Action |= ACTION_ERROR;
                ErrorCode = ERROR_UNKNOWN_ARG;
                break;
            }
        }
    }
}

void CAdrenalin::PrintHelp() {
    printf( "%s v%lu.%lu%s FileEchoProcessor for Win95/98/NT\n",
            PROG_NAME, PROG_VER_MAJOR, PROG_VER_MINOR, PROG_VER_SUFFIX );
    puts( "Copyright (c) 1999-2000 Andrey Kuprishov 2:5023/27.0@Fidonet\n" );
    puts( "Usage: Adrenalin.exe [options]" );
    puts( "where options are:" );
    puts( "  HELP    prints this help information" );
    puts( "  FIX     scans netmail directory for filefix requests" );
    puts( "  TOSS    scans inbound directory for TIC-files and tosses them" );
    puts( "  PURGE   removes all sent files from passthrough or specified areas," );
    puts( "          e.g. PURGE <area mask>... -<area mask>..." );
    puts( "  HATCH <file> <area> <desc> [-r <Replaces>] [-L]" );
    puts( "          hatches file into fileecho" );
    puts( "            <file>     name of file to hatch" );
    puts( "            <area>     fileecho area to be hatched into" );
    puts( "            <desc>     description of file");
    puts( "            <Replaces> optional Replaces file name");
    puts( "            -L         means that file will be left without copying it into");
    puts( "                       the fileecho directory" );
    puts( "  RELINK <addr> [<robot name>]" );
    puts( "          sends relink request to the specified link using robot name as" );
    puts( "          a receiver name (in a field TO)" );
    puts( "          default robot name is AllFix" );
//    puts( "  REPORT  generates report based on saved information\n" );
//    puts( "You can use many options together, but the sequence of their execution is" );
//    puts( "HELP, FIX, TOSS, PURGE, HATCH, REPORT" );
//    puts( "and doesn't depend of the sequence of command line parameters." );
//    puts( "Also you can't simultaneously use PURGE and HATCH commands." );
}

int CAdrenalin::Execute() {
    if (Action & ACTION_ERROR) { // prevously failed
        return (ErrorCode);
    }
    if (Action & ACTION_HELP) { // help
        PrintHelp();
    }
    if (Action & ACTION_FIX) { // fix
        ProcessFixRequests();
    }
    if (Action & ACTION_TOSS) { // toss
        ProcessInboundTics();
    }
    if (Action & ACTION_RELINK) {	// relink
	    Relink();
    }
    if (Action & ACTION_PURGE) { // purge
        Purge();
    }
    if (Action & ACTION_HATCH) { // hatch
        return (Hatch());
    }
    if (Action & ACTION_REPORT) { // report
        puts( "REPORT under construction" );
    }
    if (Action & ACTION_DEBUG) { // debug
        PrintDebugInfo();
    }
    return (0);
}

int CAdrenalin::ParseLinkAddr(char *s, SFTNAddress &Addr) {
    char FullAddr[24+MAX_DOMAIN_SIZE];
    char TempStr[24+MAX_DOMAIN_SIZE];
    char Domain[MAX_DOMAIN_SIZE];
    int  ColonPresent, SlashPresent, PointPresent, DogPresent;
    int  DogPos;
    int  i;

    if (strlen(s) >= 24+MAX_DOMAIN_SIZE)
        return (0);
    
    ColonPresent = SlashPresent = PointPresent = DogPresent = 0;
    i = 0;
    while (s[i] != '\0') {
        switch (s[i]) {
        case ':':
            ColonPresent = 1;
            break;
        case '/':
            SlashPresent = 1;
            break;
        case '.':
            PointPresent = 1;
            break;
        case '@':
            if (DogPresent == 0) {
                DogPresent = 1;
                DogPos     = i;
            }
            break;
        }
        i++;
    }
    strcpy( TempStr, s );
    if (DogPresent && !PointPresent) {
        i = strlen( TempStr + DogPos + 1);
        if (i > MAX_DOMAIN_SIZE-1)
            i = MAX_DOMAIN_SIZE-1;
        memcpy( Domain, TempStr + DogPos + 1, i );
        Domain[i] = '\0';
        TempStr[DogPos] = '\0';
    }
    if (!SlashPresent) {
        #ifdef SIGNED_ADDR
        sprintf( FullAddr, "%ld/%s", GlobalVars.MainAddr.Net, TempStr );
        #else
        sprintf( FullAddr, "%lu/%s", GlobalVars.MainAddr.Net, TempStr );
        #endif
        strcpy( TempStr, FullAddr );
    }
    if (!ColonPresent) {
        #ifdef SIGNED_ADDR
        sprintf( FullAddr, "%ld:%s", GlobalVars.MainAddr.Zone, TempStr );
        #else
        sprintf( FullAddr, "%lu:%s", GlobalVars.MainAddr.Zone, TempStr );
        #endif
    }
    else
        strcpy( FullAddr, TempStr );

    if (!ParseFTNAddress( FullAddr, Addr ))
        return (0);
    
    if (!DogPresent) {
        strcpy( Addr.Domain, GlobalVars.MainAddr.Domain );
    }
    else if (!PointPresent) {
        strcpy( Addr.Domain, Domain );
    }

    return (*(Addr.Domain) != '\0');
}

int CAdrenalin::ParseLinkAccess(char *s, SLink &Link) {
    int  i;
    char c, LastChar;

    i = 0;
    Link.Priority = 0;
    while (*s >= '0' && *s <= '9' && i < 4) {
        Link.Priority *= 10;
        Link.Priority += *s - '0';
        s++;
        i++;
    }
    if (*s != '/')
        return (0);
    if (Link.Priority > 1024)
        return (0);
    s++;
    if (*s == '\0')
        return (0);

    LastChar    = '\0';
    Link.Groups = 0;
    while (*s != '\0') {
        switch (*s) {
        case '-':
            if (LastChar == '\0')
                return (0);
            s++;
            if (!isalpha(*s))
                return (0);
            c = toupper(*s);
            if (LastChar >= c)
                return (0);
            while (LastChar <= c) {
                Link.Groups |= 1 << (LastChar - 'A');
                LastChar++;
            }
            s++;
            break;
        case '*':
            s++;
            if (*s != '\0')
                return (0);
            if (LastChar == '\0')
                Link.Groups = 0x3FFFFFF;
            else {
                while (LastChar <= 'Z') {
                    Link.Groups |= 1 << (LastChar - 'A');
                    LastChar++;
                }
            }
            break;
        default:
            if (!isalpha(*s))
                return (0);
            LastChar = toupper(*s);
            Link.Groups |= 1 << (LastChar - 'A');
            s++;
        }
    }

    return (1);
}

int CAdrenalin::ParseAreaAccess(char *s, SArea &Area) {
    int  i;

    i = 0;
    Area.Priority = 0;
    while (*s >= '0' && *s <= '9' && i < 4) {
        Area.Priority *= 10;
        Area.Priority += *s - '0';
        s++;
        i++;
    }
    
    if (Area.Priority > 1024)
        return (0);
    
    if (*s == '\0') {
        Area.Group = 0;
        return (1);
    }
    
    if (*s != '/')
        return (0);
    
    s++;
    
    if (!isalpha(*s))
        return (0);
    
    Area.Group |= 1 << (toupper(*s) - 'A');
    
    s++;
    
    return (*s == '\0');
}

void CAdrenalin::PrintDebugInfo() {
    int            i;
    SLinked       *LinkedList;
    unsigned long  bit;
    char           c;
    
    puts( "Debug information" );
    printf( "System: %ld:%ld/%ld.%ld@%s\n", GlobalVars.MainAddr.Zone,
            GlobalVars.MainAddr.Net, GlobalVars.MainAddr.Node,
            GlobalVars.MainAddr.Point, GlobalVars.MainAddr.Domain );
    for (i = 0; i < GlobalVars.AKAAddrNumber; i++) {
        printf( "   AKA: %ld:%ld/%ld.%ld@%s\n", GlobalVars.AKAAddr[i].Zone,
                GlobalVars.AKAAddr[i].Net, GlobalVars.AKAAddr[i].Node,
                GlobalVars.AKAAddr[i].Point, GlobalVars.AKAAddr[i].Domain );
    }
    printf( "SysOp: %s\n", GlobalVars.SysOp );
    puts( "Directories:" );
    printf( "  Inbound: %s\n  Outbound: %s\n  BinkOutbound: %s\n  Netmail: %s\n",
            GlobalVars.InboundPath, GlobalVars.OutboundPath,
            GlobalVars.BinkOutboundPath, GlobalVars.NetmailPath );
    puts( "Flags:" );
    printf( "  after fix:   %s\n  after toss:  %s\n  after hatch: %s\n",
            (*GlobalVars.FixFlagName != '\0') ? GlobalVars.FixFlagName : "none",
            (*GlobalVars.TossFlagName != '\0') ? GlobalVars.TossFlagName : "none",
            (*GlobalVars.HatchFlagName != '\0') ? GlobalVars.HatchFlagName : "none" );
    puts( "Alias(es):" );
    for (i = 0; i < GlobalVars.AliasesNumber; i++) {
        if (i < GlobalVars.AliasesNumber-1)
            printf( "  %s\n", GlobalVars.Aliases[i] );
        else
            printf( "  %s\n", GlobalVars.Aliases[i] );
    }
    puts( "Link(s):");
    Links.SelectFirst();
    while (Links.Selected != NULL) {
        if (Links.Selected->Next != NULL) {
            printf( "  %ld:%ld/%ld.%ld@%s\n     CfgFile: %s\n     Line: %d\n     Password: %s\n     Route: %ld\n     BSO: %ld\n     AreaCreation: %ld\n     Access: %ld/",
                    Links.Selected->Address.Zone, Links.Selected->Address.Net,
                    Links.Selected->Address.Node, Links.Selected->Address.Point,
                    Links.Selected->Address.Domain, Links.Selected->cfgFileName, Links.Selected->Line, Links.Selected->Password,
                    Links.Selected->RouteType, Links.Selected->BSOType,
                    Links.Selected->AllowAreaCreation, Links.Selected->Priority );
            bit = 1;
            c   = 'A';
            while (c <= 'Z') {
                if (Links.Selected->Groups & bit)
                    printf( "%c", c );
                bit <<= 1;
                c++;
            }
            printf( "\n" );
            if (Links.Selected->defaultGroup == 0)
                printf( "     no default group\n");
            else {
                printf( "     default group: ");
                bit = 1;
                c   = 'A';
                while (c <= 'Z') {
                    if (Links.Selected->defaultGroup & bit) {
                        printf( "%c\n", c );
                        break;
                    }
                    bit <<= 1;
                    c++;
                }
            }
            
            printf( "     our address: %ld:%ld/%ld.%ld@%s\n",
                    Links.Selected->OurAddr->Zone,
                    Links.Selected->OurAddr->Net,
                    Links.Selected->OurAddr->Node,
                    Links.Selected->OurAddr->Point,
                    Links.Selected->OurAddr->Domain );
        }
        else {
            printf( "  %ld:%ld/%ld.%ld@%s\n      CfgFile: %s\n      Line: %d\n      Password: %s\n      Route: %ld\n      BSO: %ld\n      AreaCreation: %ld\n      Access: %ld/",
                    Links.Selected->Address.Zone, Links.Selected->Address.Net,
                    Links.Selected->Address.Node, Links.Selected->Address.Point,
                    Links.Selected->Address.Domain, Links.Selected->cfgFileName, Links.Selected->Line, Links.Selected->Password,
                    Links.Selected->RouteType, Links.Selected->BSOType,
                    Links.Selected->AllowAreaCreation, Links.Selected->Priority );
            bit = 1;
            c   = 'A';
            while (c <= 'Z') {
                if (Links.Selected->Groups & bit)
                    printf( "%c", c );
                bit <<= 1;
                c++;
            }
            printf( "\n" );
            if (Links.Selected->defaultGroup == 0)
                printf( "      no default group\n");
            else {
                printf( "      default group: ");
                bit = 1;
                c   = 'A';
                while (c <= 'Z') {
                    if (Links.Selected->defaultGroup & bit) {
                        printf( "%c\n", c );
                        break;
                    }
                    bit <<= 1;
                    c++;
                }
            }
            printf( "      our address: %ld:%ld/%ld.%ld@%s\n",
                    Links.Selected->OurAddr->Zone,
                    Links.Selected->OurAddr->Net,
                    Links.Selected->OurAddr->Node,
                    Links.Selected->OurAddr->Point,
                    Links.Selected->OurAddr->Domain );
        }
	printf( "\011AvailAreas: %s\n", Links.Selected->szfnAvailAreas );
            
        Links.SelectNext();
    }
    puts( "Area(s):" );
    Areas.SelectFirst();
    while (Areas.Selected != NULL) {
        if (Areas.Selected->Next != NULL) {
            printf( "  %s\n     Path: %s\n     Desc: %s\n     Route: %ld\n     Access: %ld/",
                    Areas.Selected->Name, Areas.Selected->Path,
                    Areas.Selected->Desc, Areas.Selected->RouteType,
                    Areas.Selected->Priority );
            if (Areas.Selected->Group == 0)
                puts("any");
            else
                printf( "%c\n", (char)GetOnBit(Areas.Selected->Group) + 'A' );
            printf( "     passthrough: ");
            if (Areas.Selected->PassThrough)
                puts( "ON" );
            else
                puts( "OFF" );
            printf( "     veto manual purge: " );
            if (Areas.Selected->VetoManualPurge)
                puts( "ON" );
            else
                puts( "OFF" );
            LinkedList = Areas.Selected->LinkedList;
            if (LinkedList == NULL)
                puts("     No links");
            else {
                puts("     Linked:");
                while (LinkedList != NULL) {
                    if (LinkedList->Next != NULL) {
                        printf( "         %ld:%ld/%ld.%ld@%s, %ld\n",
                                LinkedList->LinkPtr->Address.Zone,
                                LinkedList->LinkPtr->Address.Net,
                                LinkedList->LinkPtr->Address.Node,
                                LinkedList->LinkPtr->Address.Point,
                                LinkedList->LinkPtr->Address.Domain,
                                LinkedList->RouteType );
                    }
                    else {
                        printf( "         %ld:%ld/%ld.%ld@%s, %ld\n",
                                LinkedList->LinkPtr->Address.Zone,
                                LinkedList->LinkPtr->Address.Net,
                                LinkedList->LinkPtr->Address.Node,
                                LinkedList->LinkPtr->Address.Point,
                                LinkedList->LinkPtr->Address.Domain,
                                LinkedList->RouteType );
                    }
                    LinkedList = LinkedList->Next;
                }
            }
        }
        else {
            printf( "  %s\n      Path: %s\n      Desc: %s\n      Route: %ld\n      Access: %ld/",
                    Areas.Selected->Name, Areas.Selected->Path,
                    Areas.Selected->Desc, Areas.Selected->RouteType,
                    Areas.Selected->Priority );
            if (Areas.Selected->Group == 0)
                puts("any");
            else
                printf( "%c\n", (char)GetOnBit(Areas.Selected->Group) + 'A' );
            printf( "      passthrough: " );
            if (Areas.Selected->PassThrough)
                puts( "ON" );
            else
                puts( "OFF" );
            printf( "      veto manual purge: " );
            if (Areas.Selected->VetoManualPurge)
                puts( "ON" );
            else
                puts( "OFF" );
            LinkedList = Areas.Selected->LinkedList;
            if (LinkedList == NULL)
                puts("      No links");
            else {
                puts("      Linked:");
                while (LinkedList != NULL) {
                    if (LinkedList->Next != NULL) {
                        printf( "          %ld:%ld/%ld.%ld@%s, %ld\n",
                                LinkedList->LinkPtr->Address.Zone,
                                LinkedList->LinkPtr->Address.Net,
                                LinkedList->LinkPtr->Address.Node,
                                LinkedList->LinkPtr->Address.Point,
                                LinkedList->LinkPtr->Address.Domain,
                                LinkedList->RouteType );
                    }
                    else {
                        printf( "          %ld:%ld/%ld.%ld@%s, %ld\n",
                                LinkedList->LinkPtr->Address.Zone,
                                LinkedList->LinkPtr->Address.Net,
                                LinkedList->LinkPtr->Address.Node,
                                LinkedList->LinkPtr->Address.Point,
                                LinkedList->LinkPtr->Address.Domain,
                                LinkedList->RouteType );
                    }
                    LinkedList = LinkedList->Next;
                }
            }
        }
        Areas.SelectNext();
    }
}

void CAdrenalin::ProcessFixRequests() {
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFindFile;
    BOOL            FileFound;
    char            FindFileMask[MAX_PATH];
    char            FullFileName[MAX_PATH];
    char            LogLine[MAX_PATH];
    int             MailPathLen;
    int             ProcessedMsgNumber;
    SStringList    *MsgFileNameList, *LastMsgFileName, *TmpMsgFileName;
    CMsg            Msg;
    SMsgDataList   *CurMsgData;
    CStringList     BodyLines;
    char            Line[LINE_SIZE];
    char            TearLine[LINE_SIZE];
    char            ReplySubject[71];
    int             i;
    int             SubscribeNumber, FoundNumber;
    int             LinksChanged;
    int             NewMessagesCreated; // if created we should create FixFlag

    
    sprintf( LogLine, "Processing Fix requests in directory %s:\n",
             GlobalVars.NetmailPath );
    
    GlobalVars.Log << LogLine;
    cout << LogLine;
    
    strcpy( FindFileMask, GlobalVars.NetmailPath );
    strcpy( FullFileName, GlobalVars.NetmailPath );
    MailPathLen = strlen( GlobalVars.NetmailPath );
    strcpy( FindFileMask + MailPathLen, "*.msg" );

    ProcessedMsgNumber = 0;

    MsgFileNameList = LastMsgFileName = NULL;
    hFindFile = FindFirstFile( FindFileMask, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    while (FileFound) {
        if (MsgFileNameList) {
            LastMsgFileName->Next = new SStringList;
            LastMsgFileName = LastMsgFileName->Next;
        }
        else {
            MsgFileNameList = new SStringList;
            LastMsgFileName = MsgFileNameList;
        }
        LastMsgFileName->Next = NULL;
        LastMsgFileName->String = new char[MAX_PATH];
        strcpy( LastMsgFileName->String, FullFileName );
        strcpy( LastMsgFileName->String + MailPathLen, FindFileData.cFileName );
        FileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );
    
    sprintf( TearLine, "--- %s v%lu.%lu%s", PROG_NAME,
             PROG_VER_MAJOR, PROG_VER_MINOR, PROG_VER_SUFFIX );
    
    LinksChanged = 0;
    NewMessagesCreated = 0;
    while (MsgFileNameList != NULL) {
        if (GlobalVars.LogMsgNames == 1)
            GlobalVars.Log << "* msg: "<< MsgFileNameList->String << endl;
        if (Msg.Read( MsgFileNameList->String )) {
            if (GlobalVars.LogMsgNames == 1)
                GlobalVars.Log << "       read\n";
            if (GlobalVars.KillProcessedMail)
                Msg.Delete();
            else
                Msg.SaveAsReceived();
            // check link
            if (Links.Select(Msg.From)) {
                if (AddrCmp( Msg.To, *(Links.Selected->OurAddr) ) != 0) {
                    // ignore message;
                    // only add line into log file
                    GlobalVars.Log << "  message not for corresponding AKA:\n";
                    sprintf( LogLine,
                             "    from %lu:%lu/%lu.%lu@%s to %lu:%lu/%lu.%lu@%s\n",
                             Msg.From.Zone, Msg.From.Net, Msg.From.Node,
                             Msg.From.Point, Msg.From.Domain,
                             Msg.To.Zone, Msg.To.Net, Msg.To.Node,
                             Msg.To.Point, Msg.To.Domain );
                    GlobalVars.Log << LogLine;
                }
                else if (Links.Selected->Activity == ACTIVITY_UNAVAILABLE) {
                    // unavailable system
                    sprintf( Line, Language[LNG_UNAVAILABLE_INFO],
                             GlobalVars.SysOp );
                    BodyLines.Add( Line );
                    BodyLines.Add( TearLine );
                    Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                               Language[LNG_UNAVAILABLE_INFO_SUBJ],
                               BodyLines.Strings );
                    NewMessagesCreated = 1;
                    BodyLines.Free();
                    GlobalVars.Log << "  message from unavailable system:\n";
                    sprintf( LogLine,
                             "    %lu:%lu/%lu.%lu@%s\n",
                             Msg.From.Zone, Msg.From.Net, Msg.From.Node,
                             Msg.From.Point, Msg.From.Domain );
                    GlobalVars.Log << LogLine;
                }
                else if (stricmp( Links.Selected->Password, Msg.Header.Subject )
                    == 0)
                {
                    if (GlobalVars.MsgData == NULL) {
                        // no commands used
                        BodyLines.Add( Language[LNG_NO_COMMANDS_USED] );
                        BodyLines.Add( TearLine );
                        Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                   Language[LNG_NO_COMMANDS_USED_SUBJ],
                                   BodyLines.Strings );
                        NewMessagesCreated = 1;
                        BodyLines.Free();
                    }
                    else {
                        // process unknown commands
                        BodyLines.Add( Language[LNG_UNKNOWN_COMMANDS] );
                        SubscribeNumber = 0;
                        CurMsgData = GlobalVars.MsgData;
                        while (CurMsgData != NULL) {
                            if (CurMsgData->DataType == MSGDATA_COMMAND &&
                                stricmp( CurMsgData->Data, "help") != 0 &&
                                stricmp( CurMsgData->Data, "list") != 0 &&
                                stricmp( CurMsgData->Data, "query") != 0 &&
                                stricmp( CurMsgData->Data, "linked") != 0 &&
                                stricmp( CurMsgData->Data, "unlinked") != 0 &&
                                stricmp( CurMsgData->Data, "lock") != 0 &&
                                stricmp( CurMsgData->Data, "unlock") != 0)
                            {
                                sprintf( Line, "  %%%s", CurMsgData->Data );
                                BodyLines.Add( Line );
                                SubscribeNumber++;
                                CurMsgData->DataType = MSGDATA_NONE;
                            }
                            CurMsgData = CurMsgData->Next;
                        }
                        if (SubscribeNumber > 0) {
                            BodyLines.Add( "" );
                            BodyLines.Add( TearLine );
                            Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                       Language[LNG_UNKNOWN_COMMANDS_SUBJ],
                                       BodyLines.Strings );
                            NewMessagesCreated = 1;
                        }
                        BodyLines.Free();
                        // process known commands and subscriptions
                        BodyLines.Add( Language[LNG_SUBSCRIPTION_CHANGES] );
                        SubscribeNumber = 0;
                        CurMsgData = GlobalVars.MsgData;
                        while (CurMsgData != NULL) {
                            switch (CurMsgData->DataType) {
                            case MSGDATA_SUBSCRIBE:
                                SubscribeNumber++;
                                Areas.SelectFirst();
                                FoundNumber = 0;
                                while (Areas.Selected != NULL) {
                                    if (CheckMask( Areas.Selected->Name,
                                                   CurMsgData->Data))
                                    {
                                        if (Areas.Linked( Links.Selected )) {
                                            sprintf( Line, Language[LNG_AREA_ALREADY_LINKED],
                                                     Areas.Selected->Name );
                                        }
                                        else {
                                            if (Areas.Link( Links.Selected ))
                                                sprintf( Line, Language[LNG_AREA_LINKED],
                                                         Areas.Selected->Name );
                                            else
                                                sprintf( Line, Language[LNG_AREA_NO_ACCESS],
                                                         Areas.Selected->Name );
                                        }
                                        BodyLines.Add( Line );
                                        FoundNumber++;
                                    }
                                    Areas.SelectNext();
                                }
                                if (FoundNumber == 0) {
					if (IsMask(CurMsgData->Data)) {
						sprintf( Line, Language[LNG_AREA_MASK_NOT_FOUND],
							 CurMsgData->Data );
						BodyLines.Add( Line );
					}
					else {
						char	szLine[LINE_SIZE];
						// trying to forward request
						Links.SelectFirst();
						while (Links.Selected != NULL) {
							// skip requestor
							if (AddrCmp( Links.Selected->Address, Msg.From ) == 0) {
								Links.SelectNext();
								continue;
							}
							// check AvailAreas
							if (Links.Selected->szfnAvailAreas[0] == '\0') {
								Links.SelectNext();
								continue;
							}
							// open areas file
							ifstream	in(Links.Selected->szfnAvailAreas);
							if (in.fail()) {
								Links.SelectNext();
								continue;
							}
							bool	fAreaFound = false;
							while (!in.eof() && !fAreaFound) {
								in.getline( szLine, LINE_SIZE );
								if (in.fail())
									break;
								fAreaFound = (stricmp( szLine, CurMsgData->Data ) == 0);
							}
							in.close();
							if (fAreaFound)
								break;
							// check next link
							Links.SelectNext();
						}
						if (Links.Selected != NULL) {
							// link found
							// write message to this link
							CMsg		msg;
							CStringList	lstBody;
							SLink*		pLink = Links.Selected;
							// print log message
							sprintf( szLine, "Forwarding request from %lu:%lu/%lu.%lu@%s to %lu:%lu/%lu.%lu@%s\n\011requesting %s\n",
								Msg.From.Zone, Msg.From.Net, Msg.From.Node, Msg.From.Point, Msg.From.Domain,
								pLink->Address.Zone, pLink->Address.Net, pLink->Address.Node, pLink->Address.Point, pLink->Address.Domain,
								CurMsgData->Data );
							GlobalVars.Log << szLine;

							sprintf( szLine, "+%s", CurMsgData->Data );
							lstBody.Add( szLine );
							lstBody.Add( "" );
							sprintf( szLine, "--- %s v%lu.%lu%s", PROG_NAME, PROG_VER_MAJOR, PROG_VER_MINOR, PROG_VER_SUFFIX );
							lstBody.Add( szLine );
							MakeDirStructure( GlobalVars.NetmailPath );
							// from/to fields reversed because we will use reply method
							msg.From = pLink->Address;
							msg.To   = *(pLink->OurAddr);
							// FIX: add robot names
							if (pLink->m_szRobotName[0] != '\0')
								strcpy( msg.Header.FromUserName, pLink->m_szRobotName );
							else
								strcpy( msg.Header.FromUserName, "AllFix" );
							strcpy( msg.Header.ToUserName, GlobalVars.SysOp );
							msg.FreeMsgNumber = Msg.FreeMsgNumber;
							msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, pLink->Password, lstBody.Strings );
							Msg.FreeMsgNumber++;
							GlobalVars.Log << "Forward request message created.\n";
							// add line
							sprintf( Line, Language[LNG_REQUEST_FORWARDED],
								CurMsgData->Data );
							BodyLines.Add( Line );
							// 2. Let's keep track of requestors
							if (!m_fForwardsRead) {
								LoadForwards();
							}
							// 3. Add new forward info, format of date is DDMMYY
							SForwardInfo	fwdInfo;
							SYSTEMTIME	systemTime;
							GetLocalTime( &systemTime );
							sprintf( fwdInfo.m_szDate, "%02d%02d%02d", systemTime.wDay, systemTime.wMonth, systemTime.wYear % 100 );
							sprintf( fwdInfo.m_szLink, "%d:%d/%d.%d@%s", Msg.From.Zone, Msg.From.Net, Msg.From.Node, Msg.From.Point, Msg.From.Domain );
							strcpy( fwdInfo.m_szArea, CurMsgData->Data );
							m_vectForwardInfo.push_back( fwdInfo );
							m_fForwardsChanged = true;
						}
						else {
							sprintf( Line, Language[LNG_AREA_NOT_FOUND],
								CurMsgData->Data );
							BodyLines.Add( Line );
						}
						// restore selection
						Links.Select( Msg.From );

					}
                                    
                                }
                                break;
                            case MSGDATA_UNSUBSCRIBE:
                                SubscribeNumber++;
                                Areas.SelectFirst();
                                FoundNumber = 0;
                                while (Areas.Selected != NULL) {
                                    if (CheckMask( Areas.Selected->Name,
                                                   CurMsgData->Data))
                                    {
                                        if (Areas.Unlink( Links.Selected )) {
                                            sprintf( Line, Language[LNG_AREA_UNLINKED],
                                                     Areas.Selected->Name );
                                            BodyLines.Add( Line );
                                            FoundNumber++;
                                        }
                                        else if (!IsMask(CurMsgData->Data)) {
                                            sprintf( Line, Language[LNG_AREA_NOT_LINKED],
                                                     Areas.Selected->Name );
                                            BodyLines.Add( Line );
                                            FoundNumber++;
                                        }
                                    }
                                    Areas.SelectNext();
                                }
                                if (FoundNumber == 0) {
                                    if (IsMask(CurMsgData->Data))
                                        sprintf( Line, Language[LNG_AREA_MASK_NOT_FOUND],
                                                 CurMsgData->Data );
                                    else
                                        sprintf( Line, Language[LNG_AREA_NOT_FOUND],
                                                 CurMsgData->Data );
                                    BodyLines.Add( Line );
                                }
                                break;
                            case MSGDATA_COMMAND:
                                if (SubscribeNumber > 0) {
                                    BodyLines.Add( "" );
                                    BodyLines.Add( TearLine );
                                    Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                               Language[LNG_SUBSCRIPTION_CHANGES_SUBJ],
                                               BodyLines.Strings );
                                    NewMessagesCreated = 1;
                                    SubscribeNumber = 0;
                                }
                                BodyLines.Free();
                                BodyLines.Add("");
                                if (stricmp( CurMsgData->Data, "help" ) == 0) {
                                    strcpy( ReplySubject, Language[LNG_HELP_SUBJ] );
                                    BodyLines.Add( Language[LNG_HELP_HEAD] );
                                    if (GlobalVars.AliasesNumber > 1) {
                                        BodyLines.Add( Language[LNG_HELP_ALIASES] );
                                        for (i = 1; i < GlobalVars.AliasesNumber; i++) {
                                            sprintf( Line, "  %s", GlobalVars.Aliases[i] );
                                            BodyLines.Add( Line );
                                        }
                                        BodyLines.Add( Language[LNG_HELP] );
                                    }
                                }
                                else if (stricmp( CurMsgData->Data, "list" ) == 0) {
                                    strcpy( ReplySubject, Language[LNG_LIST_SUBJ] );
                                    i = 0;
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.CanAccess( Links.Selected ))
                                            i++;
                                        Areas.SelectNext();
                                    }
                                    if (i == 0)
                                        BodyLines.Add( Language[LNG_LIST_EMPTY] );
                                    else {
                                        sprintf( Line, Language[LNG_LIST_TITLE], i );
                                        BodyLines.Add( Line );
                                    }
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.CanAccess( Links.Selected )) {
                                            Areas.GetNameDesc( Line );
                                            if (Areas.Linked( Links.Selected )) {
                                                switch (Areas.GetRouteType( Links.Selected)) {
                                                case ROUTE_RECEIVE:
                                                    Line[0] = 'R';
                                                    break;
                                                case ROUTE_SEND:
                                                    Line[0] = 'S';
                                                    break;
                                                case ROUTE_BIDIRECT:
                                                    Line[0] = 'B';
                                                    break;
                                                }
                                            }
                                            BodyLines.Add( Line );
                                        }
                                        Areas.SelectNext();
                                    }
                                    if (i > 0) {
                                        BodyLines.Add( Language[LNG_LIST_MARKS_TITLE] );
                                        BodyLines.Add( Language[LNG_MARKS] );
                                    }
                                }
                                else if (stricmp( CurMsgData->Data, "query" ) == 0 ||
                                         stricmp( CurMsgData->Data, "linked" ) == 0)
                                {
                                    strcpy( ReplySubject, Language[LNG_LINKED_SUBJ] );
                                    i = 0;
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.Linked( Links.Selected ))
                                            i++;
                                        Areas.SelectNext();
                                    }
                                    if (i == 0)
                                        BodyLines.Add( Language[LNG_LINKED_EMPTY] );
                                    else {
                                        sprintf( Line, Language[LNG_LINKED_TITLE], i );
                                        BodyLines.Add( Line );
                                    }
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.Linked( Links.Selected )) {
                                            Areas.GetNameDesc( Line );
                                            if (Areas.Linked( Links.Selected )) {
                                                switch (Areas.GetRouteType( Links.Selected)) {
                                                case ROUTE_RECEIVE:
                                                    Line[0] = 'R';
                                                    break;
                                                case ROUTE_SEND:
                                                    Line[0] = 'S';
                                                    break;
                                                case ROUTE_BIDIRECT:
                                                    Line[0] = 'B';
                                                    break;
                                                }
                                            }
                                            BodyLines.Add( Line );
                                        }
                                        Areas.SelectNext();
                                    }
                                    if (i > 0) {
                                        BodyLines.Add( Language[LNG_LINKED_MARKS_TITLE] );
                                        BodyLines.Add( Language[LNG_MARKS] );
                                    }
                                }
                                else if (stricmp( CurMsgData->Data, "unlinked" ) == 0) {
                                    strcpy( ReplySubject, Language[LNG_LINKED_SUBJ] );
                                    i = 0;
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.CanAccess( Links.Selected) &&
                                            !Areas.Linked( Links.Selected ))
                                            i++;
                                        Areas.SelectNext();
                                    }
                                    if (i == 0)
                                        BodyLines.Add( Language[LNG_UNLINKED_EMPTY]);
                                    else {
                                        sprintf( Line, Language[LNG_UNLINKED_TITLE], i );
                                        BodyLines.Add( Line );
                                    }
                                    Areas.SelectFirst();
                                    while (Areas.Selected != NULL) {
                                        if (Areas.CanAccess( Links.Selected) &&
                                            !Areas.Linked( Links.Selected ))
                                        {
                                            Areas.GetNameDesc( Line );
                                            BodyLines.Add( Line );
                                        }
                                        Areas.SelectNext();
                                    }
                                }
                                else if (stricmp( CurMsgData->Data, "lock" ) == 0) {
                                    strcpy( ReplySubject, Language[LNG_LOCK_SUBJ] );
                                    if (Links.Selected->Activity == ACTIVITY_USUAL) {
                                        Links.Selected->Activity = ACTIVITY_PASSIVE;
                                        LinksChanged = 1;
                                        BodyLines.Add( Language[LNG_LOCK] );
                                    }
                                    else {
                                        BodyLines.Add( Language[LNG_ALREADY_LOCKED] );
                                    }
                                    BodyLines.Add( Language[LNG_UNLOCK_NOTIFY] );
                                }
                                else if (stricmp( CurMsgData->Data, "unlock" ) == 0) {
                                    strcpy( ReplySubject, Language[LNG_UNLOCK_SUBJ] );
                                    if (Links.Selected->Activity == ACTIVITY_PASSIVE) {
                                        Links.Selected->Activity = ACTIVITY_USUAL;
                                        LinksChanged = 1;
                                        BodyLines.Add( Language[LNG_UNLOCK] );
                                    }
                                    else {
                                        BodyLines.Add( Language[LNG_NOT_LOCKED] );
                                    }
                                    BodyLines.Add( Language[LNG_LOCK_NOTIFY] );
                                }
                                BodyLines.Add( "" );
                                BodyLines.Add( TearLine );
                                Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                           ReplySubject, BodyLines.Strings );
                                NewMessagesCreated = 1;
                                BodyLines.Free();
                                BodyLines.Add( Language[LNG_SUBSCRIPTION_CHANGES] );
                                SubscribeNumber = 0;
                                break;
                            };
                            CurMsgData = CurMsgData->Next;
                        }
                        if (SubscribeNumber > 0) {
                            BodyLines.Add( "" );
                            BodyLines.Add( TearLine );
                            Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                                       Language[LNG_SUBSCRIPTION_CHANGES_SUBJ],
                                       BodyLines.Strings );
                            NewMessagesCreated = 1;
                            SubscribeNumber = 0;
                        }
                    }
                }
                else {
                    // incorrect password
                    sprintf( Line, Language[LNG_INCORRECT_PASSWORD],
                             Msg.Header.Subject, GlobalVars.SysOp );
                    BodyLines.Add( Line );
                    BodyLines.Add( TearLine );
                    Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                               Language[LNG_ACCESS_DENIED_SUBJ],
                               BodyLines.Strings );
                    NewMessagesCreated = 1;
                    BodyLines.Free();
                }
            }
            else {
                // link is not listed
                sprintf( Line, Language[LNG_UNKNOWN_USER],
                         Msg.Header.ToUserName, GlobalVars.SysOp );
                BodyLines.Add( Line );
                BodyLines.Add( TearLine );
                Msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0,
                           Language[LNG_ACCESS_DENIED_SUBJ],
                           BodyLines.Strings );
                NewMessagesCreated = 1;
                BodyLines.Free();
            }
            GlobalVars.FreeMsgData();
            ProcessedMsgNumber++;
        }
        else {
            if (GlobalVars.LogMsgNames)
                GlobalVars.Log << "       failed\n";
        }
        TmpMsgFileName = MsgFileNameList->Next;
        delete [] MsgFileNameList->String;
        delete MsgFileNameList;
        MsgFileNameList = TmpMsgFileName;
    }
    if (NewMessagesCreated && *GlobalVars.FixFlagName != '\0')
        CreateFileFlag( GlobalVars.FixFlagName );
    if (LinksChanged)
        SaveCfg();
    GlobalVars.Log << "Fix done.\n";
}


void CAdrenalin::MoveTicToBads(char *FullTicName, int TicNameFrom) {
    char DestTicName[MAX_PATH];
    char LogLine[LINE_SIZE];
    int  i, j, Len;
    int  NameFound;
    int  NameChanged;

    strcpy( DestTicName, GlobalVars.BadTicsPath );
    strcat( DestTicName, FullTicName + TicNameFrom );

    Len = strlen(DestTicName);

    NameFound   = 0;
    NameChanged = 0;
    
    do {

        // trying to check the file
        if (access( DestTicName, 0) != 0)
            NameFound = 1;
        else {
            
            // form next name
            i = Len-1;
            j = 0;
        
            while (j <= 2) {
                if (DestTicName[i] < '0') {
                    DestTicName[i] = '0';
                    break;
                }
            
                if (DestTicName[i] < '9') {
                    DestTicName[i]++;
                    break;
                }
            
                if (DestTicName[i] < 'A') {
                    DestTicName[i] = 'a';
                    break;
                }
                
                if (DestTicName[i] < 'Z') {
                    DestTicName[i]++;
                    DestTicName[i] += 'a' - 'A';
                    break;
                }
            
                if (DestTicName[i] < 'a') {
                    DestTicName[i] = 'a';
                    break;
                }
            
                if (DestTicName[i] < 'z') {
                    DestTicName[i]++;
                    break;
                }
        
                DestTicName[i] = '0';
                i--;
                j++;
            }

            // check if failed to form new name
            if (j > 2)
                break;
            NameChanged = 1;
        }
    } while (!NameFound);

    if (!NameFound || !MoveFile( FullTicName, DestTicName ))
        sprintf( LogLine, "                failed to move TIC into BadTics\n");
    else {
        if (NameChanged) {
            sprintf( LogLine, "                TIC was renamed to %s and moved into BadTics\n",
                     DestTicName + TicNameFrom + Len - strlen(FullTicName) );
        }
        else
            sprintf( LogLine, "                TIC was moved into BadTics\n");
    }
    
    GlobalVars.Log << LogLine;
    cout << LogLine;
}


void CAdrenalin::ProcessInboundTics() {
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFindFile;
    BOOL            FileFound;
    char            FindFileMask[MAX_PATH];
    char            FullFileName[MAX_PATH];
    char            DestFileName[MAX_PATH];
    char            NewAreasFileName[MAX_PATH];
    int             InboundPathLen;
    int             ProcessedTicNumber;
    SStringList    *TicsList, *LastTic, *TmpTic;
    CTic            Tic;
    int             i;
    SLinked        *Linked;
    SLinked        *CurLinked;
    SAddrList      *CurSeenby;
    SLinked        *ForwardList, *LastForward;
    ofstream        out;
    ofstream        NewAreas;
    UINT            RealCRC32;
    int             TicNameFrom;
    char            LogLine[LINE_SIZE];
    char            TimeLine[LINE_SIZE];
    int            *AKASeen;
    int             ReplaceDone;
    char            ReplaceLine[LINE_SIZE];
    ifstream        in;
    CCfgParser     *TrafficParser;
    char           *s;
    int             Traffic;
    int             NewFilesArrived;
    int             TossedForLinks; // if yes, we should create TossFlag
    
    sprintf( LogLine, "Tossing inbound TICs in directory %s:\n",
             GlobalVars.InboundPath );
    GlobalVars.Log << LogLine;
    cout << LogLine;

    // make all fecho directories and set areas traffic to zero
    Areas.SelectFirst();
    while (Areas.Selected != NULL) {
        MakeDirStructure( Areas.Selected->Path );
        Areas.Selected->Traffic = 0;
        Areas.SelectNext();
    }

    // make tics directories
    MakeDirStructure( GlobalVars.BadTicsPath );
    MakeDirStructure( GlobalVars.NewTicsPath );

    // read areas traffic
    
    strcpy( FullFileName, StartupPath );
    strcat( FullFileName, "traffic.dat" );

    *TimeLine = '\0';
    TrafficParser = new CCfgParser( FullFileName );
    if (!TrafficParser->Failed()) {
        // read date line
        TrafficParser->ReadLine();
        if (!TrafficParser->Failed())
            strcpy( TimeLine, TrafficParser->String() );
        // read data lines
        while (!TrafficParser->Eof()) {
            TrafficParser->ReadLine();
            if (TrafficParser->Failed())
                break;
            if (TrafficParser->Count() != 2)
                break;
            if (!Areas.Select( (*TrafficParser)[0] ))
                continue;
            s = (*TrafficParser)[1];
            Traffic = 0;
            while (isdigit(*s)) {
                Traffic *= 10;
                Traffic += (*s - '0');
                s++;
            }
            if (*s != '\0')
                break;
            Areas.Selected->Traffic = Traffic;
        }
    }
    delete TrafficParser;
    
    strcpy( FindFileMask, GlobalVars.InboundPath );
    strcpy( FullFileName, GlobalVars.InboundPath );
    InboundPathLen = strlen( GlobalVars.InboundPath );
    strcat( FindFileMask, "*.ti?" );

    ProcessedTicNumber = 0;

    TicsList = LastTic = NULL;
    hFindFile = FindFirstFile( FindFileMask, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    while (FileFound) {
        if (!(FindFileData.dwFileAttributes &
              (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
               FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_READONLY |
               FILE_ATTRIBUTE_TEMPORARY)))
        {
            if (TicsList) {
                LastTic->Next = new SStringList;
                LastTic       = LastTic->Next;
            }
            else {
                TicsList = new SStringList;
                LastTic  = TicsList;
            }
            LastTic->Next = NULL;
            LastTic->String = new char[MAX_PATH];
            strcpy( LastTic->String, FullFileName );
            strcpy( LastTic->String + InboundPathLen,
                    FindFileData.cFileName );
        }
        FileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );

    // allocate memory for seen AKA
    AKASeen = new int[GlobalVars.AKAAddrNumber + 1];
    // open newfiles list
    strcpy( FullFileName, StartupPath );
    strcat( FullFileName, "newfiles.txt" );
    out.open( FullFileName, ios::app );
    if (!out)
        out.open( FullFileName );
    // prepare file name for new areas
    strcpy( NewAreasFileName, StartupPath );
    strcat( NewAreasFileName, "newareas.txt" );
    // do not update traffic.dat if no new files arrived
    NewFilesArrived = 0;
    TossedForLinks  = 0;
    // process TICs
    while (TicsList != NULL) {
        // get name of tic without path
        for ( TicNameFrom = -1, i = 0; TicsList->String[i] != '\0';
              i++)
        {
            if (TicsList->String[i] == '\\')
                TicNameFrom = i;
        }
        TicNameFrom++;
        // trying to parse TIC
        if (Tic.Read( TicsList->String )) {
            do {
                // check link
                
                if (!Links.Select( Tic.From )) {
                    sprintf( LogLine,
                             "! %12s: unknown sender (%lu:%lu/%lu.%lu@%s)\n",
                             TicsList->String + TicNameFrom, Tic.From.Zone,
                             Tic.From.Net, Tic.From.Node, Tic.From.Point,
                             Tic.From.Domain );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    MoveTicToBads( TicsList->String, TicNameFrom );
                    break;
                }
                // check link activity
                
                if (Links.Selected->Activity == ACTIVITY_UNAVAILABLE) {
                    sprintf( LogLine,
                             "! %12s: unavailable sender (%lu:%lu/%lu.%lu@%s)\n",
                             TicsList->String + TicNameFrom, Tic.From.Zone,
                             Tic.From.Net, Tic.From.Node, Tic.From.Point,
                             Tic.From.Domain );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    MoveTicToBads( TicsList->String, TicNameFrom );
                    break;
                }
                // check password
                
                if (stricmp( Links.Selected->Password, Tic.PW ) != 0) {
                    sprintf( LogLine,
                             "! %12s: password error (%s instead of %s)\n",
                             TicsList->String + TicNameFrom, Tic.PW,
                             Links.Selected->Password );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    MoveTicToBads( TicsList->String, TicNameFrom );
                    break;
                }
                // form full file name of inbound file
                strcpy( FullFileName, GlobalVars.InboundPath );
                strcat( FullFileName, Tic.File );
                // check file
                
                hFindFile = FindFirstFile( FullFileName, &FindFileData );
                FileFound = (hFindFile != INVALID_HANDLE_VALUE);
                if (!FileFound) {
                    sprintf( LogLine, "! %12s: file %s not found\n",
                             TicsList->String + TicNameFrom, Tic.File );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                FindClose( hFindFile );
                // check file finished or not
                if (FindFileData.dwFileAttributes &
                    (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
                     FILE_ATTRIBUTE_OFFLINE   | FILE_ATTRIBUTE_READONLY |
                     FILE_ATTRIBUTE_TEMPORARY))
                {
                    sprintf( LogLine, "! %12s: file %s is not accesible\n",
                             TicsList->String + TicNameFrom, Tic.File );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                // check size
                
                if (Tic.Size != -1 && Tic.Size !=
                    ((FindFileData.nFileSizeHigh * MAXDWORD) +
                     FindFileData.nFileSizeLow))
                {
                    sprintf( LogLine,
                             "! %12s: file size error (%ld instead of %ld)\n",
                             TicsList->String + TicNameFrom,
                             (FindFileData.nFileSizeHigh * MAXDWORD) +
                             FindFileData.nFileSizeLow, Tic.Size );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                // check crc
                
                if (!CalcCRC32( FullFileName, RealCRC32 )) {
                    sprintf( LogLine,
                             "! %12s: failed to calculate CRC32 of file %s\n",
                             TicsList->String + TicNameFrom, Tic.File );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                if (RealCRC32 != Tic.CRC32) {
                    sprintf( LogLine,
                             "! %12s: CRC32 error (%08lX instead of %08lX)\n",
                             TicsList->String + TicNameFrom, RealCRC32,
                             Tic.CRC32 );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                // check area
                
                if (Areas.Select( Tic.Area )) {
                    Linked = Areas.Linked( Links.Selected );
                    if (Linked == NULL) {
                        sprintf( LogLine,
                                 "! %12s: %lu:%lu/%lu.%lu@%s isn't linked to %s\n",
                                 TicsList->String + TicNameFrom, Tic.From.Zone,
                                 Tic.From.Net, Tic.From.Node, Tic.From.Point,
                                 Tic.From.Domain, Areas.Selected->Name );
                        GlobalVars.Log << LogLine;
                        cout << LogLine;
                        MoveTicToBads( TicsList->String, TicNameFrom );
                        break;
                    }
                    if ((Linked->RouteType & ROUTE_SEND) == 0) {
                        sprintf( LogLine,
                                 "! %12s: %lu:%lu/%lu.%lu@%s can't send into %s\n",
                                 TicsList->String + TicNameFrom, Tic.From.Zone,
                                 Tic.From.Net, Tic.From.Node, Tic.From.Point,
                                 Tic.From.Domain, Areas.Selected->Name );
                        GlobalVars.Log << LogLine;
                        cout << LogLine;
                        MoveTicToBads( TicsList->String, TicNameFrom );
                        break;
                    }
                }
                else {
                    if (Links.Selected->AllowAreaCreation) {
                        if (!Areas.IsValidName( Tic.Area )) {
                            sprintf( LogLine, "! %12s: area name %s is wrong\n",
                                     TicsList->String + TicNameFrom,
                                     Tic.Area );
                            GlobalVars.Log << LogLine;
                            cout << LogLine;
                            MoveTicToBads( TicsList->String, TicNameFrom );
                            break;
                        }
                        Areas.AutoCreate( Tic.Area, Links.Selected );
                        Linked = Areas.Linked( Links.Selected );
                        sprintf( LogLine,
                                 "+ area %s created by %lu:%lu/%lu.%lu@%s\n",
                                 Areas.Selected->Name, Tic.From.Zone,
                                 Tic.From.Net, Tic.From.Node, Tic.From.Point,
                                 Tic.From.Domain );
                        GlobalVars.Log << LogLine;
                        cout << LogLine;
                        NewAreas.open( NewAreasFileName, ios::app );
                        if (!NewAreas)
                            NewAreas.open( NewAreasFileName );
                        NewAreas << "Fileecho area " << Areas.Selected->Name
                            << " created by " << Tic.From.Zone << ":"
                            << Tic.From.Net << "/" << Tic.From.Node;
                        if (Tic.From.Point != 0)
                            NewAreas << "." << Tic.From.Point;
                        NewAreas << "@" << Tic.From.Domain << endl;
			// 2. We should check for previously kept forwards
			// 2.1. Read forwards, if have to
			if (!m_fForwardsRead) {
				LoadForwards();
			}
			int	nForwardersLinked = 0;
			for (CForwardInfoVector::iterator iter = m_vectForwardInfo.begin(); iter != m_vectForwardInfo.end();) {
				// 2.4.5. Compare area with just created
				if (stricmp( iter->m_szArea, Tic.Area ) == 0) {
					// 2.5. Check address
					SFTNAddress	addr;
					if (ParseFTNAddress( iter->m_szLink, addr )) {
						// 2.5.1. Link area
						if (Links.Select( addr )) {
							if (Areas.Link( Links.Selected )) {
								// 2.5.2. Count links
								nForwardersLinked++;
								if (nForwardersLinked) {
									NewAreas << "\nAutomatically linked:\n";
								}
								// 2.5.3. Prepare address string
								char	szLine[LINE_SIZE];
								sprintf( szLine, "%d:%d/%d", Links.Selected->Address.Zone, Links.Selected->Address.Net, Links.Selected->Address.Node );
								if (Links.Selected->Address.Point != 0) {
									sprintf( szLine + strlen(szLine), ".%d", Links.Selected->Address.Point );
								}
								strcat( szLine, "@" );
								strcat( szLine, Links.Selected->Address.Domain );
								// 2.5.4. Add newareas.txt line
								NewAreas << "  " << szLine << endl;
								// 2.5.5. Add log line
								GlobalVars.Log << "  linked " << szLine << endl;
							}
							else {
								// FIX: we should send message about failure
							}
							// remove info from forwards
							iter = m_vectForwardInfo.erase(iter);
							m_fForwardsChanged = true;
							continue;
						}
						else {
							GlobalVars.Log << "Unknown link " 
								<< addr.Zone << ":" << addr.Net 
								<< "/" << addr.Node << "." << addr.Point 
								<< "@" << addr.Domain 
								<< " from " << szForwardsFileName << endl;
						}
					}
					else {
						GlobalVars.Log << "Failed to parse address " << iter->m_szLink << " from " << szForwardsFileName << endl;
					}
				}
				++iter;
			}
			if (nForwardersLinked > 0) {
				NewAreas << endl << endl;
			}
			// Restore link
			Links.Select( Tic.From );
                        NewAreas.close();
                    }
                    else {
                        sprintf( LogLine, "! %12s: unknown area %s\n",
                                 TicsList->String + TicNameFrom,
                                 Tic.Area );
                        GlobalVars.Log << LogLine;
                        cout << LogLine;
                        MoveTicToBads( TicsList->String, TicNameFrom );
                        break;
                    }
                }

                // handle replaces if needed
                ReplaceDone = 0;
                if (GlobalVars.HandleReplaces && *(Tic.Replaces) != '\0') {
                    strcpy( DestFileName, Areas.Selected->Path );
                    strcat( DestFileName, Tic.Replaces );
                    // check for file existance
                    if (access( DestFileName, 0 ) == 0) {
                        // remove file
                        if (unlink(DestFileName) != 0) {
                            sprintf( ReplaceLine,
                                     "? %12s: can't delete replaced file %s\n",
                                     TicsList->String + TicNameFrom,
                                     DestFileName );
                        }
                        else {
                            Areas.ClearDescFile( Tic.Replaces );
                            sprintf( ReplaceLine,
                                     "                replaced file %s\n",
                                     Tic.Replaces );
                        }
                        ReplaceDone = 1;
                    }
                }
                // move file into fecho directory
                strcpy( DestFileName, Areas.Selected->Path );
                if (*Tic.FullName != '\0')
                    strcat( DestFileName, Tic.FullName );
                else
                    strcat( DestFileName, Tic.File );
                if (!MoveFile( FullFileName, DestFileName)) {
                    sprintf( LogLine,
                             "! %12s: failed to move file %s into file %s\n",
                             TicsList->String + TicNameFrom, Tic.File,
                             DestFileName );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                    break;
                }
                Areas.Selected->Traffic +=
                    (FindFileData.nFileSizeHigh * MAXDWORD) +
                    FindFileData.nFileSizeLow;
                NewFilesArrived = 1;
                // output information in logfile and on screen, if needed
                sprintf( LogLine, "+ %12s: %-12s in %s\n",
                         TicsList->String + TicNameFrom, Tic.File,
                         Areas.Selected->Name );
                GlobalVars.Log << LogLine;
                if (GlobalVars.ShowMustGoOn)
                    cout << LogLine;
                if (ReplaceDone) {
                    GlobalVars.Log << ReplaceLine;
                    if (GlobalVars.ShowMustGoOn)
                        cout << ReplaceLine;
                }
                // update descriptions
                if (*Tic.FullName != '\0') {
                    GlobalVars.Log << "                renamed to \"" <<
                        Tic.FullName << "\"\n";
                    Areas.UpdateDescFile( Tic.FullName, Tic.Desc,
                                          Tic.LDescList );
                }
                else
                    Areas.UpdateDescFile( Tic.File, Tic.Desc, Tic.LDescList );
                if (unlink( TicsList->String ) != 0) {
                    sprintf( LogLine, "? %12s: can't delete TIC\n",
                             TicsList->String + TicNameFrom );
                    GlobalVars.Log << LogLine;
                    cout << LogLine;
                }
                // update newfiles list
                out << "Area " << Areas.Selected->Name << endl;
                out << "File ";
                if (*Tic.FullName != '\0')
                    out << Tic.FullName;
                else
                    out << Tic.File;
                out << endl;
                out << "Size "
                    << (FindFileData.nFileSizeHigh * MAXDWORD) +
                    FindFileData.nFileSizeLow << endl;
                out << "Desc " << Tic.Desc << endl;
                out << "Origin " << Tic.Origin.Zone << ":" << Tic.Origin.Net
                    << "/" << Tic.Origin.Node << "." << Tic.Origin.Point << endl;
                out << "Uplink "  << Linked->LinkPtr->Address.Zone << ":" << Linked->LinkPtr->Address.Net
                    << "/" << Linked->LinkPtr->Address.Node << "." << Linked->LinkPtr->Address.Point << endl;
                // add sender to seenby list, if absent
                CurSeenby = Tic.SeenbyList;
                while (CurSeenby != NULL) {
                    if (AddrCmp( Linked->LinkPtr->Address,
                                 CurSeenby->Addr ) == 0)
                        break;
                    CurSeenby = CurSeenby->Next;
                }
                if (CurSeenby == NULL) {
                    if (Tic.SeenbyList == NULL) {
                        Tic.SeenbyList = new SAddrList;
                        Tic.LastSeenby = Tic.SeenbyList;
                    }
                    else {
                        Tic.LastSeenby->Next = new SAddrList;
                        Tic.LastSeenby       = Tic.LastSeenby->Next;
                    }
                    Tic.LastSeenby->Next = NULL;
                    Tic.LastSeenby->Addr.Zone  = Linked->LinkPtr->Address.Zone;
                    Tic.LastSeenby->Addr.Net   = Linked->LinkPtr->Address.Net;
                    Tic.LastSeenby->Addr.Node  = Linked->LinkPtr->Address.Node;
                    Tic.LastSeenby->Addr.Point = Linked->LinkPtr->Address.Point;
                    strcpy( Tic.LastSeenby->Addr.Domain,
                            Linked->LinkPtr->Address.Domain );
                }
                // add us to seenby list, if absent
                CurSeenby = Tic.SeenbyList;
                while (CurSeenby != NULL) {
                    if (AddrCmp( *(Linked->LinkPtr->OurAddr),
                                 CurSeenby->Addr ) == 0)
                        break;
                    CurSeenby = CurSeenby->Next;
                }
                if (CurSeenby == NULL) {
                    if (Tic.SeenbyList == NULL) {
                        Tic.SeenbyList = new SAddrList;
                        Tic.LastSeenby = Tic.SeenbyList;
                    }
                    else {
                        Tic.LastSeenby->Next = new SAddrList;
                        Tic.LastSeenby       = Tic.LastSeenby->Next;
                    }
                    Tic.LastSeenby->Next = NULL;
                    Tic.LastSeenby->Addr.Zone  = Linked->LinkPtr->OurAddr->Zone;
                    Tic.LastSeenby->Addr.Net   = Linked->LinkPtr->OurAddr->Net;
                    Tic.LastSeenby->Addr.Node  = Linked->LinkPtr->OurAddr->Node;
                    Tic.LastSeenby->Addr.Point = Linked->LinkPtr->OurAddr->Point;
                    strcpy( Tic.LastSeenby->Addr.Domain,
                            Linked->LinkPtr->OurAddr->Domain );
                }
                // add absent AKAs if AllAKASeen switched to Yes
                if (GlobalVars.AllAKASeen == 1) {
                    for (i = 0; i <= GlobalVars.AKAAddrNumber; i++)
                        AKASeen[i] = 0;
                    CurSeenby = Tic.SeenbyList;
                    while (CurSeenby != NULL) {
                        // check main addr
                        if (AddrCmp( GlobalVars.MainAddr,
                                     CurSeenby->Addr ) == 0)
                            AKASeen[GlobalVars.AKAAddrNumber] = 1;
                        // check AKAs
                        for (i = 0; i < GlobalVars.AKAAddrNumber; i++) {
                            if (AddrCmp( GlobalVars.AKAAddr[i],
                                         CurSeenby->Addr ) == 0)
                            {
                                AKASeen[i] = 1;
                                break;
                            }
                        }
                        CurSeenby = CurSeenby->Next;
                    }
                    // add absent addresses
                    for (i = 0; i < GlobalVars.AKAAddrNumber; i++) {
                        if (AKASeen[i] == 0) {
                            Tic.LastSeenby->Next = new SAddrList;
                            Tic.LastSeenby       = Tic.LastSeenby->Next;
                            Tic.LastSeenby->Next = NULL;
                            Tic.LastSeenby->Addr.Zone  = GlobalVars.AKAAddr[i].Zone;
                            Tic.LastSeenby->Addr.Net   = GlobalVars.AKAAddr[i].Net;
                            Tic.LastSeenby->Addr.Node  = GlobalVars.AKAAddr[i].Node;
                            Tic.LastSeenby->Addr.Point = GlobalVars.AKAAddr[i].Point;
                            strcpy( Tic.LastSeenby->Addr.Domain,
                                    GlobalVars.AKAAddr[i].Domain );
                        }
                    }
                    if (AKASeen[GlobalVars.AKAAddrNumber] == 0) {
                        Tic.LastSeenby->Next = new SAddrList;
                        Tic.LastSeenby       = Tic.LastSeenby->Next;
                        Tic.LastSeenby->Next = NULL;
                        Tic.LastSeenby->Addr.Zone  = GlobalVars.MainAddr.Zone;
                        Tic.LastSeenby->Addr.Net   = GlobalVars.MainAddr.Net;
                        Tic.LastSeenby->Addr.Node  = GlobalVars.MainAddr.Node;
                        Tic.LastSeenby->Addr.Point = GlobalVars.MainAddr.Point;
                        strcpy( Tic.LastSeenby->Addr.Domain,
                                GlobalVars.MainAddr.Domain );
                    }
                }
                // do tics for all connected links absent in seenbys
                // first of all prepare list of links
                ForwardList = NULL;
                for (CurLinked = Areas.Selected->LinkedList;
                     CurLinked != NULL;
                     CurLinked = CurLinked->Next)
                {
                    // skip passive or unavailable seenby
                    if (CurLinked->LinkPtr->Activity != ACTIVITY_USUAL)
                        continue;
                    // skip sender
                    if (AddrCmp( CurLinked->LinkPtr->Address,
                                 Linked->LinkPtr->Address) == 0)
                        continue;
                    // skip links without receive status
                    if ((CurLinked->RouteType & ROUTE_RECEIVE) == 0)
                        continue;
                    // check seenby
                    CurSeenby = Tic.SeenbyList;
                    while (CurSeenby != NULL) {
                        if (AddrCmp( CurLinked->LinkPtr->Address,
                                     CurSeenby->Addr ) == 0)
                            break;
                        CurSeenby = CurSeenby->Next;
                    }
                    if (CurSeenby != NULL)
                        continue;
                    // add link to list
                    if (ForwardList == NULL) {
                        ForwardList = new SLinked;
                        LastForward = ForwardList;
                    }
                    else {
                        LastForward->Next = new SLinked;
                        LastForward       = LastForward->Next;
                    }
                    LastForward->Next = NULL;
                    LastForward->LinkPtr = CurLinked->LinkPtr;
                    // doesn't matter what is the value of RouteType
                }
                // now do tics
                if (ForwardList != NULL)
                    TossedForLinks = 1;
                LastForward = ForwardList;
                while (LastForward != NULL) {
                    Tic.Forward( LastForward->LinkPtr, Areas.Selected->Path,
                                 ForwardList );
                    LastForward = LastForward->Next;
                }
                // free forwards list
                while (ForwardList != NULL) {
                    LastForward = ForwardList->Next;
                    delete ForwardList;
                    ForwardList = LastForward;
                }
                break;
                // using pragmas to avoid warning message "unreachable code"
//#pragma warning 13 5
            } while (TRUE);
//#pragma warning 13 4

        }
        else {
            sprintf( LogLine, "! %12s: error parsering file\n",
                     TicsList->String + TicNameFrom );
            GlobalVars.Log << LogLine;
	    GlobalVars.Log << "                " << Tic.GetLastError() << endl;
            cout << LogLine;
	    cout << "                " << Tic.GetLastError() << endl;
        }
        TmpTic = TicsList->Next;
        delete [] TicsList->String;
        delete TicsList;
        TicsList = TmpTic;
        cout.flush();
    }
    out.close();
    // check if newfiles.txt has no data (zero size) and remove it if it hasn't
    strcpy( FullFileName, StartupPath );
    strcat( FullFileName, "newfiles.txt" );
    hFindFile = FindFirstFile( FullFileName, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    if (FileFound) {
        FindClose( hFindFile );
        if (((FindFileData.nFileSizeHigh * MAXDWORD) +
             FindFileData.nFileSizeLow) == 0)
            unlink( FullFileName);
    }
    // save traffic, if changed
    if (NewFilesArrived) {
        strcpy( FullFileName, StartupPath );
        strcat( FullFileName, "traffic.dat" );
        out.open( FullFileName );
        if (!out) {
            sprintf( LogLine, "! Error writing traffic base file %s\n",
                     FullFileName );
            GlobalVars.Log << LogLine;
            cout << LogLine;
        }
        else {
            if (*TimeLine == '\0') {
                time_t CurTime = time(NULL);
                out << ctime( &CurTime );
            }
            else
                out << TimeLine << endl;
            Areas.SelectFirst();
            while (Areas.Selected != NULL) {
                if (Areas.Selected->Traffic > 0)
                    out << Areas.Selected->Name << " "
                        << Areas.Selected->Traffic << endl;
                Areas.SelectNext();
            }
            out.close();
        }
        // check if traffic.dat has no data (zero size) and remove it if it hasn't
        hFindFile = FindFirstFile( FullFileName, &FindFileData );
        FileFound = (hFindFile != INVALID_HANDLE_VALUE);
        if (FileFound) {
            FindClose( hFindFile );
            if (((FindFileData.nFileSizeHigh * MAXDWORD) +
                 FindFileData.nFileSizeLow) == 0)
                unlink( FullFileName);
        }
    }
    delete [] AKASeen;
    // create flag, if tossed
    if (TossedForLinks && *GlobalVars.TossFlagName != '\0')
        CreateFileFlag( GlobalVars.TossFlagName );
}

int CAdrenalin::Hatch() {
    char        DestFileName[MAX_PATH];
    char	DestFilePath[MAX_PATH];
    char        ShortFileName[MAX_PATH];
    char        FullFileName[MAX_PATH];
    char        LogLine[LINE_SIZE];
    char        TimeLine[LINE_SIZE];
    int         i , j, NameFrom;
    CTic        Tic;
    ofstream    out;
    SLinked    *CurLinked;
    SLinked    *ForwardList, *LastForward;
    ifstream    in;
    CCfgParser *TrafficParser;
    char       *s;
    int         Traffic;
    int         DistributedForLinks;

    GlobalVars.Log << "Hatch:\n";
    
    // make all fecho directories and set areas traffic to zero
    Areas.SelectFirst();
    while (Areas.Selected != NULL) {
        MakeDirStructure( Areas.Selected->Path );
        Areas.Selected->Traffic = 0;
        Areas.SelectNext();
    }

    // make tics directories
    MakeDirStructure( GlobalVars.BadTicsPath );
    MakeDirStructure( GlobalVars.NewTicsPath );

    // check area
    if (!Areas.Select( HatchArea )) {
        GlobalVars.Log << "  area " << HatchArea << " doesn't exist\n";
        cout << "Area " << HatchArea << " doesn't exist\n";
        return (ERROR_AREA_NOT_FOUND);
    }
    // check file
    if (access( HatchFile, 0) != 0) {
        GlobalVars.Log << "  file " << HatchFile << " not found\n";
        cout << "File " << HatchFile << " not found\n";
        return (ERROR_FILE_ABSENT);
    }
    // calc crc & size
    Tic.Size = CalcCRC32( HatchFile, Tic.CRC32 );
    if (Tic.Size == -1) {
        GlobalVars.Log << "  failed to calculate CRC32 of " << HatchFile
                       << endl;
        cout << "Failed to calculate CRC32 of " << HatchFile
             << endl;
        return (ERROR_CRC32_FAILED);
    }
    // move file into fecho directory if have to
    // first let's find name without path
    for (i = 0, NameFrom = -1; HatchFile[i] != '\0'; i++) {
        if (HatchFile[i] == '\\')
            NameFrom = i;
    }
    NameFrom++;
        if (leaveHatchedFile) {
        	if (HatchFile[1] == ':')
        		strcpy( DestFileName, HatchFile );
        	else {
        		strcpy( DestFileName, StartupPath );
        		strcat( DestFileName, HatchFile );
        	}
                for (i = 0, j = -1; DestFileName[i] != '\0'; i++) {
                    if (DestFileName[i] == '\\')
                        j = i;
                }
                j++;
                strcpy( DestFilePath, DestFileName );
                DestFilePath[j] = '\0';
        }
        else {
                strcpy( DestFileName, Areas.Selected->Path );
                strcat( DestFileName, HatchFile + NameFrom );
                if (!CopyFile( HatchFile, DestFileName )) {
                    GlobalVars.Log << "  failed to move file " << HatchFile << " into "
                                   << DestFileName << endl;
                    cout << "Failed to move file " << HatchFile << " into "
                         << DestFileName << endl;
                    return (ERROR_MOVE_FAILED);
                }
                strcpy( DestFilePath, Areas.Selected->Path );
        }
    GlobalVars.Log << "  file: " << HatchFile << "\n  area: "
                   << Areas.Selected->Name << endl;
    if (*HatchDesc != '\0')
        GlobalVars.Log << "  desc: \"" << HatchDesc << "\"\n";
    else
        GlobalVars.Log << "  no description\n";
    if (*HatchReplaces != '\0')
        GlobalVars.Log << "  repl: " << HatchReplaces << endl;
    else
        GlobalVars.Log << "  no Replaces\n";

        if (!leaveHatchedFile) {
                // update descriptions
                GlobalVars.Log << "  updating description file...";
                Areas.UpdateDescFile( HatchFile + NameFrom, HatchDesc );
                GlobalVars.Log << "done\n";
        }

        // prepare list of links

        ForwardList = NULL;
        for (CurLinked = Areas.Selected->LinkedList; CurLinked != NULL; CurLinked = CurLinked->Next) {
                // skip passive or unavailable seenby
                if (CurLinked->LinkPtr->Activity != ACTIVITY_USUAL)
                	continue;
                // skip links without receive status
                if ((CurLinked->RouteType & ROUTE_RECEIVE) == 0)
                	continue;
                // add link to list
                if (ForwardList == NULL) {
                        ForwardList = new SLinked;
                        LastForward = ForwardList;
                }
                else {
                        LastForward->Next = new SLinked;
                        LastForward       = LastForward->Next;
                }
                LastForward->Next = NULL;
                LastForward->LinkPtr = CurLinked->LinkPtr;
                // doesn't matter what is the value of RouteType
        }

        SFTNAddress*	addr;

        // if only one link is connected to area, set hatch address to its AKA
        if (ForwardList != NULL && ForwardList->Next == NULL)
        	addr = ForwardList->LinkPtr->OurAddr;
        else
        	addr = &GlobalVars.MainAddr;

        // open newfiles list
        GlobalVars.Log << "  updating newfiles.txt file...";
        strcpy( ShortFileName, StartupPath );
        strcat( ShortFileName, "newfiles.txt" );
        out.open( ShortFileName, ios::app );
        if (!out)
        	out.open( ShortFileName );
        // update newfiles list
        out << "Area " << Areas.Selected->Name << endl;
        out << "File " << HatchFile + NameFrom << endl;
        out << "Size " << Tic.Size << endl;
        out << "Desc " << HatchDesc << endl;
        out << "Origin " << addr->Zone << ":" << addr->Net << "/" << addr->Node << "." << addr->Point << endl;
        
        out << "Uplink " << addr->Zone << ":" << addr->Net << "/" << addr->Node << "." << addr->Point << endl;
        out.close();
        GlobalVars.Log << "done\n";
    
    // update traffic base
    
    // read areas traffic
    
    strcpy( FullFileName, StartupPath );
    strcat( FullFileName, "traffic.dat" );

    *TimeLine = '\0';
    TrafficParser = new CCfgParser( FullFileName );
    if (!TrafficParser->Failed()) {
        // read date line
        TrafficParser->ReadLine();
        if (!TrafficParser->Failed())
            strcpy( TimeLine, TrafficParser->String() );
        // read data lines
        while (!TrafficParser->Eof()) {
            TrafficParser->ReadLine();
            if (TrafficParser->Failed())
                break;
            if (TrafficParser->Count() != 2)
                break;
            if (!Areas.Select( (*TrafficParser)[0] ))
                continue;
            s = (*TrafficParser)[1];
            Traffic = 0;
            while (isdigit(*s)) {
                Traffic *= 10;
                Traffic += (*s - '0');
                s++;
            }
            if (*s != '\0')
                break;
            Areas.Selected->Traffic = Traffic;
        }
    }
    delete TrafficParser;

    Areas.Select( HatchArea );
    assert( Areas.Selected != NULL );
    Areas.Selected->Traffic += Tic.Size;
    
    out.open( FullFileName );
    if (!out) {
        sprintf( LogLine, "! Error writing traffic base file %s\n",
                 FullFileName );
        GlobalVars.Log << LogLine;
        cout << LogLine;
    }
    else {
        if (*TimeLine == '\0') {
            time_t CurTime = time(NULL);
            out << ctime( &CurTime );
        }
        else
            out << TimeLine << endl;
        Areas.SelectFirst();
        while (Areas.Selected != NULL) {
            if (Areas.Selected->Traffic > 0)
                out << Areas.Selected->Name << " "
                    << Areas.Selected->Traffic << endl;
            Areas.SelectNext();
        }
        out.close();
    }

    Areas.Select( HatchArea );

    
    GlobalVars.Log << "  creating TICs for connected links...";
    // set area name
    strcpy( Tic.Area, Areas.Selected->Name );
    // check and set file name
    ConvertNameToShort( DestFileName, ShortFileName );
    for (i = 0, j = -1; ShortFileName[i] != '\0'; i++) {
        if (ShortFileName[i] == '\\')
            j = i;
    }
    j++;
    strcpy( Tic.File, ShortFileName + j );
    if (stricmp( HatchFile + NameFrom, ShortFileName + j ) != 0)
        strcpy( Tic.FullName, HatchFile + NameFrom );
    // set desc
    if (*HatchDesc != '\0')
        strcpy( Tic.Desc, HatchDesc );
    // set Replaces
    if (*HatchReplaces != '\0')
        strcpy( Tic.Replaces, HatchReplaces );
    // set Origin to main address, if have to
    if (GlobalVars.HatchOriginMain || GlobalVars.AKAAddrNumber == 0) {
        Tic.Origin.Zone  = GlobalVars.MainAddr.Zone;
        Tic.Origin.Net   = GlobalVars.MainAddr.Net;
        Tic.Origin.Node  = GlobalVars.MainAddr.Node;
        Tic.Origin.Point = GlobalVars.MainAddr.Point;
        strcpy( Tic.Origin.Domain, GlobalVars.MainAddr.Domain );
    }
    // add all us to seenby list, if have to
    if (GlobalVars.AllAKASeen == 1) {
        Tic.SeenbyList = new SAddrList;
        Tic.LastSeenby = Tic.SeenbyList;
        Tic.LastSeenby->Next = NULL;
        Tic.LastSeenby->Addr.Zone  = GlobalVars.MainAddr.Zone;
        Tic.LastSeenby->Addr.Net   = GlobalVars.MainAddr.Net;
        Tic.LastSeenby->Addr.Node  = GlobalVars.MainAddr.Node;
        Tic.LastSeenby->Addr.Point = GlobalVars.MainAddr.Point;
        strcpy( Tic.LastSeenby->Addr.Domain, GlobalVars.MainAddr.Domain );
        for (i = 0; i < GlobalVars.AKAAddrNumber; i++) {
            Tic.LastSeenby->Next = new SAddrList;
            Tic.LastSeenby = Tic.LastSeenby->Next;
            Tic.LastSeenby->Next = NULL;
            Tic.LastSeenby->Addr.Zone  = GlobalVars.AKAAddr[i].Zone;
            Tic.LastSeenby->Addr.Net   = GlobalVars.AKAAddr[i].Net;
            Tic.LastSeenby->Addr.Node  = GlobalVars.AKAAddr[i].Node;
            Tic.LastSeenby->Addr.Point = GlobalVars.AKAAddr[i].Point;
            strcpy( Tic.LastSeenby->Addr.Domain, GlobalVars.AKAAddr[i].Domain );
        }
    }
    // do tics for all connected links
    // now do tics
    DistributedForLinks = (ForwardList != NULL);
    LastForward = ForwardList;
    while (LastForward != NULL) {
        // set Origin to AKA, if have to
        if (GlobalVars.AKAAddrNumber > 0 && !GlobalVars.HatchOriginMain) {
            Tic.Origin.Zone  = LastForward->LinkPtr->OurAddr->Zone;
            Tic.Origin.Net   = LastForward->LinkPtr->OurAddr->Net;
            Tic.Origin.Node  = LastForward->LinkPtr->OurAddr->Node;
            Tic.Origin.Point = LastForward->LinkPtr->OurAddr->Point;
            strcpy( Tic.Origin.Domain, LastForward->LinkPtr->OurAddr->Domain );
        }
       	Tic.Forward( LastForward->LinkPtr, DestFilePath, ForwardList );
        LastForward = LastForward->Next;
    }
    // free forwards list
    while (ForwardList != NULL) {
        LastForward = ForwardList->Next;
        delete ForwardList;
        ForwardList = LastForward;
    }
    // create flag
    if (DistributedForLinks && *GlobalVars.HatchFlagName != '\0')
        CreateFileFlag( GlobalVars.HatchFlagName );
    GlobalVars.Log << "done\n";
    cout << "File " << HatchFile << " was successfully hatched into area "
        << Areas.Selected->Name << endl;
    return (ERROR_NONE);
}

void CAdrenalin::ChooseAKA(SLink *Link) {
    int i, Choosed;
    int MaxSameness, Sameness;
    // sameness coefficient is:
    // if zones are same == 1
    // if nets too       == 2
    // if nodes too      == 3
    // if points too     == 4

    Choosed = -1; // MainAddr
    MaxSameness = 0;
    if (stricmp( GlobalVars.MainAddr.Domain, Link->Address.Domain ) == 0) {
        if (GlobalVars.MainAddr.Zone == Link->Address.Zone) {
            MaxSameness++;
            if (GlobalVars.MainAddr.Net == Link->Address.Net) {
                MaxSameness++;
                if (GlobalVars.MainAddr.Node == Link->Address.Node) {
                   MaxSameness++;
                   if (GlobalVars.MainAddr.Point == Link->Address.Point)
                       MaxSameness++;
                }
            }
        }
    }
    for (i = 0; i < GlobalVars.AKAAddrNumber; i++) {
        if (stricmp( GlobalVars.AKAAddr[i].Domain, Link->Address.Domain ) == 0)
        {
            Sameness = 0;
            if (GlobalVars.AKAAddr[i].Zone == Link->Address.Zone) {
                Sameness++;
                if (GlobalVars.AKAAddr[i].Net == Link->Address.Net) {
                    Sameness++;
                    if (GlobalVars.AKAAddr[i].Node == Link->Address.Node) {
                       Sameness++;
                       if (GlobalVars.AKAAddr[i].Point == Link->Address.Point)
                           Sameness++;
                    }
                }
            }
            if (Sameness > MaxSameness) {
                Choosed = i;
                MaxSameness = Sameness;
            }
        }
    }
    // fill OurAddr field
    if (Choosed == -1)
        Link->OurAddr = &(GlobalVars.MainAddr);
    else
        Link->OurAddr = GlobalVars.AKAAddr + Choosed;
}

void CAdrenalin::AddStartupPath(char *Location) {
    char TempPath[MAX_PATH];

    if (Location[0] == '\0' || Location[1] != ':') {
        strcpy( TempPath, Location );
        strcpy( Location, StartupPath );
        strcat( Location, TempPath );
    }
}

int CAdrenalin::SaveCfg() {
	struct LinksList {
		SLink*		link;
		LinksList*	prev;
		LinksList*	next;
	};

	struct LinksFileList {
		char*		cfgFileName;
		LinksList*	list;
		LinksFileList*	next;
	};

	LinksFileList*	fileList;
	LinksFileList*	item;
	LinksFileList*	lastItem;

	LinksList*	linkItem;

	SLink*		link;

	ofstream      out;
	ifstream      in;
	char          tempFileName[MAX_PATH];
	char          buf[LINE_SIZE];

	int           Line;
	int           CompareNext;
	char          c, firstc;
	unsigned long bit;

	// prepare temp file name

    	strcpy( tempFileName, StartupPath );
    	strcat( tempFileName, "adrenalin.tmp" );
    	if (!GetTmpFileName(tempFileName))
        	return (0);

        // prepare links file list

        fileList = NULL;
        lastItem = NULL;
        Links.SelectFirst();
        while (Links.Selected != NULL) {
        	// search link cfg file name in file list
        	item = fileList;
        	while (item != NULL && strcmp( item->cfgFileName, Links.Selected->cfgFileName ) != 0)
        		item = item->next;
        	// if corresponding item wasn't found we shell add new one
        	if (item == NULL) {
        		if (fileList == NULL) {
        			fileList = new LinksFileList;
        			lastItem = fileList;
        		}
        		else {
        			lastItem->next = new LinksFileList;
        			lastItem       = lastItem->next;
        		}
        		lastItem->cfgFileName = Links.Selected->cfgFileName;
        		lastItem->list = new LinksList;
        		lastItem->list->link = Links.Selected;
        		lastItem->list->prev = NULL;
        		lastItem->list->next = NULL;
        		lastItem->next = NULL;
        	}
        	else {
        		// lets insert link into correct position based on line number
        		linkItem = item->list;
        		while (linkItem->link->Line < Links.Selected->Line && linkItem->next != NULL)
        			linkItem = linkItem->next;
        		// if linkItem->link->Line less than current link line than
        		// it means that this item is the last item in list
        		if (linkItem->link->Line < Links.Selected->Line) {
        			linkItem->next = new LinksList;
        			linkItem->next->prev = linkItem;
        			linkItem = linkItem->next;
        			linkItem->link = Links.Selected;
        			linkItem->next = NULL;
        		}
        		else {
        			if (linkItem->prev == NULL) {
        				lastItem->list = new LinksList;
        				lastItem->list->prev = NULL;
        				lastItem->list->link = Links.Selected;
        				lastItem->list->next = linkItem;
        				linkItem->prev = lastItem->list;
        			}
        			else {
        				linkItem->prev->next = new LinksList;
        				linkItem->prev->next->prev = linkItem->prev;
        				linkItem->prev->next->next = linkItem;
        				linkItem->prev->next->link = Links.Selected;
        				linkItem->prev = linkItem->prev->next;
        			}
        		}
        	}
        	Links.SelectNext();
        }

        // go through file list and rewrite all files

        item = fileList;
        while (item != NULL) {
        	do {
        		// open input and output files
                	in.open( item->cfgFileName );
                	if (!in)
                		break;
                	out.open( tempFileName );
                	if (!out) {
                		in.close();
                		break;
                	}
                	// setup links
                	linkItem = item->list;
                        link = linkItem->link;
                	Line = 0;
                	// read, write and rewritelines
			while (!in.eof() && linkItem != NULL) {
                                in.getline( buf, LINE_SIZE );
                                if (!in)
                                	continue;
                                Line++;
                                if (Line < link->Line)
                                    	out << buf << endl;
                                else {
                                        out << "Link " << setw(24) << setiosflags(ios::left);
                                        // address
                                        *buf = '\0';
                                        if (stricmp( GlobalVars.MainAddr.Domain, link->Address.Domain ) == 0) {
                                         	CompareNext = 1;
	                                        if (GlobalVars.MainAddr.Zone != link->Address.Zone) {
        	                                	sprintf( buf, "%lu:", link->Address.Zone );
                	                        	CompareNext = 0;
                        	                }
                                                if (!CompareNext || GlobalVars.MainAddr.Net != link->Address.Net)
                                                	sprintf( buf + strlen(buf), "%lu/", link->Address.Net );
                                                sprintf( buf + strlen(buf), "%lu", link->Address.Node );
                                                if (link->Address.Point != 0)
                                                	sprintf( buf + strlen(buf), ".%lu", link->Address.Point );
                                        }
                                	else
                                    		sprintf( buf, "%lu:%lu/%lu.%lu@%s", link->Address.Zone, link->Address.Net, link->Address.Node, link->Address.Point, link->Address.Domain );
                                	out << buf;
                                	// password
                                	out << " " << setw(15) << setiosflags(ios::left) << link->Password << " ";
                                	// route type
                                	switch (link->RouteType) {
                                                case ROUTE_SEND:
                                                        out << "S";
                                                        break;
                                                case ROUTE_RECEIVE:
                                                        out << "R";
                                                        break;
                                                case ROUTE_BIDIRECT:
                                                        out << "B";
                                                        break;
                                        }
                                        // BSO type
                                        out << " ";
                                        switch (link->BSOType) {
                                        case BSO_NORMAL:
                                                out << "N";
                                                break;
                                        case BSO_DIRECT:
                                                out << "D";
                                                break;
                                        case BSO_HOLD:
                                                out << "H";
                                                break;
                                        case BSO_CRASH:
                                                out << "C";
                                                break;
                                        case BSO_IMMEDIATE:
                                                out << "I";
                                                break;
                                        }
                                        out << " ";
                                        // allow area creation
                                        if (link->AllowAreaCreation)
                                                out << "A ";
                                        // passive
                                        switch (link->Activity) {
                                        case ACTIVITY_PASSIVE:
                                            out << "P ";
                                            break;
                                        case ACTIVITY_UNAVAILABLE:
                                            out << "U ";
                                            break;
                                        }
                                        // priority
                                        out << link->Priority << "/";
                                        // groups
                                        firstc = '\0';
                                        c      = 'A';
                                        bit    = 1;
                                        while (c <= 'Z') {
                                                if ((link->Groups & bit) != 0) {
                                                        if (firstc == '\0')
                                                                firstc = c;
                                                }
                                                else {
                                                        if (firstc != '\0') {
                                                                out << firstc;
                                                                if (firstc != c-1)
                                                                        out << "-" << (char)(c-1);
                                                                firstc = '\0';
                                                        }
                                                }
                                                c++;
                                                bit <<= 1;
                                        }
                                        if (firstc != '\0') {
                                                if (firstc == 'A')
                                                        out << "*";
                                                else {
                                                        out << firstc;
                                                        if (firstc != c-1)
                                                                out << "*";
                                                }
                                        }
                                        out << endl;
                                        // free link memory
                                        if (linkItem->next == NULL) {
                                        	delete linkItem;
                                        	linkItem = NULL;
                                        }
                                        else {
                                                linkItem = linkItem->next;
                                                delete linkItem->prev;
                                                linkItem->prev = NULL;
                                               	link = linkItem->link;
                                        }
                        	}
			}
                        while (!in.eof()) {
                                in.getline( buf, LINE_SIZE );
                                if (!in)
                                        continue;
                                out << buf << endl;
                        }
                        in.close();
                        out.close();
                        unlink( item->cfgFileName);
                        rename( tempFileName, item->cfgFileName );
                } while (FALSE);
        	
        	// free item memory and shift to next item
        	lastItem = item->next;
        	delete item;
        	item = lastItem;
        }

        return (1);
}


void CAdrenalin::Purge() {
    struct SAreaFiles {
        char         AreaName[AREA_NAME_SIZE];
        SStringList *Files, *LastFile;
        SAreaFiles  *Next;
    };
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFindFile;
    BOOL            FileFound;
    char            FindFileMask[MAX_PATH];
    char            FullFileName[MAX_PATH];
    char            FileName[MAX_PATH];
    int             NamePos;
    CTic            Tic;
    SAreaFiles     *AreaFiles, *LastAreaFiles, *CurAreaFiles, *PrevAreaFiles;
    SStringList    *CurFile, *LastFile;
    ifstream        in;
    ofstream        tempout;
    char            TempFileName[MAX_PATH];
    char            buf[LINE_SIZE];
    int             i;
    int             PurgedNumber, TotalPurgedNumber;
    int             PurgedSize, TotalPurgedSize;
    int             SkipLongDesc;
    SStringList    *CurPurgeArea;
    int             HaveToAddFile;
    int             AreaAsMask;

    GlobalVars.Log << "Purge:\n";
    
    AreaFiles = LastAreaFiles = NULL;
    if (PurgeAreasList != NULL) {
        for (CurPurgeArea = PurgeAreasList; CurPurgeArea != NULL;
             CurPurgeArea = CurPurgeArea->Next)
        {
            if (CurPurgeArea->String[0] == '-') { // exclude from arealist
                CurAreaFiles = AreaFiles;
                PrevAreaFiles = NULL;
                while (CurAreaFiles != NULL) {
                    if (CheckMask( CurAreaFiles->AreaName,
                                   CurPurgeArea->String + 1 ))
                    {
                        // remove area from list
                        if (PrevAreaFiles == NULL) {
                            AreaFiles = CurAreaFiles->Next;
                            delete CurAreaFiles;
                            CurAreaFiles = AreaFiles;
                            if (AreaFiles == NULL)
                                LastAreaFiles = NULL;
                        }
                        else {
                            PrevAreaFiles->Next = CurAreaFiles->Next;
                            delete CurAreaFiles;
                            CurAreaFiles = PrevAreaFiles->Next;
                            if (CurAreaFiles == NULL)
                                LastAreaFiles = PrevAreaFiles;
                        }
                    }
                    else {
                        if (PrevAreaFiles == NULL)
                            PrevAreaFiles = AreaFiles;
                        else
                            PrevAreaFiles = PrevAreaFiles->Next;
                        CurAreaFiles = CurAreaFiles->Next;
                    }
                }
            }
            else { // add to arealist, if one or more areas meet mask
                i = 0; // number of added areas
                AreaAsMask = IsMask( CurPurgeArea->String );
                for (Areas.SelectFirst(); Areas.Selected != NULL;
                     Areas.SelectNext())
                {
                    if (AreaAsMask) {
                        if (Areas.Selected->VetoManualPurge == 1)
                            continue;
                        if (!CheckMask( Areas.Selected->Name,
                                        CurPurgeArea->String))
                            continue;
                    }
                    else if (stricmp( Areas.Selected->Name,
                                      CurPurgeArea->String) != 0)
                        continue;
                    // add only unique areas
                    CurAreaFiles = AreaFiles;
                    while (CurAreaFiles != NULL) {
                        if (stricmp( CurAreaFiles->AreaName,
                                     Areas.Selected->Name ) == 0)
                            break;
                        CurAreaFiles = CurAreaFiles->Next;
                    }
                    if (CurAreaFiles == NULL) { // not found
                        if (AreaFiles == NULL) {
                            AreaFiles = new SAreaFiles;
                            LastAreaFiles = AreaFiles;
                        }
                        else {
                            LastAreaFiles->Next = new SAreaFiles;
                            LastAreaFiles = LastAreaFiles->Next;
                        }
                        LastAreaFiles->Next = NULL;
                        strcpy( LastAreaFiles->AreaName,
                                Areas.Selected->Name );
                        LastAreaFiles->Files = LastAreaFiles->LastFile
                            = NULL;
                        i++;
                    }
                }
                if (i == 0) { // no areas added
                    if (AreaAsMask)
                        sprintf( buf,
                                 "! No purgable areas found for mask %s\n",
                                 CurPurgeArea->String );
                    else
                        sprintf( buf, "! Area %s not found.\n",
                                 CurPurgeArea->String );
                    GlobalVars.Log << buf;
                    cout << buf;
                }
            }
        }
    }
    
    strcpy( FindFileMask, GlobalVars.NewTicsPath );
    strcat( FindFileMask, "*.ti?" );
    
    hFindFile = FindFirstFile( FindFileMask, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    strcpy( FullFileName, GlobalVars.NewTicsPath );
    NamePos = strlen( FullFileName );
    while (FileFound) {
        strcpy( FullFileName + NamePos, FindFileData.cFileName );
        if (Tic.Read( FullFileName )) {
            HaveToAddFile = 0;
            do {
                if (!Areas.Select( Tic.Area ))
                    break;
                // try to find area
                CurAreaFiles = AreaFiles;
                while (CurAreaFiles != NULL) {
                    if (stricmp( CurAreaFiles->AreaName, Tic.Area) == 0)
                        break;
                    CurAreaFiles = CurAreaFiles->Next;
                }
                if (CurAreaFiles != NULL) { // area found
                    HaveToAddFile = 1;
                    break;
                }
                if (PurgeAreasList == NULL &&
                    Areas.Selected->PassThrough == 1)
                {
                    // we will add area
                    HaveToAddFile = 1;
                    break;
                }
                break;
                // using pragmas to avoid warning message "unreachable code"
//#pragma warning 13 5
            } while (1);
//#pragma warning 13 4
            if (HaveToAddFile) {
                // prepare file name
                if (*Tic.FullName != '\0')
                    strcpy( FileName, Tic.FullName );
                else
                    strcpy( FileName, Tic.File );
                // add file into area files list,
                // with this purpose find area or add it, if it's absent
                if (CurAreaFiles == NULL) {
                    // add area
                    if (AreaFiles == NULL) {
                        AreaFiles = new SAreaFiles;
                        LastAreaFiles = AreaFiles;
                    }
                    else {
                        LastAreaFiles->Next = new SAreaFiles;
                        LastAreaFiles = LastAreaFiles->Next;
                    }
                    strcpy( LastAreaFiles->AreaName, Tic.Area );
                    // add file
                    LastAreaFiles->Files = new SStringList;
                    LastAreaFiles->Files->String =
                        new char[strlen(FileName) + 1];
                    strcpy( LastAreaFiles->Files->String, FileName );
                    LastAreaFiles->Files->Next = NULL;
                    LastAreaFiles->LastFile = LastAreaFiles->Files;
                    LastAreaFiles->Next = NULL;
                }
                else {
                    // try to find file to exclude duplicates
                    CurFile = CurAreaFiles->Files;
                    while (CurFile != NULL) {
                        if (stricmp( CurFile->String, FileName ) == 0)
                            break;
                        CurFile = CurFile->Next;
                    }
                    if (CurFile == NULL) {
                        // add file
                        if (CurAreaFiles->Files == NULL) {
                            CurAreaFiles->Files = new SStringList;
                            CurAreaFiles->LastFile = CurAreaFiles->Files;
                        }
                        else {
                            CurAreaFiles->LastFile->Next = new SStringList;
                            CurAreaFiles->LastFile =
                                CurAreaFiles->LastFile->Next;
                        }
                        CurAreaFiles->LastFile->String =
                            new char[strlen(FileName) + 1];
                        strcpy( CurAreaFiles->LastFile->String, FileName );
                        CurAreaFiles->LastFile->Next = NULL;
                    }
                }
            }
        }
        FileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );
    // add passthrough areas if not directly specified
    if (PurgeAreasList == NULL) {
        Areas.SelectFirst();
        while (Areas.Selected != NULL) {
            if (Areas.Selected->PassThrough == 1) {
                CurAreaFiles = AreaFiles;
                while (CurAreaFiles != NULL) {
                    if (stricmp( CurAreaFiles->AreaName,
                                 Areas.Selected->Name) == 0)
                        break;
                    CurAreaFiles = CurAreaFiles->Next;
                }
                if (CurAreaFiles == NULL) {
                    // add area
                    if (AreaFiles == NULL) {
                        AreaFiles = new SAreaFiles;
                        LastAreaFiles = AreaFiles;
                    }
                    else {
                        LastAreaFiles->Next = new SAreaFiles;
                        LastAreaFiles = LastAreaFiles->Next;
                    }
                    strcpy( LastAreaFiles->AreaName, Areas.Selected->Name );
                    LastAreaFiles->Files = NULL;
                    LastAreaFiles->LastFile = NULL;
                    LastAreaFiles->Next = NULL;
                }
            }
            Areas.SelectNext();
        }
    }
    // go through all selected areas
    TotalPurgedSize   = 0;
    TotalPurgedNumber = 0;
    CurAreaFiles      = AreaFiles;
    while (CurAreaFiles != NULL) {
        sprintf( buf, "* Area: %s\n", CurAreaFiles->AreaName );
        GlobalVars.Log << buf;
        cout << buf;
        // select area
        Areas.Select( CurAreaFiles->AreaName );
        assert( Areas.Selected != NULL );
        // prepare directory
        strcpy( FullFileName, Areas.Selected->Path );
        NamePos = strlen( FullFileName );
        // prepare mask to look files
        strcpy( FindFileMask, Areas.Selected->Path );
        strcat( FindFileMask, "*.*" );
        // process fecho files
        PurgedNumber = 0; // number of removed files
        PurgedSize   = 0; // size of purged files in this area
        hFindFile    = FindFirstFile( FindFileMask, &FindFileData );
        FileFound    = (hFindFile != INVALID_HANDLE_VALUE);
        while (FileFound) {
            // skip directories, special files and "files.bbs"
            if (!(FindFileData.dwFileAttributes &
                  (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
                   FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_READONLY |
                   FILE_ATTRIBUTE_TEMPORARY))
                &&
                stricmp( FindFileData.cFileName, "files.bbs" ) != 0)
            {
                // let's check, if we have TIC for this file
                CurFile = CurAreaFiles->Files;
                while (CurFile != NULL) {
                    if (stricmp( FindFileData.cFileName, CurFile->String) == 0)
                        break;
                    CurFile = CurFile->Next;
                }
                // remove file, if it's not used
                if (CurFile == NULL) {
                    // form full name
                    strcpy( FullFileName + NamePos, FindFileData.cFileName );
                    if (unlink( FullFileName ) == 0) {
                        sprintf( buf, "        Removed file %s\n",
                                 FindFileData.cFileName );
                        PurgedNumber++;
                        PurgedSize += (FindFileData.nFileSizeHigh * MAXDWORD) +
                            FindFileData.nFileSizeLow;
                    }
                    else
                        sprintf( buf, "        Error removing file %s\n",
                                 FindFileData.cFileName );
                    GlobalVars.Log << buf;
                    cout << buf;
                }
            }
            FileFound = FindNextFile( hFindFile, &FindFileData );
        }
        if (hFindFile != INVALID_HANDLE_VALUE)
            FindClose( hFindFile );
        sprintf( buf, "        %ld file(s) of size ", PurgedNumber );
        FormatSize( PurgedSize, buf + strlen(buf) );
        strcat( buf, " byte(s).\n" );
        GlobalVars.Log << buf;
        cout << buf;
        // change "files.bbs"
        // set file names
        strcpy( FullFileName + NamePos, "files.bbs" );
        strcpy( TempFileName, Areas.Selected->Path );
        strcat( TempFileName, "adrenalin.tmp");
        // check names for spaces
        CurFile = CurAreaFiles->Files;
        while (CurFile != NULL) {
            for (i = 0; CurFile->String[i] != '\0'; i++) {
                if (CurFile->String[i] == ' ')
                    break;
            }
            if (CurFile->String[i] != '\0') {
                strcpy( FileName, "\"" );
                strcat( FileName, CurFile->String );
                strcat( FileName, "\"" );
                delete [] CurFile->String;
                CurFile->String = new char[strlen(FileName) + 1];
                strcpy( CurFile->String, FileName );
            }
            CurFile = CurFile->Next;
        }
        // copy desc file into temp file, but skip lines with absent files
        BOOL overwriteSuccess = TRUE;
        do {
            // check temp file name
            if (!GetTmpFileName( TempFileName ))
                break;
            in.open( FullFileName );
            if (!in)
                break;
            tempout.open( TempFileName );
            if (!tempout) {
                in.close();
                break;
            }
            SkipLongDesc = 0;
            while (!in.eof()) {
                in.getline( buf, LINE_SIZE );
                if (!in) {
                    if (!in.eof())
                        break;
                    continue;
                }
                // check file name
                CurFile = CurAreaFiles->Files;
                while (CurFile != NULL) {
                    if (strnicmp( buf, CurFile->String,
                                  strlen(CurFile->String) ) == 0)
                        break;
                    CurFile = CurFile->Next;
                }
                if (CurFile == NULL)
                    SkipLongDesc = 1;
                else {
                    if (SkipLongDesc) {
                        if (*buf != ' ' && *buf != '>' && *buf != '+') {
                            SkipLongDesc = 0;
                            tempout << buf << endl;
                            if (!tempout) {
                                overwriteSuccess = FALSE;
                                break;
                            }
                        }
                    }
                    else {
                        tempout << buf << endl;
                        if (!tempout) {
                            overwriteSuccess = FALSE;
                            break;
                        }
                    }
                }
            }
            in.close();
            tempout.close();
            if (overwriteSuccess) {
                unlink( FullFileName );
                rename( TempFileName, FullFileName );
            }
            else {
                unlink( TempFileName );
                GlobalVars.Log << "! error updating file " << FullFileName << endl;
            }
            break;
            // using pragmas to avoid warning message "unreachable code"
//#pragma warning 13 5
        } while (1);
//#pragma warning 13 4
        // remove "files.bbs", if it's empty
        hFindFile = FindFirstFile( FullFileName, &FindFileData );
        FileFound = (hFindFile != INVALID_HANDLE_VALUE);
        if (FileFound) {
            FindClose( hFindFile );
            if (((FindFileData.nFileSizeHigh * MAXDWORD) +
                 FindFileData.nFileSizeLow) == 0)
                unlink( FullFileName);
        }
        TotalPurgedNumber += PurgedNumber;
        TotalPurgedSize   += PurgedSize;
        // free item memory
        LastAreaFiles = CurAreaFiles->Next;
        CurFile = CurAreaFiles->Files;
        while (CurFile != NULL) {
            LastFile = CurFile->Next;
            delete [] CurFile->String;
            delete CurFile;
            CurFile = LastFile;
        }
        delete CurAreaFiles;
        CurAreaFiles = LastAreaFiles;
    }
    
    sprintf( buf, "Total: %ld file(s) of size ", TotalPurgedNumber );
    FormatSize( TotalPurgedSize, buf + strlen(buf) );
    strcat( buf, " byte(s).\n" );
    GlobalVars.Log << buf;
    cout << buf;
}

void CAdrenalin::Relink() {
	CMsg		msg;
	CStringList	lstBody;
	char		szLine[256];
	SLink*		pLink;
	int		nAreas;

	// check link
	if (!Links.Select(m_addrRelink)) {
		sprintf( szLine, "Unknown link: %lu:%lu/%lu.%lu@%s\n",
			m_addrRelink.Zone, m_addrRelink.Net, 
			m_addrRelink.Node, m_addrRelink.Point, m_addrRelink.Domain );
		cerr << szLine;
		GlobalVars.Log << "Relink failed:\n" << szLine;
		return;
	}

	// print log message
	sprintf( szLine, "Relink: %lu:%lu/%lu.%lu@%s\n",
		m_addrRelink.Zone, m_addrRelink.Net, 
		m_addrRelink.Node, m_addrRelink.Point, m_addrRelink.Domain );
	GlobalVars.Log << szLine;

	pLink = Links.Selected;

	Areas.SelectFirst();
	nAreas = 0;
	while (Areas.Selected != NULL) {
		if (Areas.Linked( pLink )) {
			// add msg line
			sprintf( szLine, "+%s", Areas.Selected->Name );
			lstBody.Add( szLine );
			nAreas++;
			// print log message
			GlobalVars.Log << szLine << endl;
		}
		Areas.SelectNext();
	}

	GlobalVars.Log << nAreas << " area(s) added into the relink request.\n";

	if (nAreas > 0) {
		lstBody.Add( "" );
		sprintf( szLine, "--- %s v%lu.%lu%s", PROG_NAME, PROG_VER_MAJOR, PROG_VER_MINOR, PROG_VER_SUFFIX );
		lstBody.Add( szLine );
		MakeDirStructure( GlobalVars.NetmailPath );
		// from/to fields reversed because we will use reply method
		msg.From = pLink->Address;
		msg.To   = *(pLink->OurAddr);
		if (m_szRelinkRobotName[0] == '\0') {
			if (pLink->m_szRobotName[0] == '\0')
				strcpy( msg.Header.FromUserName, "AllFix" );
			else
				strcpy( msg.Header.FromUserName, pLink->m_szRobotName );
		}
		else
			strcpy( msg.Header.FromUserName, m_szRelinkRobotName );
		strcpy( msg.Header.ToUserName, GlobalVars.SysOp );
		msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, pLink->Password, lstBody.Strings );
		GlobalVars.Log << "Relink request message created.\n";
	}
	else {
		GlobalVars.Log << "There is no need to create relink request message.\n";
	}
}

void CAdrenalin::LoadForwards() {
	int		i;
	char		szfnForwards[MAX_PATH];
	char*		szFrom;
	SForwardInfo	fwdInfo;
	int		nCurrentDays;

	// 0. Clear vector and prepare time
	m_vectForwardInfo.clear();
	SYSTEMTIME	systemTime;
	GetLocalTime( &systemTime );
	nCurrentDays = GetDays( systemTime.wDay, systemTime.wMonth, systemTime.wYear %= 100 );

	// 1. Prepare filename
	strcpy( szfnForwards, StartupPath );
	strcat( szfnForwards, szForwardsFileName );
	// 2. Open file for reading
	ifstream	in(szfnForwards);
	if (!in.fail()) {
		// 3. consequently read lines
		char	szLine[LINE_SIZE];
		while (!in.eof()) {
			in.getline( szLine, LINE_SIZE );
			if (in.fail())
				break;
			// 3.1. Get date
			StrNCpy( fwdInfo.m_szDate, szLine, 6 );
			// 3.1.1. Check date 
			if (GlobalVars.m_nForwardExpirationTime != 0) {
				WORD	wDay   = (fwdInfo.m_szDate[0] - '0')*10 + (fwdInfo.m_szDate[1] - '0');
				WORD	wMonth = (fwdInfo.m_szDate[2] - '0')*10 + (fwdInfo.m_szDate[3] - '0');
				WORD	wYear  = (fwdInfo.m_szDate[4] - '0')*10 + (fwdInfo.m_szDate[5] - '0');
				int	nDays = GetDays( wDay, wMonth, wYear );
				// skip expired request
				if (nCurrentDays - nDays > GlobalVars.m_nForwardExpirationTime) {
					m_fForwardsChanged = true;
					continue;
				}
			}
			// 3.2. Get link
			i = 0;
			szFrom = szLine + 7;
			while (*szFrom != ' ') {
				fwdInfo.m_szLink[i] = *szFrom;
				i++;
				szFrom++;
			}
			fwdInfo.m_szLink[i] = '\0';
			// 3.3. Get area
			i = 0;
			szFrom++;
			while (*szFrom != '\0') {
				fwdInfo.m_szArea[i] = *szFrom;
				i++;
				szFrom++;
			}
			fwdInfo.m_szArea[i] = '\0';
			m_vectForwardInfo.push_back(fwdInfo);
		}
		in.close();
	}

	// in spite of read file result we should set flag
	m_fForwardsRead = true;
}

void CAdrenalin::SaveForwards() {
	char		szfnForwards[MAX_PATH];

	// 1. Prepare filename
	strcpy( szfnForwards, StartupPath );
	strcat( szfnForwards, szForwardsFileName );
	// 2. Open file for writing
	ofstream	out(szfnForwards);
	if (!out.fail()) {
		CForwardInfoVector::iterator	i;
		for (i = m_vectForwardInfo.begin(); i != m_vectForwardInfo.end(); ++i) {
			out << i->m_szDate << " " << i->m_szLink << " " << i->m_szArea << endl;
		}
		out.close();
	}
}

void CAdrenalin::GetErrorMessage(int nErrorCode, char* szMessage) {
	switch (nErrorCode) {
	case ERROR_UNKNOWN_ARG:
	    strcpy( szMessage, "Unknown argument" );
	    break;
	case ERROR_HATCH_NO_FILE_DEF:
	    strcpy( szMessage, "No file defined to hatch" );
	    break;		
	case ERROR_HATCH_NO_AREA_DEF:
	    strcpy( szMessage, "No area defined to hatch" );
	    break;
	case ERROR_HATCH_EXTRA_ARG:
	    strcpy( szMessage, "Extra arguments for hatch" );
	    break;
	case ERROR_SYNTAX:
	    strcpy( szMessage, "Syntax error" );
	    break;
	case ERROR_NUMBER:
	    strcpy( szMessage, "Wrong number of arguments" );
	    break;
	case ERROR_DUPLICATE:
	    strcpy( szMessage, "Duplicate usage of a keyword" );
	    break;
	case ERROR_FILE_ABSENT:
	    strcpy( szMessage, "File absent" );
	    break;
	case ERROR_UNKNOWN_KEYWORD:
	    strcpy( szMessage, "Unknown keyword" );
	    break;
	case ERROR_FORWARD_DEF:
	    strcpy( szMessage, "Forward definition" );
	    break;
	case ERROR_ABSENT_DEF:
	    strcpy( szMessage, "Definition is absent" );
	    break;
	case ERROR_MEMORY:
	    strcpy( szMessage, "Not enough memory" );
	    break;
	case ERROR_AREA_NOT_FOUND:
	    strcpy( szMessage, "Area not found" );
	    break;
	case ERROR_MOVE_FAILED:
	    strcpy( szMessage, "Failed to move file" );
	    break;
	case ERROR_CRC32_FAILED:
	    strcpy( szMessage, "Failed to calculate CRC32" );
	    break;
	case ERROR_ALREADY_WORK:
	    strcpy( szMessage, "Already started" );
	    break;
	case ERROR_UNKNOWN_AKA:
	    strcpy( szMessage, "Unknown AKA" );
	    break;
	case ERROR_LNG_READ:
	    strcpy( szMessage, "Error reading language file" );
	    break;
	case ERROR_LNG_FILE_NOT_FOUND:
	    strcpy( szMessage, "Language file not found" );
	    break;
	case ERROR_LNG_NUMBER:
	    strcpy( szMessage, "Wrong number of language messages" );
	    break;
	case ERROR_HATCH_NO_DESC_DEF:
	    strcpy( szMessage, "Description not defined for hatching" );
	    break;
	case ERROR_HATCH_NO_REPLACES:
	    strcpy( szMessage, "REPLACES value absent for hatching" );
	    break;
	case ERROR_UNKNOWN_OPT:
	    strcpy( szMessage, "Unknown option" );
	    break;
	case ERROR_LOW_DISK_SPACE:
	    strcpy( szMessage, "Low disk space" );
	    break;
	case ERROR_UNDEFINED_ADDRESS:
	    strcpy( szMessage, "Undefined address for relink" );
	    break;
	case ERROR_WRONG_ADDRESS:
	    strcpy( szMessage, "Wrong relink address" );
	    break;
	default:
		strcpy( szMessage, "Unknown error" );
	}
}
