// changes log:
// 02.02.2000 01:19 added m_addrRelink and m_szRelinkRobotName fields, Relink method into 
//                  CAdrenalin

#ifndef __ADRENALIN__HEADER__
#define __ADRENALIN__HEADER__

#include "pch.h"

#include "FTNStruc.hpp"
#include "Common.hpp"
#include "Links.hpp"
#include "Areas.hpp"
#include "Lng.hpp"

#define PROG_NAME       "Adrenalin"
#define PROG_VER_MAJOR  (unsigned long)0
#define PROG_VER_MINOR  (unsigned long)65
#define PROG_VER_SUFFIX "release"

const
    // Action
    ACTION_NONE   = 0,
    ACTION_HELP   = 1,
    ACTION_FIX    = 2,
    ACTION_TOSS   = 4,
    ACTION_HATCH  = 8,
    ACTION_REPORT = 16,
    ACTION_ERROR  = 32,
    ACTION_DEBUG  = 64,
    ACTION_PURGE  = 128,
    ACTION_RELINK = 256,
    // ErrorCode
    ERROR_NONE               = 0,
    ERROR_UNKNOWN_ARG        = 1,
    ERROR_HATCH_NO_FILE_DEF  = 2,
    ERROR_HATCH_NO_AREA_DEF  = 3,
    ERROR_HATCH_EXTRA_ARG    = 4,
    ERROR_SYNTAX             = 5,
    ERROR_NUMBER             = 6,
    ERROR_DUPLICATE          = 7,
    ERROR_FILE_ABSENT        = 8,
    ERROR_UNKNOWN_KEYWORD    = 9,
    ERROR_FORWARD_DEF        = 10,
    ERROR_ABSENT_DEF         = 11,
    ERROR_MEMORY             = 12,
    ERROR_AREA_NOT_FOUND     = 13,
    ERROR_MOVE_FAILED        = 14,
    ERROR_CRC32_FAILED       = 15,
    ERROR_ALREADY_WORK       = 16,
    ERROR_UNKNOWN_AKA        = 17,
    ERROR_LNG_READ           = 18,
    ERROR_LNG_FILE_NOT_FOUND = 19,
    ERROR_LNG_NUMBER         = 20,
    ERROR_HATCH_NO_DESC_DEF  = 21,
    ERROR_HATCH_NO_REPLACES  = 22,
    ERROR_UNKNOWN_OPT        = 23,
    ERROR_LOW_DISK_SPACE     = 24,
    ERROR_UNDEFINED_ADDRESS  = 25,	// RELINK without address
    ERROR_WRONG_ADDRESS      = 26;	// RELINK with wrong address


struct SForwardInfo {
	char	m_szDate[7];
	char	m_szLink[80];
	char	m_szArea[MAX_AREA_SIZE];
};

typedef vector<SForwardInfo>	CForwardInfoVector;

struct DiskLimitsList {
	char		root[MAX_PATH];
	DWORD		limit;
	DiskLimitsList*	next;
};

class CAdrenalin {
public:
    CAdrenalin(int argc, char **argv);
    ~CAdrenalin();
    int Execute();
    CLinks       Links;
protected:
	SFTNAddress		m_addrRelink;
	char			m_szRelinkRobotName[36];

	// forwards-related stuff

	CForwardInfoVector	m_vectForwardInfo;
	bool			m_fForwardsRead;
	bool			m_fForwardsChanged;
	void			LoadForwards();
	void			SaveForwards();

	void		Relink();

	void GetErrorMessage(int nErrorCode, char* szMessage);
private:
    HANDLE       hBusyMutex;
    DWORD        Action;
    int          ErrorCode;

    char         StartupPath[MAX_PATH];
    char         LngFileName[MAX_PATH];
    
    CAreas       Areas;

    CLng         Language;

    char        HatchFile[MAX_PATH];
    char        HatchArea[MAX_PATH];
    char        HatchDesc[MAX_PATH];
    char        HatchReplaces[MAX_PATH];
    BOOL	leaveHatchedFile;

    SStringList *PurgeAreasList, *LastPurgeArea;
    
    DiskLimitsList*	diskLimits;

    void ParseAppCfg();

    void checkDiskLimits();

    void ParseAreasCfg();
    void ParseCmdLine(int argc, char **argv);
    
    int  ParseLinkAddr(char *s, SFTNAddress &Addr);
    int  ParseLinkAccess(char *s, SLink &Link);
    
    int  ParseAreaAccess(char *s, SArea &Area);
    
    void PrintHelp();
    void PrintDebugInfo();

    void ProcessFixRequests();
    void ProcessInboundTics();
    int  Hatch();
    void Purge();

    void ChooseAKA(SLink *Link);

    void AddStartupPath(char *Location);

    int  SaveCfg();

    void  MoveTicToBads(char *FullTicName, int TicNameFrom);
};

extern CAdrenalin*	pAdrenalin;

#endif