#include "pch.h"

#include "Common.hpp"
#include "Misc.hpp"
#include "Temp.hpp"
#include "CfgParse.hpp"
#include "Areas.hpp"

CAreas::CAreas() {
    AreasList = LastArea = Selected = NULL;
    MaxAreaNameLen = MaxAreaPathLen = 0;
}

CAreas::~CAreas() {
    while (AreasList != NULL) {
        while (AreasList->LinkedList != NULL) {
            AreasList->LastLinked = AreasList->LinkedList->Next;
            delete AreasList->LinkedList;
            AreasList->LinkedList = AreasList->LastLinked;
        }
        LastArea = AreasList->Next;
        delete AreasList;
        AreasList = LastArea;
    }
}

void CAreas::Add(SArea &NewArea) {
    if (AreasList == NULL) {
        AreasList = new SArea;
        LastArea  = AreasList;
    }
    else {
        LastArea->Next = new SArea;
        LastArea       = LastArea->Next;
    }
    memcpy( LastArea, &NewArea, sizeof(SArea) );
    LastArea->Next = NULL;
    if (strlen(NewArea.Name) > MaxAreaNameLen)
        MaxAreaNameLen = strlen(NewArea.Name);
    if (strlen(NewArea.Path) > MaxAreaPathLen)
        MaxAreaPathLen = strlen(NewArea.Path);
}

void CAreas::AutoCreate(char *NewAreaName, SLink *Creator) {
    int  i;
    char ShortName[AREA_NAME_SIZE];
    
    if (AreasList == NULL) {
        AreasList = new SArea;
        LastArea  = AreasList;
    }
    else {
        LastArea->Next = new SArea;
        LastArea       = LastArea->Next;
    }
    LastArea->Next = NULL;
    strcpy( LastArea->Name, NewAreaName );
    if (strlen(NewAreaName) > MaxAreaNameLen)
        MaxAreaNameLen = strlen(NewAreaName);
    // set description
    LastArea->Desc[0] = '\0';
    if (GlobalVars.AutoDescFileName[0] != '\0') {
        CCfgParser DescList( GlobalVars.AutoDescFileName );
        if (!DescList.Failed()) {
            while (!DescList.Eof()) {
                DescList.ReadLine();
                if (DescList.Failed()) {
                    if (!DescList.Eof())
                        break;
                    else
                        continue;
                }
                if (DescList.Count() < 3)
                    continue;
                if (stricmp( DescList[0], "Area" ) == 0 &&
                    stricmp( DescList[1], NewAreaName ) == 0)
                {
                    StrNCpy( LastArea->Desc, DescList[2], AREA_DESC_SIZE );
                    for (i = 0; LastArea->Desc[i] != '\0'; i++) {
                        if (LastArea->Desc[i] == '_')
                            LastArea->Desc[i] = ' ';
                    }
                    break;
                }
            }
        }
    }
    LastArea->RouteType       = ROUTE_NONE;
    LastArea->Priority        = 0;
    LastArea->Group           = Creator->defaultGroup;
    LastArea->Line            = -1;
    LastArea->PassThrough     = 0;
    LastArea->VetoManualPurge = 0;
    LastArea->Traffic         = 0;
    LastArea->LinksChanged    = 1;
    LastArea->LinkedList      = new SLinked;
    LastArea->LastLinked      = LastArea->LinkedList;
    LastArea->LastLinked->LinkPtr   = Creator;
    LastArea->LastLinked->RouteType = ROUTE_SEND | Creator->RouteType;
    LastArea->LastLinked->Next      = NULL;
    // form fecho path
    char	szPath[MAX_PATH];
    if (Creator->szAutoCreatedPathName[0] != '\0')
	    strcpy( szPath, Creator->szAutoCreatedPathName );
    else
	    strcpy( szPath, GlobalVars.AutoCreatedPath );
 
    if (GlobalVars.LongPaths)
        sprintf( LastArea->Path, "%s%s\\", szPath, NewAreaName );
    else {
        i = strlen(NewAreaName);
        if (i > 8)
            i = 8;
        memcpy( ShortName, NewAreaName, i );
        ShortName[i] = '\0';
        for (i = 0; ShortName[i] != '\0'; i++) {
            if (ShortName[i] == '.')
                ShortName[i] = '_';
        }
        sprintf( LastArea->Path, "%s%s\\", szPath, ShortName );
    }
    if (access( LastArea->Path, 0 ) == 0) {
        i = 0;
        do {
            if (GlobalVars.LongPaths)
                sprintf( LastArea->Path, "%s%s.%03ld\\", szPath,
                         NewAreaName, i );
            else
                sprintf( LastArea->Path, "%s%s.%03lX\\", szPath,
                         ShortName, i );
            i++;
        } while (access( LastArea->Path, 0 ) == 0);
    }
    MakeDirStructure( LastArea->Path );
    if (strlen(LastArea->Path) > MaxAreaPathLen)
        MaxAreaPathLen = strlen(LastArea->Path);
    Selected = LastArea;
}

int CAreas::SelectFirst() {
    Selected = AreasList;
    return (Selected != NULL);
}

int CAreas::SelectNext() {
    if (Selected != NULL)
        Selected = Selected->Next;
    return (Selected != NULL);
}

int CAreas::Select(char *AreaName) {
    Selected = AreasList;
    while (Selected != NULL) {
        if (stricmp( Selected->Name, AreaName ) == 0)
            break;
        Selected = Selected->Next;
    }
    return (Selected != NULL);
}

int CAreas::CanAccess(SLink *Link) {
    if (Selected == NULL)
        return (0);
    if (Link->Priority < Selected->Priority)
        return (0);
    if (Selected->Group != 0 && (Selected->Group & Link->Groups) == 0)
        return (0);
    return (1);
}

SLinked* CAreas::Linked(SLink *Link) {
    SLinked *CurLinked;
    
    if (Selected == NULL)
        return (0);
    CurLinked = Selected->LinkedList;
    while (CurLinked != NULL) {
        if (CurLinked->LinkPtr->Address.Zone  == Link->Address.Zone  &&
            CurLinked->LinkPtr->Address.Net   == Link->Address.Net   &&
            CurLinked->LinkPtr->Address.Node  == Link->Address.Node  &&
            CurLinked->LinkPtr->Address.Point == Link->Address.Point &&
            stricmp( CurLinked->LinkPtr->Address.Domain, Link->Address.Domain )
            == 0)
            break;
        CurLinked = CurLinked->Next;
    }
    return (CurLinked);
}

int CAreas::Link(SLink *Link) {
    if (Selected == NULL)
        return (0);
    if (Link->Priority < Selected->Priority)
        return (0);
    if (Selected->Group != 0 && (Selected->Group & Link->Groups) == 0)
        return (0);
    if (Selected->LinkedList == NULL) {
        Selected->LinkedList = new SLinked;
        Selected->LastLinked = Selected->LinkedList;
    }
    else {
        Selected->LastLinked->Next = new SLinked;
        Selected->LastLinked = Selected->LastLinked->Next;
    }
    Selected->LastLinked->Next = NULL;
    Selected->LastLinked->LinkPtr = Link;
    if (Selected->RouteType == ROUTE_NONE)
        Selected->LastLinked->RouteType = Link->RouteType;
    else
        Selected->LastLinked->RouteType = Selected->RouteType;
    Selected->LinksChanged = 1;
    return (1);
}

int CAreas::Unlink(SLink *Link) {
    SLinked *PrevLinked;
    SLinked *CurLinked;

    if (Selected == NULL)
        return (0);
    PrevLinked = NULL;
    CurLinked = Selected->LinkedList;
    while (CurLinked != NULL) {
        if (CurLinked->LinkPtr->Address.Zone  == Link->Address.Zone  &&
            CurLinked->LinkPtr->Address.Net   == Link->Address.Net   &&
            CurLinked->LinkPtr->Address.Node  == Link->Address.Node  &&
            CurLinked->LinkPtr->Address.Point == Link->Address.Point &&
            stricmp( CurLinked->LinkPtr->Address.Domain, Link->Address.Domain )
            == 0)
            break;
        PrevLinked = CurLinked;
        CurLinked = CurLinked->Next;
    }
    if (CurLinked == NULL)
        return (0);
    if (PrevLinked == NULL)
        Selected->LinkedList = CurLinked->Next;
    else
        PrevLinked->Next = CurLinked->Next;
    delete CurLinked;
    Selected->LinksChanged = 1;
    return (1);
}

int CAreas::LinksChanged() {
    SArea *CurArea;

    CurArea = AreasList;
    while (CurArea != NULL) {
        if (CurArea->LinksChanged)
            break;
        CurArea = CurArea->Next;
    }
    return (CurArea != NULL);
}

int CAreas::SaveToFile(char *FileName) {
    char      Line[LINE_SIZE];
    char      SubLine[LINE_SIZE];
    int       LineLen, Len;
    int       SubLineLen;
    SArea    *CurArea;
    ofstream  out;
    char      RouteChar;
    SLinked  *CurLinked;

    out.open( FileName );
    if (!out)
        return (0);

    CurArea = AreasList;
    while (CurArea != NULL) {
        switch (CurArea->RouteType) {
        case ROUTE_NONE:
            RouteChar = ' ';
            break;
        case ROUTE_RECEIVE:
            RouteChar = 'R';
            break;
        case ROUTE_SEND:
            RouteChar = 'S';
            break;
        case ROUTE_BIDIRECT:
            RouteChar = 'B';
            break;
        }
        sprintf( Line, "Area %s", CurArea->Name );
        LineLen = strlen( Line );
        Len = strlen(CurArea->Name);
        while (Len < MaxAreaNameLen) {
            Line[LineLen++] = ' ';
            Len++;
        }
        Line[LineLen] = '\0';
        sprintf( Line + LineLen, " %s", CurArea->Path );
        LineLen = strlen( Line );
        Len = strlen( CurArea->Path );
        while (Len < MaxAreaPathLen) {
            Line[LineLen++] = ' ';
            Len++;
        }
        Line[LineLen] = '\0';
        sprintf( Line + LineLen, " %c", RouteChar );
        LineLen += 2;
        if (CurArea->PassThrough == 1) {
            strcpy( Line + LineLen, " P" );
            LineLen += 2;
        }
        if (CurArea->VetoManualPurge == 1) {
            strcpy( Line + LineLen, " V" );
            LineLen += 2;
        }
        if (CurArea->Group != 0) {
            sprintf( Line + LineLen, " %ld/%c", CurArea->Priority,
                     (char)GetOnBit(CurArea->Group) + 'A' );
        }
        else if (CurArea->Priority != 0)
            sprintf( Line + LineLen, " %ld", CurArea->Priority );
        LineLen = strlen( Line );
        if (CurArea->Desc[0] != '\0')
            sprintf( Line + LineLen, " \"%s\"", CurArea->Desc );
        out << Line << endl;
        strcpy( Line, "Linked" );
        CurLinked = CurArea->LinkedList;
        while (CurLinked != NULL) {
            LineLen = strlen( Line );
            if (CurLinked->RouteType != CurLinked->LinkPtr->RouteType) {
                switch (CurLinked->RouteType) {
                case ROUTE_RECEIVE:
                    RouteChar = 'R';
                    break;
                case ROUTE_SEND:
                    RouteChar = 'S';
                    break;
                case ROUTE_BIDIRECT:
                    RouteChar = 'B';
                    break;
                }
                sprintf( SubLine, " %c", RouteChar );
            }
            else
                strcpy( SubLine, " " );
            SubLineLen = strlen( SubLine );
            if (stricmp( GlobalVars.MainAddr.Domain,
                         CurLinked->LinkPtr->Address.Domain ) != 0)
            {
                sprintf( SubLine + SubLineLen, "%ld:%ld/%ld.%ld@%s",
                         CurLinked->LinkPtr->Address.Zone,
                         CurLinked->LinkPtr->Address.Net,
                         CurLinked->LinkPtr->Address.Node,
                         CurLinked->LinkPtr->Address.Point,
                         CurLinked->LinkPtr->Address.Domain );
            }
            else {
                if (CurLinked->LinkPtr->Address.Zone ==
                    GlobalVars.MainAddr.Zone)
                {
                    if (CurLinked->LinkPtr->Address.Net ==
                        GlobalVars.MainAddr.Net)
                    {
                        sprintf( SubLine + SubLineLen, "%ld",
                                 CurLinked->LinkPtr->Address.Node );
                    }
                    else {
                        sprintf( SubLine + SubLineLen, "%ld/%ld",
                                 CurLinked->LinkPtr->Address.Net,
                                 CurLinked->LinkPtr->Address.Node );
                    }
                }
                else {
                    sprintf( SubLine + SubLineLen, "%ld:%ld/%ld",
                             CurLinked->LinkPtr->Address.Zone,
                             CurLinked->LinkPtr->Address.Net,
                             CurLinked->LinkPtr->Address.Node );
                }
                if (CurLinked->LinkPtr->Address.Point != 0) {
                    SubLineLen = strlen( SubLine );
                    sprintf( SubLine + SubLineLen, ".%ld",
                             CurLinked->LinkPtr->Address.Point );
                }
            }
            if (strlen(Line) + strlen(SubLine) > 80) {
                Line[LineLen] = '\0';
                out << Line << endl;
                Line[6] = '\0'; // Line == "Linked"
            }
            else {
                strcat( Line, SubLine );
                CurLinked = CurLinked->Next;
            }
        }
        if (strlen(Line) > 6)
            out << Line << endl;
        CurArea = CurArea->Next;
        if (CurArea != NULL)
            out << endl;
    }
    out.close();
    return (1);
}


void CAreas::GetNameDesc(char *NameDesc) {
	int	Len;
	char	desc[1024];
    
        if (Selected == NULL)
                *NameDesc = '\0';
        else {
                strcpy( NameDesc, "  " );
                strcat( NameDesc, Selected->Name );
                Len = strlen( NameDesc );
                NameDesc[Len++] = ' ';
                while (Len < 39)
                	NameDesc[Len++] = '.';
                NameDesc[Len++] = ' ';
                if (Selected->Desc[0] != '\0') {
                	alignDescription( Selected->Desc, desc, 38, 40 );
                	strcpy( NameDesc + Len, desc );
                }
                else
                	strncpy( NameDesc + Len, "no description", 40 );
        }
}

void CAreas::alignDescription(char* desc, char* aligned, int len, int prefixLen) {
	int	i, j, pos;
	int	wordStart, wordLen;

	i   = 0;
	j   = 0;
	pos = 0;
	wordStart = -1;
	while (desc[i] != '\0') {
		if (pos == len) {
			aligned[j] = 13;
			j++;
			memset( aligned + j, ' ', prefixLen );
			j += prefixLen;
			pos = 0;
		}
		if (desc[i] == ' ') {
			if (wordStart >= 0) {
				if (wordLen > len) {
					memcpy( aligned + j, desc + wordStart, len - pos );
					j += len - pos;
					aligned[j] = 13;
					j++;
					wordStart += len - pos;
					wordLen -= len - pos;
					while (wordLen > len) {
						memset( aligned + j, ' ', prefixLen );
						j += prefixLen;
						memcpy( aligned + j, desc + wordStart, len );
						j += len;
						aligned[j] = 13;
						j++;
						wordLen -= len;
						wordStart += len;
					}
					if (wordLen > 0) {
       						memset( aligned + j, ' ', prefixLen );
       						j += prefixLen;
						memcpy( aligned + j, desc + wordStart, wordLen );
						j += wordLen;
					}
					pos = wordLen;
				}
				else {
					if (pos + wordLen < len) {
        					memcpy( aligned + j, desc + wordStart, wordLen );
        					j += wordLen;
        					pos += wordLen;
					}
					else {
						memset( aligned + j, ' ', len - pos );
						j += (len - pos);
						aligned[j] = 13;
						j++;
						memset( aligned + j, ' ', prefixLen );
						j += prefixLen;
						memcpy( aligned + j, desc + wordStart, wordLen );
						j += wordLen;
						pos = wordLen;
					}
				}
				wordStart = -1;
			}
			if (pos > 0 && pos < len) {
				aligned[j] = ' ';
				j++;
				pos++;
			}
		}
		else {
			if (wordStart < 0) {
				wordStart = i;
				wordLen   = 1;
			}
			else
				wordLen++;
		}
		i++;
	}
 	if (wordStart >= 0) {
 		if (wordLen > len) {
			memcpy( aligned + j, desc + wordStart, len - pos );
			j += len - pos;
			aligned[j] = 13;
			j++;
			wordStart += len - pos;
			wordLen -= len - pos;
 			while (wordLen > len) {
				memset( aligned + j, ' ', prefixLen );
 				j += prefixLen;
 				memcpy( aligned + j, desc + wordStart, len );
 				j += len;
 				aligned[j] = 13;
 				j++;
 				wordLen -= len;
 				wordStart += len;
 			}
 			if (wordLen > 0) {
					memset( aligned + j, ' ', prefixLen );
					j += prefixLen;
 				memcpy( aligned + j, desc + wordStart, wordLen );
 				j += wordLen;
 			}
 			pos = wordLen;
 		}
 		else {
 			if (pos + wordLen < len) {
 				memcpy( aligned + j, desc + wordStart, wordLen );
 				j += wordLen;
 				pos += wordLen;
 			}
 			else {
 				memset( aligned + j, ' ', len - pos );
 				j += (len - pos);
 				aligned[j] = 13;
 				j++;
 				memset( aligned + j, ' ', prefixLen );
 				j += prefixLen;
 				memcpy( aligned + j, desc + wordStart, wordLen );
 				j += wordLen;
 				pos = wordLen;
 			}
 		}
 	}
	aligned[j] = '\0';
}

int CAreas::GetRouteType(SLink *Link) {
    SLinked *CurLinked;
    
    if (Selected == NULL)
        return (ROUTE_NONE);
    CurLinked = Selected->LinkedList;
    while (CurLinked != NULL) {
        if (CurLinked->LinkPtr->Address.Zone  == Link->Address.Zone  &&
            CurLinked->LinkPtr->Address.Net   == Link->Address.Net   &&
            CurLinked->LinkPtr->Address.Node  == Link->Address.Node  &&
            CurLinked->LinkPtr->Address.Point == Link->Address.Point &&
            stricmp( CurLinked->LinkPtr->Address.Domain, Link->Address.Domain )
            == 0)
            break;
        CurLinked = CurLinked->Next;
    }
    return (CurLinked != NULL ? CurLinked->RouteType : ROUTE_NONE );
}

int CAreas::UpdateDescFile(char *FileName, char *FileDesc,
                           SStringList* LDescList)
{
    ofstream     out;
    ifstream     in;
    ofstream     tempout;
    char         DescFileName[MAX_PATH];
    char         TempFileName[MAX_PATH];
    char         WorkFileName[MAX_PATH];
    char         buf[LINE_SIZE];
    int          Len;
    int          i;
    int          SkipLongDesc;
    SStringList *CurLDesc;
    
    if (Selected == NULL)
        return (0);
    // set file names
    strcpy( DescFileName, Selected->Path );
    strcat( DescFileName, "files.bbs" );
    strcpy( TempFileName, Selected->Path );
    strcat( TempFileName, "adrenalin.tmp");
    Len = strlen( FileName );
    // check name for spaces
    for (i = 0; FileName[i] != '\0'; i++) {
        if (FileName[i] == ' ')
            break;
    }
    if (FileName[i] != '\0') {
        strcpy( WorkFileName, "\"" );
        strcat( WorkFileName, FileName );
        strcat( WorkFileName, "\"" );
        Len += 2;
    }
    else
        strcpy( WorkFileName, FileName );
    // copy desc file into temp file, but skip lines with WorkFileName
    BOOL successOverwrite = TRUE;
    do {
        // check temp file name
        if (!GetTmpFileName( TempFileName ))
            break;
        in.open( DescFileName );
        if (!in)
            break;
        tempout.open( TempFileName );
        if (!tempout) {
            in.close();
            break;
        }
        SkipLongDesc = 0;
        while (!in.eof()) {
            in.getline( buf, LINE_SIZE );
            if (!in) {
                if (!in.eof())
                    break;
                continue;
            }
            if (strnicmp( buf, WorkFileName, Len) == 0)
                SkipLongDesc = 1;
            else {
                if (SkipLongDesc) {
                    if (*buf != ' ' && *buf != '>' && *buf != '+') {
                        SkipLongDesc = 0;
                        tempout << buf << endl;
                        if (!tempout) {
                            successOverwrite = FALSE;
                            break;
                        }
                    }
                }
                else {
                    tempout << buf << endl;
                    if (!tempout) {
                        successOverwrite = FALSE;
                        break;
                    }
                }
            }
        }
        in.close();
        tempout.close();
        if (successOverwrite) {
            unlink( DescFileName );
            rename( TempFileName, DescFileName );
        } else {
            unlink( TempFileName );
            GlobalVars.Log << "! error updating " << DescFileName << endl;
        }
        break;
        // using pragmas to disable "unreachable code" warning
//#pragma warning 13 5
    } while (1);
//#pragma warning 13 4
    
    out.open( DescFileName, ios::app );
    if (!out)
        out.open( DescFileName );
    if (!out)
        return (0);
    out << WorkFileName;
    if (*FileDesc != '\0') {
        while (Len < 12) {
            out << " ";
            Len++;
        }
        out << " ";
        out << FileDesc;
    }
    out << endl;
    CurLDesc = LDescList;
    while (CurLDesc != NULL) {
        out << GlobalVars.LongDescPrefix << CurLDesc->String << endl;
        CurLDesc = CurLDesc->Next;
    }
    out.close();
    return (1);
}

int CAreas::IsValidName(char *Name) {
    while (*Name != '\0') {
        if (*Name == '\\' || *Name == '/' || *Name == ':' ||
            *Name == '*' || *Name == '?' || *Name == '<' || *Name == '>'
            || *Name == '\'' || *Name == '"')
            break;
        Name++;
    }
    return (*Name == '\0');
}

void CAreas::ClearDescFile(char *FileName) {
    ifstream     in;
    ofstream     tempout;
    char         DescFileName[MAX_PATH];
    char         TempFileName[MAX_PATH];
    char         WorkFileName[MAX_PATH];
    char         buf[LINE_SIZE];
    int          Len;
    int          i;
    int          SkipLongDesc;
    
    if (Selected == NULL)
        return;
    // set file names
    strcpy( DescFileName, Selected->Path );
    strcat( DescFileName, "files.bbs" );
    strcpy( TempFileName, Selected->Path );
    strcat( TempFileName, "adrenalin.tmp");
    Len = strlen( FileName );
    // check name for spaces
    for (i = 0; FileName[i] != '\0'; i++) {
        if (FileName[i] == ' ')
            break;
    }
    if (FileName[i] != '\0') {
        strcpy( WorkFileName, "\"" );
        strcat( WorkFileName, FileName );
        strcat( WorkFileName, "\"" );
        Len += 2;
    }
    else
        strcpy( WorkFileName, FileName );
    // copy desc file into temp file, but skip lines with WorkFileName
    do {
        // check temp file name
        if (!GetTmpFileName( TempFileName ))
            break;
        in.open( DescFileName );
        if (!in)
            break;
        tempout.open( TempFileName );
        if (!tempout) {
            in.close();
            break;
        }
        SkipLongDesc = 0;
        while (!in.eof()) {
            in.getline( buf, LINE_SIZE );
            if (!in)
                continue;
            if (strnicmp( buf, WorkFileName, Len) == 0)
                SkipLongDesc = 1;
            else {
                if (SkipLongDesc) {
                    if (*buf != ' ' && *buf != '>' && *buf != '+') {
                        SkipLongDesc = 0;
                        tempout << buf << endl;
                    }
                }
                else
                    tempout << buf << endl;
            }
        }
        in.close();
        tempout.close();
        unlink( DescFileName );
        rename( TempFileName, DescFileName );
        break;
        // using pragmas to disable "unreachable code" warning
//#pragma warning 13 5
    } while (1);
//#pragma warning 13 4
}

void CAreas::Sort() {
    char   SelAreaName[AREA_NAME_SIZE];
    SArea *PrevPrevArea, *PrevArea, *CurArea;
    int    WasExchange;

    if (AreasList == NULL)
        return;
    
    // memorize selected area
    if (Selected != NULL)
        strcpy( SelAreaName, Selected->Name );
    else
        *SelAreaName = '\0';

    do {
        WasExchange  = 0;
        PrevPrevArea = NULL;
        PrevArea     = AreasList;
        CurArea      = AreasList->Next;
        while (CurArea != NULL) {
            // compare current and previous areas
            if (PrevArea->Group > CurArea->Group ||
                (PrevArea->Group == CurArea->Group &&
                 stricmp( PrevArea->Name, CurArea->Name ) > 0))
            {
                // previous area > current
                // exchange previous and current areas
                if (PrevPrevArea == NULL)
                    AreasList = CurArea;
                else
                    PrevPrevArea->Next = CurArea;
                PrevArea->Next = CurArea->Next;
                CurArea->Next  = PrevArea;
                CurArea        = PrevArea->Next;
                WasExchange = 1;
            }
            else {
                PrevArea = PrevArea->Next;
                CurArea  = CurArea->Next;
            }
            if (PrevPrevArea == NULL)
                PrevPrevArea = AreasList;
            else
                PrevPrevArea = PrevPrevArea->Next;
        }
    } while (WasExchange);
    
    // restore selected area
    if (*SelAreaName != '\0')
        Select( SelAreaName );

    // restore pointer to last area
    LastArea = AreasList;
    while (LastArea->Next != NULL)
        LastArea = LastArea->Next;
}
