#ifndef __ADRENALIN__FTNSTRUC__HEADER__
#define __ADRENALIN__FTNSTRUC__HEADER__

// uncomment definition below to use signed addresses
#define SIGNED_ADDR

// uncomment definition below if you have AKA addresses
// in this case you should also define AKA[] in Msg.cpp
// #define HAVE_AKA

const MAX_DOMAIN_SIZE = 20;
const MAX_AREA_SIZE   = 96;  // Maximum area name size (with '\0')
const MAX_KLUDGE_SIZE = 512; // Maximum kludge line size
const MAX_FLAG_SIZE   = 20;  // Maximum flag size of ^AFLAGS kludge

typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned long  uint32;
typedef signed char    int8;
typedef signed short   int16;
typedef signed long    int32;

#ifdef SIGNED_ADDR
typedef signed short   addr;
#else
typedef unsigned short addr;
#endif

const ATTR_PRIVATE            = 1,
      ATTR_CRASH              = 2,
      ATTR_RECEIVED           = 4,
      ATTR_SENT               = 8,
      ATTR_FILE_ATTACHED      = 16,
      ATTR_TRANSIT            = 32,
      ATTR_ORPHAN             = 64,
      ATTR_KILL_SENT          = 128,
      ATTR_LOCAL              = 256,
      ATTR_HOLD_FOR_PICKUP    = 512,
      // 1024 is unused
      ATTR_FILE_REQ           = 2048,
      ATTR_RETURN_RECEIPT_REQ = 4096,
      ATTR_IS_RETURN_RECEIPT  = 8192,
      ATTR_AUDIT_REQ          = 16384,
      ATTR_FILE_UPDATE_REQ    = 32768,
      // flags
      FLAG_DIRECT             = 1,
      FLAG_IMMEDIATE          = 2,
      FLAG_KILL_FILE_SENT     = 4,
      FLAG_TRUNCATE_FILE      = 8,
      FLAG_LOCK               = 16,
      FLAG_CONFIRM_REQ        = 32,
      // BSO flags
      BSO_NORMAL              = 0,
      BSO_CRASH               = 1,
      BSO_DIRECT              = 2,
      BSO_HOLD                = 4,
      BSO_IMMEDIATE           = 8,
      BSO_KILL_FILE_SENT      = 16,
      BSO_TRUNCATE_FILE       = 32,
      BSO_IRNORE              = 64,
      BSO_TYPE                = 15,  // mask for BSO ?ut type
      BSO_LINE_TYPE           = 112; // mask for BSO ?lo line type
          
#pragma pack(push, 1)

struct SMsgHeader {
    char   FromUserName[36];
    char   ToUserName[36];
    char   Subject[72];
    char   DateTime[20];
    uint16 timesRead;
    addr   DestNode;
    addr   OrigNode;
    uint16 cost;
    addr   OrigNet;
    addr   DestNet;
    addr   DestZone;
    addr   OrigZone;
    addr   DestPoint;
    addr   OrigPoint;
    uint16 ReplyTo;
    uint16 Attrs;
    uint16 NextReply;
};

struct SPackedMsgHeader {
    uint16 MsgID; // 0002h
    addr   OrigNode;
    addr   DestNode;
    addr   OrigNet;
    addr   DestNet;
    uint16 Attrs;
    uint16 Cost;
    char   DateTime[20];
};

struct SPktHeader {
    addr   OrigNode;
    addr   DestNode;
    uint16 Year;
    uint16 Month;
    uint16 Day;
    uint16 Hour;
    uint16 Minute;
    uint16 Second;
    uint16 Baud;
    uint16 PacketType;
    addr   OrigNet;
    addr   DestNet;
    uint8  ProdCode;
    uint8  SerialNo;
    char   Password[8];
    addr   OrigZone;
    addr   DestZone;
    char   Fill[20];
};

#pragma pack(pop)

struct SFTNAddress {
    addr Zone;
    addr Net;
    addr Node;
    addr Point;
    char Domain[MAX_DOMAIN_SIZE];
};

int ParseFTNAddress(char *s, SFTNAddress &Address);
int AddrCmp( SFTNAddress &Addr1, SFTNAddress &Addr2 );

#endif
