#include "pch.h"
#include "CfgParse.hpp"
#include "Common.hpp"
#include "Misc.hpp"
#include "Lng.hpp"

CLng::CLng() {
    LinesNumber = 0;
    Lines = NULL;
}

CLng::~CLng() {
    Free();
}

void CLng::Free() {
    int i;
    
    if (LinesNumber > 0) {
        for (i = 0; i < LinesNumber; i++)
            delete [] Lines[i];
        delete [] Lines;
    }
    LinesNumber = 0;
    Lines = NULL;
}

int CLng::Read(char *LngFileName) {
    SStringList *SublinesList, *LastSubline;
    SStringList *LinesList, *LastLine;
    int          WaitingSubline, CurLen, Len;
    int          Number;
    int          ParserError;
    int          i;
    
    Free();
    
    CCfgParser LngParser( LngFileName );
    if (LngParser.Failed())
        return (0);
    LinesList = NULL;
    SublinesList = NULL;
    WaitingSubline = 0;
    Number = 0;
    ParserError = 0;
    while (!LngParser.Eof()) {
        LngParser.ReadLine();
        if (LngParser.Failed()) {
            if (!LngParser.Eof())
                ParserError = 1;
            break;
        }
        if (LngParser.Count() == 0)
            continue;
        // check syntax
        if (LngParser.Count() > 2 || LngParser[0][0] != '"' ||
            (LngParser.Count() == 2 && stricmp(LngParser[1], "\\") != 0))
        {
            ParserError = 1;
            break;
        }
        if (WaitingSubline) {
            LastSubline->Next = new SStringList;
            LastSubline = LastSubline->Next;
            LastSubline->Next = NULL;
            Len = strlen( LngParser[0] ) - 2;
            LastSubline->String = new char[Len + 1];
            StrNCpy( LastSubline->String, LngParser[0]+1, Len );
            CurLen += (Len + 1); // +1 because of "\r"
        }
        else {
            SublinesList = new SStringList;
            LastSubline = SublinesList;
            LastSubline->Next = NULL;
            Len = strlen( LngParser[0] ) - 2;
            LastSubline->String = new char[Len + 1];
            StrNCpy( LastSubline->String, LngParser[0]+1, Len );
            CurLen = Len;
        }
        if (LngParser.Count() == 2) {
            WaitingSubline = 1;
        }
        else {
            WaitingSubline = 0;
            if (LinesList == NULL) {
                LinesList = new SStringList;
                LastLine = LinesList;
            }
            else {
                LastLine->Next = new SStringList;
                LastLine = LastLine->Next;
            }
            LastLine->Next = NULL;
            LastLine->String = new char[CurLen + 1];
            LastLine->String[0] = '\0';
            while (SublinesList != NULL) {
                strcat( LastLine->String, SublinesList->String );
                LastSubline = SublinesList->Next;
                delete [] SublinesList->String;
                delete SublinesList;
                SublinesList = LastSubline;
                if (SublinesList != NULL)
                    strcat( LastLine->String, "\r" );
            }
            ConvertSoftCR( LastLine->String );
            Number++;
            CurLen = 0;
        }
    }
    if (ParserError || WaitingSubline) {
        // free memory
        while (LinesList != NULL) {
            LastLine = LinesList->Next;
            delete [] LinesList->String;
            delete LinesList;
            LinesList = LastLine;
        }
        while (SublinesList != NULL) {
            LastSubline = SublinesList->Next;
            delete [] SublinesList->String;
            delete SublinesList;
            SublinesList = LastSubline;
        }
        return (0);
    }
    if (Number == 0)
        return (0);
    Lines = new char*[Number];
    LinesNumber = Number;
    i = 0;
    while (LinesList != NULL) {
        Lines[i] = LinesList->String;
        i++;
        LastLine = LinesList->Next;
        delete LinesList;
        LinesList = LastLine;
    }
    assert( i == Number );
    return (1);
}

char* CLng::operator [](int i) {
    assert(i >= 0 && i < LinesNumber);
    return (Lines[i]);
}
