#include "pch.h"
#include <sys\types.h>
#include <direct.h>
#include <errno.h>
#include "Misc.hpp"

char *MonthName[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

void GetCurrentDateTime(char *s) {
    SYSTEMTIME  LocTime;

    GetLocalTime( &LocTime );

    sprintf( s, "%02hu %s %02hu  %02hu:%02hu:%02hu", LocTime.wDay,
             MonthName[LocTime.wMonth-1], LocTime.wYear %= 100, LocTime.wHour,
             LocTime.wMinute, LocTime.wSecond );
}

void GetCurrentDateTimeY2K(char *s) {
    SYSTEMTIME  LocTime;

    GetLocalTime( &LocTime );

    sprintf( s, "%02hu %s %04hu %02hu:%02hu:%02hu", LocTime.wDay,
             MonthName[LocTime.wMonth-1], LocTime.wYear, LocTime.wHour,
             LocTime.wMinute, LocTime.wSecond );
}

long GetFileNumber(char *s) {
    int  i;       // reading not more than 8 numbers
    long Number;

    i = 0;
    Number = 0;
    while (i < 8 && *s >= '0' && *s <= '9') {
        Number *= 10;
        Number += *s - '0';
        s++;
        i++;
    }
    if (*s != '.')
        Number = 0;

    return (Number);
}


long GetFileNumberHex(char *s) {
    int  i;       // reading not more than 8 numbers
    long Number;

    i = 0;
    Number = 0;
    while (i < 8 && isalnum(*s)) {
        Number <<= 4; // *16
        if (isdigit(*s)) {
            Number += *s - '0';
        }
        else {
            Number += toupper(*s) - 'A' + 10;
        }
        s++;
        i++;
    }
    if (*s != '.')
        Number = 0;

    return (Number);
}

int GetNumberOfEnclosedNames(char *s) {
    int Number;
    int i;

    if (*s == '\0')
        return (0);
    
    Number = 0;
    i = 0;
    while (s[i] != '\0') {
        if (s[i] == '\\')
            Number++;
        i++;
    }
    if (s[i-1] != '\\')
        Number++;

    return (Number);
}

int MakeDirStructure(char *s, int SkipNumber) {
    int i;
    static char Dir[MAX_PATH];
    int Success;

    // skips names
    i = 0;
    while (SkipNumber > 0) {
        while (s[i] != '\0' && s[i] != '\\') {
            Dir[i] = s[i];
            i++;
        }
        Dir[i] = s[i];
        i++;
        SkipNumber--;
    }
    // make structure
    Success = 1;
    while (s[i] != '\0') {
        while (s[i] != '\0' && s[i] != '\\') {
            Dir[i] = s[i];
            i++;
        }
        Dir[i] = s[i];
        if (s[i] != '\0') {
            i++;
            Dir[i] = '\0';
        }
        if (mkdir( Dir ) != 0) {
            Success &= (errno == EACCES);
        }
    }
    return (Success);
}

long GetFreeDirNumber(char *FindFileMask) {
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFindFile;
    BOOL            FileFound;
    long            MaxNumber, Number;
    
    MaxNumber = 0;
    hFindFile = FindFirstFile( FindFileMask, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    while (FileFound) {
        Number = GetFileNumber( FindFileData.cFileName );
        if (Number > MaxNumber)
            MaxNumber = Number;
        FileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );
    return (MaxNumber+1);
}

long GetFreeDirNumberHex(char *FindFileMask) {
    WIN32_FIND_DATA FindFileData;
    HANDLE          hFindFile;
    BOOL            FileFound;
    long            MaxNumber, Number;
    
    MaxNumber = 0;
    hFindFile = FindFirstFile( FindFileMask, &FindFileData );
    FileFound = (hFindFile != INVALID_HANDLE_VALUE);
    while (FileFound) {
        Number = GetFileNumberHex( FindFileData.cFileName );
        if (Number > MaxNumber)
            MaxNumber = Number;
        FileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );
    return (MaxNumber+1);
}

long int GetFileSize(FILE *f) {
    long int save_pos, size_of_file;

    save_pos = ftell( f );
    fseek( f, 0L, SEEK_END );
    size_of_file = ftell( f );
    fseek( f, save_pos, SEEK_SET );
    return( size_of_file );
    
}

int IsAlpha(char *s) {
    if (*s == '\0')
        return (0);
    while (*s != '\0') {
        if (!isalpha(*s))
            break;
        s++;
    }
    return (*s == '\0');
}

void RemoveBoundaryQuotes(char *s) {
    int len;
    int i;

    len = strlen(s);
    if (len < 2)
        return;
    len--;
    s[len] = '\0';
    for (i = 0; i < len; i++)
        s[i] = s[i+1];
}

int GetOnBit(unsigned long bitarray) {
    int i;
    unsigned long bit;

    i   = 0;
    bit = 1;
    while ((bitarray & bit) == 0 && i < 32) {
        bit <<= 1;
        i++;
    }
    return (i < 32 ? i : -1);
}

int CheckMask(char *s, char *mask) {
    while (*mask) {
        switch (*mask) {
        case '?':
            mask++;
            if (*s)
                s++;
            break;
        case '*':
            mask++;
            while (*s) {
                if (CheckMask(s, mask))
                    break;
                else
                    s++;
            }
            if (*s || CheckMask(s, mask))
                return (1);
            else
                return (0);
        default:
            if (toupper(*mask) != toupper(*s))
                return (0);
            mask++;
            s++;
        }
    }
    return (*s == '\0');
}

int IsMask(char *s) {
    while (*s != '\0') {
        if (*s == '*' || *s == '?')
            break;
        s++;
    }
    return (*s != '\0');
}


const MOVE_BUFFER_SIZE = 32768;
char  MoveBuffer[MOVE_BUFFER_SIZE];


int MoveFile(char *SourceFileName, char *DestFileName) {
    FILE   *sf;
    FILE   *df;
    size_t  BytesRead, BytesWritten;

    // check if on the same drive
    if (SourceFileName[0] == DestFileName[0]) {
        unlink( DestFileName );
        return (rename( SourceFileName, DestFileName ) == 0);
    }
    
    sf = fopen( SourceFileName, "rb" );
    if (sf == NULL)
        return (0);
    df = fopen( DestFileName, "wb" );
    if (df == NULL) {
        fclose( sf );
        return (0);
    }
    do {
        BytesRead = fread( MoveBuffer, 1, MOVE_BUFFER_SIZE, sf );
        if (BytesRead > 0)
            BytesWritten = fwrite( MoveBuffer, 1, BytesRead, df );
    } while (BytesRead > 0 && BytesWritten == BytesRead);
    fclose( sf );
    fclose( df );
    if (BytesRead == 0) {
        if (unlink( SourceFileName ) != 0)
            cout << "Can't delete file " << SourceFileName << endl;
        return (1);
    }
    unlink( DestFileName );
    return (0);
}

int CopyFile(char *SourceFileName, char *DestFileName) {
    FILE   *sf;
    FILE   *df;
    size_t  BytesRead, BytesWritten;

    sf = fopen( SourceFileName, "rb" );
    if (sf == NULL)
        return (0);
    df = fopen( DestFileName, "wb" );
    if (df == NULL) {
        fclose( sf );
        return (0);
    }
    do {
        BytesRead = fread( MoveBuffer, 1, MOVE_BUFFER_SIZE, sf );
        if (BytesRead > 0)
            BytesWritten = fwrite( MoveBuffer, 1, BytesRead, df );
    } while (BytesRead > 0 && BytesWritten == BytesRead);
    fclose( sf );
    fclose( df );
    if (BytesRead == 0)
        return (1);
    unlink( DestFileName );
    return (0);
}

void ConvertNameToShort(char *Src,char *Dest) {
    char ShortName[MAX_PATH];
    char AnsiName[MAX_PATH];
    BOOL ApisWereANSI;

    ApisWereANSI = AreFileApisANSI();
    if (!ApisWereANSI)
        SetFileApisToANSI();
    OemToChar( Src, AnsiName );
    if (GetShortPathName( AnsiName, ShortName, sizeof(ShortName) ))
        CharToOem( ShortName, Dest );
    else
        strcpy( Dest, Src );
    if (!ApisWereANSI)
        SetFileApisToOEM();
}

void StrNCpy(char *Dest, const char *Source, int Number) {
    while (Number > 0 && *Source != '\0') {
        *Dest = *Source;
        Source++;
        Dest++;
        Number--;
    }
    *Dest = '\0';
}

void ConvertSoftCR(char *String) {
    while (*String != '\0') {
        if (*String == 0x8D)
            *String = 0x48;
        String++;
    }
}

void FormatSize(int size, char *buf) {
    int i, j, len, count;

    sprintf( buf, "%ld", size );
    len = strlen(buf);

    count = 0; 
    for (i = len-1; i >= 0; i--) {
        count++;
        if ((count == 3) && (i > 0)) { // insert ','
            count = 0;
            // shift reminder left
            for (j = len+1; j > i; j--)
                buf[j] = buf[j-1];
            len++;
            buf[i] = ',';
        }
    }
}

int CreateFileFlag(char *FlagName) {
    FILE *f;

    f = fopen( FlagName, "rb" );
    if (f != NULL) {
        // file already exist
        fclose(f);
        return (1);
    }
    f = fopen( FlagName, "wb" );
    if (f == NULL) {
        // failed
        return (0);
    }
    // success
    fclose(f);
    return (1);
}

BOOL parseIntWithCommas(char* str, DWORD& d) {
	d = 0;

	while (*str != '\0') {
		if (*str == ',') {
			str++;
			continue;
		}
		if (!isdigit(*str))
			return FALSE;
		d *= 10;
		d += *str - '0';
		str++;
	}
	return TRUE;
}

void	removeBoundarySpaces(char* source, char* dest) {
	while (*source == ' ')
		source++;

	int	len = strlen(source);
	int	i   = len-1;
	while (len > 0 && source[i] == ' ') {
		len--;
		i--;
	}

	if (len > 0)
		memcpy( dest, source, len );
	dest[len] = '\0';
}

int	arrMonthDays[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

int GetDays(WORD wDay, WORD wMonth, WORD wYear) {
	int	nDays = wYear*365;
	// consider previous leap-years
	nDays += (wYear+3) / 4;
	for (int i = 0; i < wMonth-1; i++) {
		nDays += arrMonthDays[i];
	}
	// check leap-year
	if ((wYear % 4) == 0 && wMonth > 2)
		nDays++;
	nDays += wDay;
	return (nDays);
}
