#include "pch.h"
#include "Mem.hpp"
#include "Misc.hpp"
#include "Temp.hpp"
#include "Common.hpp"
#include "Msg.hpp"


CMsg::CMsg() : MsgIdHandler() {
    char FindFileMask[MAX_PATH];
    
    memset( &Header, 0, sizeof(Header) );
    *Area      = '\0';
    *MsgIdLine = '\0';
    Attrs = 0;

    strcpy( FindFileMask, GlobalVars.NetmailPath );
    strcat( FindFileMask, "*.msg" );
    FreeMsgNumber = GetFreeDirNumber( FindFileMask );

    f = NULL;
}


CMsg::~CMsg() {
	if (f)
		fclose(f);
}


static char	userNameWithoutBoundarySpaces[36];

int CMsg::Read(char *MsgFName) {
    size_t        BytesRead;
    int           BufPos, CurPos;
    int           i, j;
    static char   Kludge[MAX_KLUDGE_SIZE];
    int           MiddleOfLine;
    int           MiddleOfKludge;
    SMsgDataList *LastMsgData;
    char          KludgeType; // 0x01, '%', '+', '-'
    int           ToProcessor;

    
    *Area        = '\0';
    Flags        = 0;
    FlagsExist   = 0;
    *FlagsKludge = '\0';
    FMPTExist    = 0;
    TOPTExist    = 0;
    INTLExist    = 0;
    DOMAINExist  = 0;
    MsgIdExist   = 0;
    MsgIdDecoded = 0;

    GlobalVars.MsgData = NULL;
    
    // keep message file name and message directory
    strcpy( MsgFileName, MsgFName );
    j = -1;
    for (i = 0; MsgFName[i] != '\0'; i++) {
        if (MsgFName[i] == '\\')
            j = i;
    }
    j++;
    memcpy( MsgFileDir, MsgFName, j );
    MsgFileDir[j] = '\0';

    // open file
    f = fopen( MsgFileName, "r+b" );
    if (!f)
        return (0);

    // read header
    BytesRead = fread( &Header, 1, sizeof(Header), f );
    if (BytesRead != sizeof(Header)) {
        fclose(f);
	f = NULL;
        return (0);
    }

    Attrs = Header.Attrs;

    // check attributes
    if (Attrs & ATTR_RECEIVED) {
        fclose(f);
	f = NULL;
        return (0);
    }

    // check addressee
    Header.ToUserName[35] = '\0';
    removeBoundarySpaces( Header.ToUserName, userNameWithoutBoundarySpaces );
    for (i = 0; i < GlobalVars.AliasesNumber; i++) {
        if (stricmp( userNameWithoutBoundarySpaces, GlobalVars.Aliases[i] ) == 0) {
            strcpy( Header.ToUserName, GlobalVars.Aliases[i] );
            break;
        }
    }
    if (i >= GlobalVars.AliasesNumber) {
        fclose(f);
	f = NULL;
        return (0);
    }
    
    BufPos = 0;
    MiddleOfLine = 0;
    MiddleOfKludge = 0;
    BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize, f );
    MsgInBuf = 1;
    BodySize = BytesRead;
    if (BytesRead > 5) {
        if (strncmp( GlobalMsgBuf, "AREA:", 5 ) == 0) {
            BufPos += 5;
            i = 0;
            while (BufPos < BytesRead && i < (MAX_AREA_SIZE-1)
                   && GlobalMsgBuf[BufPos] != 0x0D)
            {
                BufPos++;
                i++;
            }
            if (i > 0) {
                memcpy( Area, GlobalMsgBuf, i );
                Area[i] = '\0';
            }
            while (BufPos < BytesRead && GlobalMsgBuf[BufPos] != 0x0D)
                BufPos++;
            if (BufPos < BytesRead) // 0x0D
                BufPos++;
            else
                MiddleOfLine = 1;
        }
    }
    while (BytesRead > 0) {
        while (BufPos < BytesRead) {
            if (MiddleOfKludge) {
                j = i;
                CurPos = BufPos;
                while (BufPos < BytesRead && i < (MAX_KLUDGE_SIZE-1)
                       && GlobalMsgBuf[BufPos] != 0x0D)
                {
                    BufPos++;
                    i++;
                }
                if (i > j)
                    memcpy( Kludge+j, GlobalMsgBuf+CurPos, i-j );
                Kludge[i] = '\0';
                if (!(BufPos == BytesRead && i < (MAX_KLUDGE_SIZE-1))) {
                    MiddleOfKludge = 0;
                    switch (KludgeType) {
                    case 0x01:
                        // check kludges
                        if (!FMPTExist && strncmp( Kludge, "FMPT ", 5) == 0) {
                            if (ParsePT( Kludge+5, FMPT ))
                                FMPTExist = 1;
                        }
                        else if (!TOPTExist && strncmp( Kludge, "TOPT ", 5) == 0) {
                            if (ParsePT( Kludge+5, TOPT ))
                                TOPTExist = 1;
                        }
                        else if (!INTLExist && strncmp( Kludge, "INTL ", 5) == 0) {
                            if (ParseINTL( Kludge+5 ))
                                INTLExist = 1;
                        }
                        else if (!FlagsExist && strncmp( Kludge, "FLAGS ", 6) == 0) {
                            if (ParseFlags( Kludge+6 ))
                                FlagsExist = 1;
                        }
                        else if (!DOMAINExist &&
                                 strncmp( Kludge, "DOMAIN ", 7) == 0)
                        {
                            if (ParseDOMAIN( Kludge+7 ))
                                DOMAINExist = 1;
                        }
                        else if (!MsgIdExist &&
                                 strncmp( Kludge, "MSGID: ", 7) == 0)
                        {
                            MsgIdExist = 1;
                            if (ParseMsgId( Kludge+7 ))
                                MsgIdDecoded = 1;
                        }
                        break;
                    case '%':
                        j = 0;
                        while (Kludge[j] != '\0' && Kludge[j] != ' ')
                            j++;
                        if (j == 0)
                            break;
                        
                        if (GlobalVars.MsgData == NULL) {
                            GlobalVars.MsgData = new SMsgDataList;
                            LastMsgData = GlobalVars.MsgData;
                        }
                        else {
                            LastMsgData->Next = new SMsgDataList;
                            LastMsgData = LastMsgData->Next;
                        }
                        LastMsgData->Next = NULL;
                        LastMsgData->DataType = MSGDATA_COMMAND;
                        LastMsgData->Data = new char[j + 1];
                        memcpy( LastMsgData->Data, Kludge, j);
                        LastMsgData->Data[j] = '\0';
                        break;
                    case '+':
                        j = 0;
                        while (Kludge[j] != '\0' && Kludge[j] != ' ')
                            j++;
                        if (j == 0)
                            break;
                        
                        if (GlobalVars.MsgData == NULL) {
                            GlobalVars.MsgData = new SMsgDataList;
                            LastMsgData = GlobalVars.MsgData;
                        }
                        else {
                            LastMsgData->Next = new SMsgDataList;
                            LastMsgData = LastMsgData->Next;
                        }
                        LastMsgData->Next = NULL;
                        LastMsgData->DataType = MSGDATA_SUBSCRIBE;
                        LastMsgData->Data = new char[j + 1];
                        memcpy( LastMsgData->Data, Kludge, j);
                        LastMsgData->Data[j] = '\0';
                        
                        break;
                    case '-':
                        // skip tearline
                        if (Kludge[0] == '-' && Kludge[1] == '-')
                            break;
                        j = 0;
                        while (Kludge[j] != '\0' && Kludge[j] != ' ')
                            j++;
                        if (j == 0)
                            break;
                        
                        if (GlobalVars.MsgData == NULL) {
                            GlobalVars.MsgData = new SMsgDataList;
                            LastMsgData = GlobalVars.MsgData;
                        }
                        else {
                            LastMsgData->Next = new SMsgDataList;
                            LastMsgData = LastMsgData->Next;
                        }
                        LastMsgData->Next = NULL;
                        LastMsgData->DataType = MSGDATA_UNSUBSCRIBE;
                        LastMsgData->Data = new char[j + 1];
                        memcpy( LastMsgData->Data, Kludge, j);
                        LastMsgData->Data[j] = '\0';
                        
                        break;
                    }
                }
            }
            else if (MiddleOfLine) {
                while (BufPos < BytesRead && GlobalMsgBuf[BufPos] != 0x0D)
                    BufPos++;
                if (BufPos < BytesRead) { // 0x0D
                    BufPos++;
                    if (BufPos < BytesRead && GlobalMsgBuf[BufPos] == 0x0A)
                        BufPos++;
                    MiddleOfLine = 0;
                }
            }
            else {
                KludgeType = GlobalMsgBuf[BufPos];
                if (KludgeType == 0x01 || KludgeType == '%' ||
                    KludgeType == '+'  || KludgeType == '-')
                {
                    BufPos++;
                    i = 0;
                    CurPos = BufPos;
                    while (BufPos < BytesRead && i < (MAX_KLUDGE_SIZE-1)
                           && GlobalMsgBuf[BufPos] != 0x0D)
                    {
                        BufPos++;
                        i++;
                    }
                    if (i > 0 || BufPos == BytesRead) {
                        if (i > 0)
                            memcpy( Kludge, GlobalMsgBuf+CurPos, i );
                        Kludge[i] = '\0';
                        if (BufPos == BytesRead && i < (MAX_KLUDGE_SIZE-1))
                            MiddleOfKludge = 1;
                        else {
                            switch (KludgeType) {
                            case 0x01:
                                // check kludges
                                if (!FMPTExist && strncmp( Kludge, "FMPT ", 5) == 0) {
                                    if (ParsePT( Kludge+5, FMPT ))
                                        FMPTExist = 1;
                                }
                                else if (!TOPTExist && strncmp( Kludge, "TOPT ", 5) == 0) {
                                    if (ParsePT( Kludge+5, TOPT ))
                                        TOPTExist = 1;
                                }
                                else if (!INTLExist && strncmp( Kludge, "INTL ", 5) == 0) {
                                    if (ParseINTL( Kludge+5 ))
                                        INTLExist = 1;
                                }
                                else if (!FlagsExist && strncmp( Kludge, "FLAGS ", 6) == 0) {
                                    if (ParseFlags( Kludge+6 ))
                                        FlagsExist = 1;
                                }
                                else if (!DOMAINExist &&
                                         strncmp( Kludge, "DOMAIN ", 7) == 0)
                                {
                                    if (ParseDOMAIN( Kludge+7 ))
                                        DOMAINExist = 1;
                                }
                                else if (!MsgIdExist &&
                                         strncmp( Kludge, "MSGID: ", 7) == 0)
                                {
                                    MsgIdExist = 1;
                                    if (ParseMsgId( Kludge+7 ))
                                        MsgIdDecoded = 1;
                                }
                                break;
                            case '%':
                                j = 0;
                                while (Kludge[j] != '\0' && Kludge[j] != ' ')
                                    j++;
                                if (j == 0)
                                    break;
                                
                                if (GlobalVars.MsgData == NULL) {
                                    GlobalVars.MsgData = new SMsgDataList;
                                    LastMsgData = GlobalVars.MsgData;
                                }
                                else {
                                    LastMsgData->Next = new SMsgDataList;
                                    LastMsgData = LastMsgData->Next;
                                }
                                LastMsgData->Next = NULL;
                                LastMsgData->DataType = MSGDATA_COMMAND;
                                LastMsgData->Data = new char[j + 1];
                                memcpy( LastMsgData->Data, Kludge, j);
                                LastMsgData->Data[j] = '\0';
                                
                                break;
                            case '+':
                                j = 0;
                                while (Kludge[j] != '\0' && Kludge[j] != ' ')
                                    j++;
                                if (j == 0)
                                    break;
                        
                                if (GlobalVars.MsgData == NULL) {
                                    GlobalVars.MsgData = new SMsgDataList;
                                    LastMsgData = GlobalVars.MsgData;
                                }
                                else {
                                    LastMsgData->Next = new SMsgDataList;
                                    LastMsgData = LastMsgData->Next;
                                }
                                LastMsgData->Next = NULL;
                                LastMsgData->DataType = MSGDATA_SUBSCRIBE;
                                LastMsgData->Data = new char[j + 1];
                                memcpy( LastMsgData->Data, Kludge, j);
                                LastMsgData->Data[j] = '\0';
                        
                                break;
                            case '-':
                                // skip tearline
                                if (Kludge[0] == '-' && Kludge[1] == '-')
                                    break;
                                j = 0;
                                while (Kludge[j] != '\0' && Kludge[j] != ' ')
                                    j++;
                                if (j == 0)
                                    break;
                                
                                if (GlobalVars.MsgData == NULL) {
                                    GlobalVars.MsgData = new SMsgDataList;
                                    LastMsgData = GlobalVars.MsgData;
                                }
                                else {
                                    LastMsgData->Next = new SMsgDataList;
                                    LastMsgData = LastMsgData->Next;
                                }
                                LastMsgData->Next = NULL;
                                LastMsgData->DataType = MSGDATA_UNSUBSCRIBE;
                                LastMsgData->Data = new char[j + 1];
                                memcpy( LastMsgData->Data, Kludge, j);
                                LastMsgData->Data[j] = '\0';
                                
                                break;
                            }
                        }
                    }
                }
                while (BufPos < BytesRead && GlobalMsgBuf[BufPos] != 0x0D)
                    BufPos++;
                if (BufPos < BytesRead) { // 0x0D
                    BufPos++;
                    if (BufPos < BytesRead && GlobalMsgBuf[BufPos] == 0x0A)
                        BufPos++;
                }
                else {
                    if (!MiddleOfKludge)
                        MiddleOfLine = 1;
                }
            }
        }
        BufPos = 0;
        BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize, f );
        if (BytesRead > 0)
            MsgInBuf = 0;
        BodySize += BytesRead;
    }
    
    if (MiddleOfKludge) {
        switch (KludgeType) {
        case 0x01:
            // check kludges
            if (!FMPTExist && strncmp( Kludge, "FMPT ", 5) == 0) {
                if (ParsePT( Kludge+5, FMPT ))
                    FMPTExist = 1;
            }
            else if (!TOPTExist && strncmp( Kludge, "TOPT ", 5) == 0) {
                if (ParsePT( Kludge+5, TOPT ))
                    TOPTExist = 1;
            }
            else if (!INTLExist && strncmp( Kludge, "INTL ", 5) == 0) {
                if (ParseINTL( Kludge+5 ))
                    INTLExist = 1;
            }
            else if (!FlagsExist && strncmp( Kludge, "FLAGS ", 6) == 0) {
                if (ParseFlags( Kludge+6 ))
                    FlagsExist = 1;
            }
            else if (!DOMAINExist &&
                     strncmp( Kludge, "DOMAIN ", 7) == 0)
            {
                if (ParseDOMAIN( Kludge+7 ))
                    DOMAINExist = 1;
            }
            else if (!MsgIdExist &&
                     strncmp( Kludge, "MSGID: ", 7) == 0)
            {
                MsgIdExist = 1;
                if (ParseMsgId( Kludge+7 ))
                    MsgIdDecoded = 1;
            }
            break;
        case '%':
            j = 0;
            while (Kludge[j] != '\0' && Kludge[j] != ' ')
                j++;
            if (j == 0)
                break;
            
            if (GlobalVars.MsgData == NULL) {
                GlobalVars.MsgData = new SMsgDataList;
                LastMsgData = GlobalVars.MsgData;
            }
            else {
                LastMsgData->Next = new SMsgDataList;
                LastMsgData = LastMsgData->Next;
            }
            LastMsgData->Next = NULL;
            LastMsgData->DataType = MSGDATA_COMMAND;
            LastMsgData->Data = new char[j + 1];
            memcpy( LastMsgData->Data, Kludge, j);
            LastMsgData->Data[j] = '\0';
            
            break;
        case '+':
            j = 0;
            while (Kludge[j] != '\0' && Kludge[j] != ' ')
                j++;
            if (j == 0)
                break;
            
            if (GlobalVars.MsgData == NULL) {
                GlobalVars.MsgData = new SMsgDataList;
                LastMsgData = GlobalVars.MsgData;
            }
            else {
                LastMsgData->Next = new SMsgDataList;
                LastMsgData = LastMsgData->Next;
            }
            LastMsgData->Next = NULL;
            LastMsgData->DataType = MSGDATA_SUBSCRIBE;
            LastMsgData->Data = new char[j + 1];
            memcpy( LastMsgData->Data, Kludge, j);
            LastMsgData->Data[j] = '\0';
            
            break;
        case '-':
            // skip tearline
            if (Kludge[0] == '-' && Kludge[1] == '-')
                break;
            j = 0;
            while (Kludge[j] != '\0' && Kludge[j] != ' ')
                j++;
            if (j == 0)
                break;

            if (GlobalVars.MsgData == NULL) {
                GlobalVars.MsgData = new SMsgDataList;
                LastMsgData = GlobalVars.MsgData;
            }
            else {
                LastMsgData->Next = new SMsgDataList;
                LastMsgData = LastMsgData->Next;
            }
            LastMsgData->Next = NULL;
            LastMsgData->DataType = MSGDATA_UNSUBSCRIBE;
            LastMsgData->Data = new char[j + 1];
            memcpy( LastMsgData->Data, Kludge, j);
            LastMsgData->Data[j] = '\0';
            
            break;
        }
    }
    
    DecodeAddresses();

    // check if message to our address or not
    ToProcessor = 0;
    if (ToAddr( GlobalVars.MainAddr ))
        ToProcessor = 1;
    for (i = 0; !ToProcessor && i < GlobalVars.AKAAddrNumber; i++) {
        if (ToAddr( GlobalVars.AKAAddr[i] ))
            ToProcessor = 1;
    }
    if (!ToProcessor) {
        GlobalVars.FreeMsgData();
        fclose(f);
	f = NULL;
        return (0);
    }
    /*
    cout << "Msg from " << From.Zone << ":" << From.Net << "/" << From.Node
         << "." << From.Point << " to " << To.Zone << ":" << To.Net << "/"
         << To.Node << "." << To.Point << endl;
    */
    return (1);
}

void CMsg::DecodeAddresses() {
    int FromZoneDefined;
    int ToZoneDefined;

    ZoneGate        = 0;
    
    if (DOMAINExist) {
        memcpy( &From, &DomainKludgeFrom, sizeof(From) );
        memcpy( &To,   &DomainKludgeTo, sizeof(To) );
        return;
    }
    
    FromZoneDefined = 0;
    ToZoneDefined   = 0;
    
    From.Net  = Header.OrigNet;
    From.Node = Header.OrigNode;
    if (FMPTExist)
        From.Point = FMPT;
    else
        From.Point = 0;
    
    To.Net  = Header.DestNet;
    To.Node = Header.DestNode;
    if (TOPTExist)
        To.Point = TOPT;
    else
        To.Point = 0;

    if (INTLExist) {
        From.Zone = INTLFrom.Zone;
        FromZoneDefined = 1;
        From.Net  = INTLFrom.Net;
        From.Node = INTLFrom.Node;
        if (To.Net == INTLTo.Net && To.Node == INTLTo.Node) {
            To.Zone = INTLTo.Zone;
            ToZoneDefined = 1;
        }
        else {
            ZoneGate = 1;
            RealTo.Zone   = INTLTo.Zone;
            RealTo.Net    = INTLTo.Net;
            RealTo.Node   = INTLTo.Node;
            RealTo.Point  = To.Point;
            To.Zone       = From.Zone;
            ToZoneDefined = 1;
            To.Point      = 0;
        }
    }
    else {
        if (MsgIdDecoded)
            From.Zone = MsgIdFrom.Zone;
    }
    
    if (!FromZoneDefined)
        From.Zone = GlobalVars.MainAddr.Zone;
    
    if (!ToZoneDefined)
        To.Zone = From.Zone;

    if (MsgIdDecoded)
        strcpy( From.Domain, MsgIdFrom.Domain );
    else
        strcpy( From.Domain, GlobalVars.MainAddr.Domain );

    strcpy( To.Domain, From.Domain );
}

int CMsg::ParsePT(char *s, addr &PT) {
    int i; // not more then 5 symbols
    
    while (*s == ' ')
        s++;
    
    #ifdef SIGNED_ADDR
    int sign;

    if (*s == '-') {
        sign = -1;
        s++;
    }
    else
        sign = 1;
    #endif
    
    PT = 0;

    i = 0;
    while (i < 5 && *s >= '0' && *s <= '9') {
        PT *= 10;
        PT += (*s - '0');
        s++;
        i++;
    }
    
    while (*s == ' ')
        s++;
    
    if (*s != '\0')
        return (0);
    
    #ifdef SIGNED_ADDR
    PT *= sign;
    #endif
    
    return (1);
}

int CMsg::ParseINTL(char *s) {
    static char AddrStr[MAX_KLUDGE_SIZE];
    int i;
    
    while (*s == ' ')
        s++;

    i = 0;
    while (i < (MAX_KLUDGE_SIZE-1) && *s != '\0' && *s != ' ') {
        AddrStr[i] = *s;
        i++;
        s++;
    }
    if (i == 0)
        return (0);
    if (*s != ' ')
        return (0);
    
    AddrStr[i] = '\0';
    if (!ParseFTNAddress( AddrStr, INTLTo ))
        return (0);
    
    while (*s == ' ')
        s++;

    i = 0;
    while (i < (MAX_KLUDGE_SIZE-1) && *s != '\0' && *s != ' ') {
        AddrStr[i] = *s;
        i++;
        s++;
    }
    if (i == 0)
        return (0);
    AddrStr[i] = '\0';
    
    while (*s == ' ')
        s++;
    if (*s != '\0')
        return(0);
    
    return (ParseFTNAddress( AddrStr, INTLFrom ));
}

int CMsg::ParseMsgId(char *s) {
    static char Str[MAX_KLUDGE_SIZE];
    int         i;
    
    while (*s == ' ')
        s++;
    
    if (*s == '\0')
        *MsgIdLine = '\0';
    else
        strcpy( MsgIdLine, s );
    
    i = 0;
    while (*s != ' ' && *s != '\0') {
        s++;
        i++;
    }
    if (i == 0)
        return (0);
    Str[i] = '\0';
    return (ParseFTNAddress( Str, MsgIdFrom ));
}

int CMsg::ParseFTNAddress(char *s, SFTNAddress &Address) {
    // parse Zone:Net/Node
    //       Zone:Net/Node.Point
    //       Zone:Net/Node.Point@Domain
    int i; // Not more then 5 symbols on number
    #ifdef SIGNED_ADDR
    int sign;
    #endif

    while (*s == ' ')
        s++;
    
    // parse zone

    #ifdef SIGNED_ADDR
    if (*s == '-') {
        sign = -1;
        s++;
    }
    else
        sign = 1;
    #endif

    Address.Zone = 0;

    i = 0;
    while (i < 5 && *s >= '0' && *s <= '9') {
        Address.Zone *= 10;
        Address.Zone += (*s - '0');
        s++;
        i++;
    }
    if (i == 0 || *s != ':')
        return (0);
    
    #ifdef SIGNED_ADDR
    Address.Zone *= sign;
    #endif
    
    s++;

    // parse net
    
    #ifdef SIGNED_ADDR
    if (*s == '-') {
        sign = -1;
        s++;
    }
    else
        sign = 1;
    #endif

    Address.Net = 0;

    i = 0;
    while (i < 5 && *s >= '0' && *s <= '9') {
        Address.Net *= 10;
        Address.Net += (*s - '0');
        s++;
        i++;
    }
    if (i == 0 || *s != '/')
        return (0);
    
    #ifdef SIGNED_ADDR
    Address.Net *= sign;
    #endif
    
    s++;
    
    // parse node

    #ifdef SIGNED_ADDR
    if (*s == '-') {
        sign = -1;
        s++;
    }
    else
        sign = 1;
    #endif

    Address.Node = 0;

    i = 0;
    while (i < 5 && *s >= '0' && *s <= '9') {
        Address.Node *= 10;
        Address.Node += (*s - '0');
        s++;
        i++;
    }
    if (i == 0)
        return (0);
    
    #ifdef SIGNED_ADDR
    Address.Node *= sign;
    #endif

    Address.Point = 0;

    if (*s == '.') {  // parse point
        s++;
        #ifdef SIGNED_ADDR
        if (*s == '-') {
            sign = -1;
            s++;
        }
        else
            sign = 1;
        #endif
        i = 0;
        while (i < 5 && *s >= '0' && *s <= '9') {
            Address.Point *= 10;
            Address.Point += (*s - '0');
            s++;
            i++;
        }
        if (i == 0)
            return (0);
        #ifdef SIGNED_ADDR
        Address.Point *= sign;
        #endif
    }

    Address.Domain[0] = '\0';
    if (*s == '@') { // parse domain
        s++;
        i = 0;
        while (i < (MAX_DOMAIN_SIZE-1) && *s != ' ' && *s != '\0') {
            Address.Domain[i] = *s;
            s++;
            i++;
        }
        if (i == 0 || (*s != ' ' && *s != '\0'))
            return (0);
        Address.Domain[i] = '\0';
        #ifdef SUBSTITUTE_DOMAIN
        if (stricmp(Address.Domain, "Fidonet.org") == 0)
            strcpy( Address.Domain, "Fidonet" );
        #endif
    }

    while (*s == ' ')
        s++;

    return (*s == '\0');
}

void CMsg::Save() {
    static char TempFileName[MAX_PATH];
    int         SuccessWrite;
    size_t      BytesRead, BytesWritten;
    FILE       *tmp;

    Header.Attrs = Attrs;
    
    strcpy( TempFileName, "TempDir" );
    strcat( TempFileName, "Tracker.tmp" );
    tmp = GetTempFile(TempFileName, "wb" );
    if (tmp) {
        SuccessWrite = SaveMsg( tmp );
        if (SuccessWrite) {
            // compare letters of disks
            if (*MsgFileName == *TempFileName) {
                fclose(f);
		f = NULL;
                remove( MsgFileName );
                rename( TempFileName, MsgFileName );
            }
            else {
                rewind(f);
                tmp = fopen( TempFileName, "rb" );
                BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                   tmp );
                while (SuccessWrite && BytesRead > 0) {
                    BytesWritten = fwrite( GlobalMsgBuf, 1, BytesRead, f );
                    SuccessWrite = (BytesWritten == BytesRead);
                    BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                       tmp );
                }
                fclose(tmp);
                remove( TempFileName );
            }
        }
    }

    if (f) {
        fclose(f);
	f = NULL;
    }
}

int CMsg::ParseDOMAIN(char *s) {
    static char Domain[MAX_DOMAIN_SIZE];
    static char Str[MAX_KLUDGE_SIZE];
    int i;
    
    while (*s == ' ')
        s++;
    
    i = 0;
    while (i < (MAX_DOMAIN_SIZE-1) && *s != ' ' && *s != '\0') {
        Domain[i] = *s;
        i++;
        s++;
    }
    if (i == 0)
        return (0);
    
    Domain[i] = '\0';
    
    while (*s == ' ')
        s++;
    
    i = 0;
    while (i < (MAX_KLUDGE_SIZE-1) && *s != ' ' && *s != '\0') {
        Str[i] = *s;
        i++;
        s++;
    }
    Str[i] = '\0';
    
    if (!ParseFTNAddress( Str, DomainKludgeTo ))
        return (0);
    if (*DomainKludgeTo.Domain != '\0')
        return (0);
    strcpy( DomainKludgeTo.Domain, Domain );
    
    while (*s == ' ')
        s++;
    
    i = 0;
    while (i < (MAX_DOMAIN_SIZE-1) && *s != ' ' && *s != '\0') {
        Domain[i] = *s;
        i++;
        s++;
    }
    if (i == 0)
        return (0);
    
    Domain[i] = '\0';
    
    if (!ParseFTNAddress( s, DomainKludgeFrom ))
        return (0);
    if (*DomainKludgeFrom.Domain != '\0')
        return (0);
    strcpy( DomainKludgeFrom.Domain, Domain );

    #ifdef SUBSTITUTE_DOMAIN
    if (stricmp(DomainKludgeFrom.Domain, "Fidonet.org") == 0)
        strcpy( DomainKludgeFrom.Domain, "Fidonet" );
    if (stricmp(DomainKludgeTo.Domain, "Fidonet.org") == 0)
        strcpy( DomainKludgeTo.Domain, "Fidonet" );
    #endif
    return (1);
}

int CMsg::ParseFlags(char *s) {
    char FlagStr[MAX_FLAG_SIZE];
    int  ParsedFlags;
    int  i;

    ParsedFlags = 0;
    *FlagsKludge = '\0';
    while (*s == ' ')
        s++;
    while (*s) {
        i = 0;
        while (i < (MAX_FLAG_SIZE-1) && *s != '\0' && *s != ' ') {
            FlagStr[i] = *s;
            i++;
            s++;
        }
        if (*s != '\0' && *s != ' ')
            break;
        FlagStr[i] = '\0';
        // check known flags
        if (strcmp( FlagStr, "PVT" ) == 0) {
            Attrs |= ATTR_PRIVATE;
        }
        else if (strcmp( FlagStr, "HLD" ) == 0) {
            Attrs |= ATTR_HOLD_FOR_PICKUP;
        }
        else if (strcmp( FlagStr, "CRA" ) == 0) {
            Attrs |= ATTR_CRASH;
        }
        else if (strcmp( FlagStr, "K/S" ) == 0) {
            Attrs |= ATTR_KILL_SENT;
        }
        else if (strcmp( FlagStr, "SNT" ) == 0) {
            Attrs |= ATTR_SENT;
        }
        else if (strcmp( FlagStr, "RCV" ) == 0) {
            Attrs |= ATTR_RECEIVED;
        }
        else if (strcmp( FlagStr, "FIL" ) == 0) {
            Attrs |= ATTR_FILE_ATTACHED;
        }
        else if (strcmp( FlagStr, "FRQ" ) == 0) {
            Attrs |= ATTR_FILE_REQ;
        }
        else if (strcmp( FlagStr, "RRQ" ) == 0) {
            Attrs |= ATTR_RETURN_RECEIPT_REQ;
        }
        else if (strcmp( FlagStr, "DIR" ) == 0) {
            Flags |= FLAG_DIRECT;
        }
        else if (strcmp( FlagStr, "IMM" ) == 0) {
            Flags |= FLAG_IMMEDIATE;
        }
        else if (strcmp( FlagStr, "KFS" ) == 0) {
            Flags |= FLAG_KILL_FILE_SENT;
        }
        else if (strcmp( FlagStr, "TFS" ) == 0) {
            Flags |= FLAG_TRUNCATE_FILE;
        }
        else if (strcmp( FlagStr, "LOK" ) == 0) {
            Flags |= FLAG_LOCK;
        }
        else if (strcmp( FlagStr, "CFM" ) == 0) {
            Flags |= FLAG_CONFIRM_REQ;
        }
        else {
            strcat( FlagsKludge, " " );
            strcat( FlagsKludge, FlagStr );
        }
        
        ParsedFlags++;
        while (*s == ' ')
            s++;
    }

    if ((Attrs & ATTR_FILE_ATTACHED) == 0 &&
        (Flags & (FLAG_KILL_FILE_SENT | FLAG_TRUNCATE_FILE) != 0))
    {
        Flags &= ~(FLAG_KILL_FILE_SENT | FLAG_TRUNCATE_FILE);
    }
    
    return (ParsedFlags > 0);
}

void CMsg::SetFlag(char *s) {
    strcat( FlagsKludge, " " );
    strcat( FlagsKludge, s );
}

int CMsg::CheckFlag(char *s) {
    int   len;
    char *fl;

    len = strlen(s);
    fl = FlagsKludge;
    while (*fl == ' ')
        fl++;
    while (*fl) {
        if (strncmp(fl, s, len) == 0) {
            if (fl[len] == '\0' || fl[len] == ' ')
                break;
        }
        while (*fl != ' ' && *fl != '\0')
            fl++;
        while (*fl == ' ')
            fl++;
    }

    return (*fl != '\0');
}

void CMsg::ClearFlag(char *s) {
    int   len, fl_len;
    int   i;
    char *fl;

    len = strlen(s);
    fl = FlagsKludge;
    fl_len = strlen(fl);
    while (*fl == ' ') {
        fl++;
        fl_len--;
    }
    while (fl_len >= len) {
        if (strncmp(fl, s, len) == 0) {
            if (fl[len] == '\0' || fl[len] == ' ') {
                fl_len -= len;
                for (i = 0; i < fl_len; i++)
                    fl[i] = fl[i+len];
                fl[fl_len] = '\0';
            }
        }
        while (*fl != ' ' && *fl != '\0') {
            fl++;
            fl_len--;
        }
        while (*fl == ' ') {
            fl++;
            fl_len--;
        }
    }
    // check if spaces only
    fl = FlagsKludge;
    while (*fl == ' ')
        fl++;
    if (*fl == '\0')
        *FlagsKludge = '\0';
}

int CMsg::Delete() {
    fclose(f);
    f = NULL;
    return (remove(MsgFileName) == 0);
}

inline int CMsg::FromNet(SFTNAddress &Addr) const {
    return (From.Zone == Addr.Zone && From.Net == Addr.Net);
}

inline int CMsg::ToNet(SFTNAddress &Addr) const {
    return (To.Zone == Addr.Zone && To.Net == Addr.Net);
}

inline int CMsg::FromNode(SFTNAddress &Addr) const {
    return (From.Zone == Addr.Zone && From.Net == Addr.Net
            && From.Node == Addr.Node);
}

inline int CMsg::ToNode(SFTNAddress &Addr) const {
    return (To.Zone == Addr.Zone && To.Net == Addr.Net
            && To.Node == Addr.Node);
}

inline int CMsg::FromPoint(SFTNAddress &Addr) const {
    return (From.Zone == Addr.Zone && From.Net == Addr.Net
            && From.Node == Addr.Node && From.Point == Addr.Point);
}

inline int CMsg::ToPoint(SFTNAddress &Addr) const {
    return (To.Zone == Addr.Zone && To.Net == Addr.Net
            && To.Node == Addr.Node && To.Point == Addr.Point);
}

int CMsg::FromAddr(SFTNAddress &Addr) {
    return (FromPoint(Addr) && (stricmp(From.Domain, Addr.Domain) == 0));
}

int CMsg::ToAddr(SFTNAddress &Addr) {
    return (ToPoint(Addr) && (stricmp(To.Domain, Addr.Domain) == 0));
}

void CMsg::Bounce( uint16 BounceAttrs, uint16 BounceFlags,
                   char *BounceTplFileName,
                   SFTNAddress &BounceAddr )
{
    FILE        *b, *tpl;
    SMsgHeader   BounceHeader;
    static char  Kludge[MAX_KLUDGE_SIZE];
    static char  BounceFileName[MAX_PATH];
    static char  Str[MAX_KLUDGE_SIZE];
    size_t       BytesRead, BytesWritten;
    int          SuccessWrite;
    int          BufPos, CurPos, i;
    int          ReadState, CRWritten;
    const MAX_MACRO_SIZE = 20;
    char         Macro[MAX_MACRO_SIZE];
    enum { TPL_READ, TPL_SKIP_LF, TPL_MACRO };

    // fill bounce header
    memset( &BounceHeader, 0, sizeof(BounceHeader) );
    strcpy( BounceHeader.FromUserName, "Tracker" );
    if (Header.FromUserName)
        strncpy( BounceHeader.ToUserName, Header.FromUserName, 35 );
    else
        *(BounceHeader.ToUserName) = '\0';
    strcpy( BounceHeader.Subject, "Message bounced");
    GetCurrentDateTime( BounceHeader.DateTime );
    BounceHeader.Attrs     = BounceAttrs;
    BounceHeader.OrigZone  = BounceAddr.Zone;
    BounceHeader.OrigNet   = BounceAddr.Net;
    BounceHeader.OrigNode  = BounceAddr.Node;
    BounceHeader.OrigPoint = BounceAddr.Point;
    BounceHeader.DestZone  = From.Zone;
    BounceHeader.DestNet   = From.Net;
    BounceHeader.DestNode  = From.Node;
    BounceHeader.DestPoint = From.Point;
    
    sprintf( BounceFileName, "%s%ld.msg", "MailOut", FreeMsgNumber );
    FreeMsgNumber++;
    b = fopen( BounceFileName, "wb" );
    if (b) {
        // write header
        BytesWritten = fwrite( &BounceHeader, 1, sizeof(BounceHeader), b );
        SuccessWrite = BytesWritten == sizeof(Header);
        // write FMPT kludge, if needed
        if (SuccessWrite) {
            if (BounceAddr.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\1FMPT %hd%c", BounceAddr.Point, 0x0D );
                #else
                sprintf( Kludge, "\1FMPT %hu%c", BounceAddr.Point, 0x0D );
                #endif
                BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
                SuccessWrite = BytesWritten == strlen(Kludge);
            }
        }
        // write TOPT kludge, if needed
        if (SuccessWrite) {
            if (From.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\1TOPT %hd%c", From.Point, 0x0D );
                #else
                sprintf( Kludge, "\1TOPT %hu%c", From.Point, 0x0D );
                #endif
                BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
                SuccessWrite = BytesWritten == strlen(Kludge);
            }
        }
        // write INTL kludge
        if (SuccessWrite) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge, "\1INTL %hd:%hd/%hd %hd:%hd/%hd%c",
                     From.Zone, From.Net, From.Node,
                     BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                     0x0D );
            #else
            sprintf( Kludge, "\1INTL %hu:%hu/%hu %hu:%hu/%hu%c",
                     From.Zone, From.Net, From.Node,
                     BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                     0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write FLAGS kludge
        if (SuccessWrite && BounceFlags != 0) {
            strcpy( Kludge, "\1FLAGS" );
            if (BounceFlags & FLAG_DIRECT) {
                strcat( Kludge, " " );
                strcat( Kludge, "DIR" );
            }
            if (BounceFlags & FLAG_IMMEDIATE) {
                strcat( Kludge, " " );
                strcat( Kludge, "IMM" );
            }
            if (BounceFlags & FLAG_KILL_FILE_SENT) {
                strcat( Kludge, " " );
                strcat( Kludge, "KFS" );
            }
            if (BounceFlags & FLAG_TRUNCATE_FILE) {
                strcat( Kludge, " " );
                strcat( Kludge, "TFS" );
            }
            if (BounceFlags & FLAG_LOCK) {
                strcat( Kludge, " " );
                strcat( Kludge, "LOK" );
            }
            if (BounceFlags & FLAG_CONFIRM_REQ) {
                strcat( Kludge, " " );
                strcat( Kludge, "CFM" );
            }
            // add 0x0D
            Kludge[strlen(Kludge)+1] = '\0';
            Kludge[strlen(Kludge)] = 0x0D;
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write MSGID: kludge
        if (SuccessWrite) {
            if (BounceAddr.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\01MSGID: %hd:%hd/%hd.%hd@%s %0lx%c",
                         BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                         BounceAddr.Point, BounceAddr.Domain,
                         MsgIdHandler.Get(), 0x0D
                       );
                #else
                sprintf( Kludge, "\01MSGID: %hu:%hu/%hu.%hu@%s %0lx%c",
                         BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                         BounceAddr.Point, BounceAddr.Domain,
                         MsgIdHandler.Get(), 0x0D
                       );
                #endif
            }
            else {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\01MSGID: %hd:%hd/%hd@%s %0lx%c",
                         BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                         BounceAddr.Domain, MsgIdHandler.Get(), 0x0D
                       );
                #else
                sprintf( Kludge, "\01MSGID: %hu:%hu/%hu%s %0lx%c",
                         BounceAddr.Zone, BounceAddr.Net, BounceAddr.Node,
                         BounceAddr.Domain, MsgIdHandler.Get(), 0x0D
                       );
                #endif
            }
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write REPLY: kludge
        if (SuccessWrite && *MsgIdLine) {
            strcpy( Kludge, "\01REPLY: " );
            strcat( Kludge, MsgIdLine );
            Kludge[strlen(Kludge)+1] = '\0';
            Kludge[strlen(Kludge)] = 0x0D;
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write DOMAIN kludge
        if (SuccessWrite && stricmp(BounceAddr.Domain, From.Domain) != 0) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge,
                     "\1DOMAIN %s %hd:%hd/%hd.%hd %s %hd:%hd/%hd.%hd%c",
                     From.Domain, From.Zone, From.Net, From.Node, From.Point,
                     BounceAddr.Domain, BounceAddr.Zone, BounceAddr.Net,
                     BounceAddr.Node, BounceAddr.Point, 0x0D );
            #else
            sprintf( Kludge,
                     "\1DOMAIN %s %hu:%hu/%hu.%hu %s %hu:%hu/%hu.%hu%c",
                     From.Domain, From.Zone, From.Net, From.Node, From.Point,
                     BounceAddr.Domain, BounceAddr.Zone, BounceAddr.Net,
                     BounceAddr.Node, BounceAddr.Point, 0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), b );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        CRWritten = 1;
        if (SuccessWrite && BounceTplFileName != NULL) {
            tpl = fopen( BounceTplFileName, "rb" );
            if (tpl) {
                BytesRead = fread( Kludge, 1, MAX_KLUDGE_SIZE, tpl );
                BufPos = 0;
                ReadState = TPL_READ;
                while (SuccessWrite && BytesRead > 0) {
                    switch (ReadState) {
                    case TPL_READ:
                        CurPos = BufPos;
                        while (BufPos < BytesRead && Kludge[BufPos] != '@'
                              && Kludge[BufPos] != 0x0D)
                            BufPos++;
                        if (BufPos >= BytesRead) {
                            if (BufPos > CurPos) {
                                BytesWritten = fwrite( Kludge + CurPos, 1,
                                                       BufPos-CurPos, b );
                                SuccessWrite = (BytesWritten == BufPos-CurPos);
                                CRWritten = 0;
                            }
                            BytesRead = fread( Kludge, 1, MAX_KLUDGE_SIZE,
                                               tpl );
                            BufPos = 0;
                            break;
                        }
                        if (Kludge[BufPos] == '@') {
                            if (BufPos > CurPos) {
                                BytesWritten = fwrite( Kludge + CurPos, 1,
                                                       BufPos-CurPos, b );
                                SuccessWrite = (BytesWritten == BufPos-CurPos);
                                CRWritten = 0;
                            }
                            ReadState = TPL_MACRO;
                            BufPos++;
                            i = 0;
                            *Macro = '\0';
                            break;
                        }
                        // 0x0D
                        BufPos++;
                        BytesWritten = fwrite( Kludge + CurPos, 1,
                                               BufPos-CurPos, b );
                        SuccessWrite = (BytesWritten == BufPos-CurPos);
                        CRWritten = 1;
                        ReadState = TPL_SKIP_LF;
                        break;
                    case TPL_MACRO:
                        while (BufPos < BytesRead && i < MAX_MACRO_SIZE-1
                               && ((Kludge[BufPos] >= 'a' &&
                                    Kludge[BufPos] <= 'z') ||
                                   (Kludge[BufPos] >= 'A' &&
                                    Kludge[BufPos] <= 'Z')))
                        {
                            Macro[i] = Kludge[BufPos];
                            BufPos++;
                            i++;
                        }
                        Macro[i] = '\0';
                        if (BufPos >= BytesRead) {
                            BytesRead = fread( Kludge, 1, MAX_KLUDGE_SIZE,
                                               tpl );
                            BufPos = 0;
                            break;
                        }
                        if (*Macro == '\0') {
                            BytesWritten = fwrite( "@", 1,
                                                   1, b );
                            SuccessWrite = (BytesWritten == 1);
                            if (Kludge[BufPos] == '@')
                                BufPos++;
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "FName" ) == 0) {
                            BytesWritten = fwrite( Header.FromUserName, 1,
                                                   strlen(Header.FromUserName),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(Header.FromUserName));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "TName" ) == 0) {
                            BytesWritten = fwrite( Header.ToUserName, 1,
                                                   strlen(Header.ToUserName),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(Header.ToUserName));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "FAddr" ) == 0) {
                            #ifdef SIGNED_ADDR
                            sprintf( Str, "%hd:%hd/%hd.%hd",
                                     From.Zone, From.Net, From.Node,
                                     From.Point );
                            #else
                            sprintf( Str, "%hu:%hu/%hu.%hu",
                                     From.Zone, From.Net, From.Node,
                                     From.Point );
                            #endif
                            BytesWritten = fwrite( Str, 1, strlen(Str), b );
                            SuccessWrite = (BytesWritten == strlen(Str));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "TAddr" ) == 0) {
                            #ifdef SIGNED_ADDR
                            sprintf( Str, "%hd:%hd/%hd.%hd",
                                     To.Zone, To.Net, To.Node, To.Point );
                            #else
                            sprintf( Str, "%hu:%hu/%hu.%hu",
                                     To.Zone, To.Net, To.Node, To.Point );
                            #endif
                            BytesWritten = fwrite( Str, 1, strlen(Str), b );
                            SuccessWrite = (BytesWritten == strlen(Str));
                        }
                        else if (stricmp( Macro, "FDomain" ) == 0) {
                            BytesWritten = fwrite( From.Domain, 1,
                                                   strlen(From.Domain),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(From.Domain));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "TDomain" ) == 0) {
                            BytesWritten = fwrite( To.Domain, 1,
                                                   strlen(To.Domain),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(To.Domain));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "Subject" ) == 0) {
                            BytesWritten = fwrite( Header.Subject, 1,
                                                   strlen(Header.Subject),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(Header.Subject));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "SName" ) == 0) {
                            BytesWritten = fwrite( "SysOpName", 1,
                                                   strlen("SysOpName"),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen("SysOp"));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "SAddr" ) == 0) {
                            #ifdef SIGNED_ADDR
                            sprintf( Str, "%hd:%hd/%hd.%hd",
                                     BounceAddr.Zone, BounceAddr.Net,
                                     BounceAddr.Node, BounceAddr.Point );
                            #else
                            sprintf( Str, "%hu:%hu/%hu.%hu",
                                     BounceAddr.Zone, BounceAddr.Net,
                                     BounceAddr.Node, BounceAddr.Point );
                            #endif
                            BytesWritten = fwrite( Str, 1, strlen(Str), b );
                            SuccessWrite = (BytesWritten == strlen(Str));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "SDomain" ) == 0) {
                            BytesWritten = fwrite( BounceAddr.Domain, 1,
                                                   strlen(BounceAddr.Domain),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(BounceAddr.Domain));
                            CRWritten = 0;
                        }
                        else if (stricmp( Macro, "DateTime" ) == 0) {
                            BytesWritten = fwrite( Header.DateTime, 1,
                                                   strlen(Header.DateTime),
                                                   b );
                            SuccessWrite = (BytesWritten ==
                                           strlen(Header.DateTime));
                            CRWritten = 0;
                        }
                        ReadState = TPL_READ;
                        break;
                    case TPL_SKIP_LF:
                        if (BufPos >= BytesRead) {
                            BytesRead = fread( Kludge, 1, MAX_KLUDGE_SIZE,
                                               tpl );
                            BufPos = 0;
                            break;
                        }
                        if (Kludge[BufPos] == 0x0A)
                            BufPos++;
                        ReadState = TPL_READ;
                        break;
                    }
                }
                if (ReadState == TPL_MACRO) {
                    if (*Macro == '\0') {
                        BytesWritten = fwrite( "@", 1,
                                               1, b );
                        SuccessWrite = (BytesWritten == 1);
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "FName" ) == 0) {
                        BytesWritten = fwrite( Header.FromUserName, 1,
                                               strlen(Header.FromUserName),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(Header.FromUserName));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "TName" ) == 0) {
                        BytesWritten = fwrite( Header.ToUserName, 1,
                                               strlen(Header.ToUserName),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(Header.ToUserName));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "FAddr" ) == 0) {
                        #ifdef SIGNED_ADDR
                        sprintf( Str, "%hd:%hd/%hd.%hd",
                                 From.Zone, From.Net, From.Node,
                                 From.Point );
                        #else
                        sprintf( Str, "%hu:%hu/%hu.%hu",
                                 From.Zone, From.Net, From.Node,
                                 From.Point );
                        #endif
                        BytesWritten = fwrite( Str, 1, strlen(Str), b );
                        SuccessWrite = (BytesWritten == strlen(Str));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "TAddr" ) == 0) {
                        #ifdef SIGNED_ADDR
                        sprintf( Str, "%hd:%hd/%hd.%hd",
                                 To.Zone, To.Net, To.Node, To.Point );
                        #else
                        sprintf( Str, "%hu:%hu/%hu.%hu",
                                 To.Zone, To.Net, To.Node, To.Point );
                        #endif
                        BytesWritten = fwrite( Str, 1, strlen(Str), b );
                        SuccessWrite = (BytesWritten == strlen(Str));
                    }
                    else if (stricmp( Macro, "FDomain" ) == 0) {
                        BytesWritten = fwrite( From.Domain, 1,
                                               strlen(From.Domain),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(From.Domain));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "TDomain" ) == 0) {
                        BytesWritten = fwrite( To.Domain, 1,
                                               strlen(To.Domain),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(To.Domain));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "Subject" ) == 0) {
                        BytesWritten = fwrite( Header.Subject, 1,
                                               strlen(Header.Subject),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(Header.Subject));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "SName" ) == 0) {
                        BytesWritten = fwrite( "SysOpName", 1,
                                               strlen( "SysopName" ),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen("SysOpName"));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "SAddr" ) == 0) {
                        #ifdef SIGNED_ADDR
                        sprintf( Str, "%hd:%hd/%hd.%hd",
                                 BounceAddr.Zone, BounceAddr.Net,
                                 BounceAddr.Node, BounceAddr.Point );
                        #else
                        sprintf( Str, "%hu:%hu/%hu.%hu",
                                 BounceAddr.Zone, BounceAddr.Net,
                                 BounceAddr.Node, BounceAddr.Point );
                        #endif
                        BytesWritten = fwrite( Str, 1, strlen(Str), b );
                        SuccessWrite = (BytesWritten == strlen(Str));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "SDomain" ) == 0) {
                        BytesWritten = fwrite( BounceAddr.Domain, 1,
                                               strlen(BounceAddr.Domain),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(BounceAddr.Domain));
                        CRWritten = 0;
                    }
                    else if (stricmp( Macro, "DateTime" ) == 0) {
                        BytesWritten = fwrite( Header.DateTime, 1,
                                               strlen(Header.DateTime),
                                               b );
                        SuccessWrite = (BytesWritten ==
                                       strlen(Header.DateTime));
                        CRWritten = 0;
                    }
                }
                fclose(tpl);
            }
        }
        fclose(b);
    }

}

int CMsg::SaveMsg(FILE *tmp) {
    size_t      BytesRead, BytesWritten;
    static char Kludge[MAX_KLUDGE_SIZE];
    int         SuccessWrite;
    int         SaveCondition;
    int         WriteKludge;
    int         ZeroWritten;
    int         EOLRead, EOLWritten;
    int         BufPos, CurPos;
    char       *buf, *kl;
    int         i;
    
    rewind(f);
    fseek( f, sizeof(Header), SEEK_SET );
    // write header
    BytesWritten = fwrite( &Header, 1, sizeof(Header), tmp );
    SuccessWrite = BytesWritten == sizeof(Header);
    // write AREA , if needed
    if (SuccessWrite && *Area) {
        sprintf( Kludge, "AREA:%s%c", Area, 0x0D );
        BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
        SuccessWrite = BytesWritten == strlen(Kludge);
    }
    // write FMPT kludge, if needed
    if (SuccessWrite) {
        if (From.Point != 0) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge, "\1FMPT %hd%c", From.Point, 0x0D );
            #else
            sprintf( Kludge, "\1FMPT %hu%c", From.Point, 0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
    }
    // write TOPT kludge, if needed
    if (SuccessWrite) {
        if (To.Point != 0) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge, "\1TOPT %hd%c", To.Point, 0x0D );
            #else
            sprintf( Kludge, "\1TOPT %hu%c", To.Point, 0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
    }
    // write INTL kludge
    if (SuccessWrite) {
        #ifdef SIGNED_ADDR
        sprintf( Kludge, "\1INTL %hd:%hd/%hd %hd:%hd/%hd%c",
                 To.Zone, To.Net, To.Node, From.Zone, From.Net, From.Node,
                 0x0D );
        #else
        sprintf( Kludge, "\1INTL %hu:%hu/%hu %hu:%hu/%hu%c",
                 To.Zone, To.Net, To.Node, From.Zone, From.Net, From.Node,
                 0x0D );
        #endif
        BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
        SuccessWrite = BytesWritten == strlen(Kludge);
    }
    // write FLAGS kludge
    if (SuccessWrite && (Flags != 0 || *FlagsKludge != '\0')) {
        strcpy( Kludge, "\1FLAGS" );
        if (Flags & FLAG_DIRECT) {
            strcat( Kludge, " " );
            strcat( Kludge, "DIR" );
        }
        if (Flags & FLAG_IMMEDIATE) {
            strcat( Kludge, " " );
            strcat( Kludge, "IMM" );
        }
        if (Flags & FLAG_KILL_FILE_SENT) {
            strcat( Kludge, " " );
            strcat( Kludge, "KFS" );
        }
        if (Flags & FLAG_TRUNCATE_FILE) {
            strcat( Kludge, " " );
            strcat( Kludge, "TFS" );
        }
        if (Flags & FLAG_LOCK) {
            strcat( Kludge, " " );
            strcat( Kludge, "LOK" );
        }
        if (Flags & FLAG_CONFIRM_REQ) {
            strcat( Kludge, " " );
            strcat( Kludge, "CFM" );
        }
        if (*FlagsKludge != '\0')
            strcat( Kludge, FlagsKludge );
        // add 0x0D
        Kludge[strlen(Kludge)+1] = '\0';
        Kludge[strlen(Kludge)] = 0x0D;
        BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
        SuccessWrite = BytesWritten == strlen(Kludge);
    }
    // write DOMAIN kludge
    if (SuccessWrite && stricmp(From.Domain, To.Domain) != 0) {
        #ifdef SIGNED_ADDR
        sprintf( Kludge,
                 "\1DOMAIN %s %hd:%hd/%hd.%hd %s %hd:%hd/%hd.%hd%c",
                 To.Domain, To.Zone, To.Net, To.Node, To.Point,
                 From.Domain, From.Zone, From.Net, From.Node, From.Point,
                 0x0D );
        #else
        sprintf( Kludge,
                 "\1DOMAIN %s %hu:%hu/%hu.%hu %s %hu:%hu/%hu.%hu%c",
                 To.Domain, To.Zone, To.Net, To.Node, To.Point,
                 From.Domain, From.Zone, From.Net, From.Node, From.Point,
                 0x0D );
        #endif
        BytesWritten = fwrite( Kludge, 1, strlen(Kludge), tmp );
        SuccessWrite = BytesWritten == strlen(Kludge);
    }
    // write from msg to tmp
    BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize, f );
    BufPos = 0;
    Kludge[0] = 0x01;
    SaveCondition = SAVE_READ;
    ZeroWritten = 0;
    EOLRead = 0;
    EOLWritten = 1;
    while (SuccessWrite && BytesRead > 0) {
        switch (SaveCondition) {
        case SAVE_READ:
            CurPos = BufPos;
            buf    = GlobalMsgBuf + CurPos;
            // while not kludge
            while (BufPos < BytesRead && *buf != 0x01) {
                if (*buf != '\0')
                    EOLRead = 0;
                // while not end of line
                while (BufPos < BytesRead && *buf != 0x0D) {
                    BufPos++;
                    buf++;
                }
                // check CRLF sequence
                if (BufPos < BytesRead) { // 0x0D
                    EOLRead = 1;
                    BufPos++;
                    buf++;
                    if (BufPos < BytesRead && *buf == 0x0A) {
                        BufPos++;
                        buf++;
                    }
                }
            }
            // write data, if more than zero bytes
            if (BufPos > CurPos) {
                BytesWritten = fwrite( GlobalMsgBuf + CurPos, 1,
                                       BufPos-CurPos, tmp );
                ZeroWritten = (GlobalMsgBuf[BufPos-1] == '\0');
                EOLWritten = EOLRead;
                SuccessWrite = (BytesWritten == BufPos-CurPos);
            }
            if (BufPos < BytesRead) { // 0x01
                i = 1;
                BufPos++;
                SaveCondition = SAVE_MIDDLE_OF_KLUDGE;
            }
            else {
                BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                   f );
                BufPos = 0;
            }
            break;
        case SAVE_MIDDLE_OF_KLUDGE:
            buf = GlobalMsgBuf + BufPos;
            while (BufPos < BytesRead && i < (MAX_KLUDGE_SIZE-1) &&
                   *buf != 0x0D)
            {
                Kludge[i] = *buf;
                i++;
                buf++;
                BufPos++;
            }
            if (BufPos < BytesRead) { // Kludge had been read
                Kludge[i] = '\0';
                // check kludge
                WriteKludge = 1; // unknown writable kludge
                kl = Kludge+1;
                if (strncmp( kl, "FMPT ", 5) == 0) {
                    WriteKludge = 0;
                }
                else if (strncmp( kl, "TOPT ", 5) == 0) {
                    WriteKludge = 0;
                }
                else if (strncmp( kl, "INTL ", 5) == 0) {
                    WriteKludge = 0;
                }
                else if (strncmp( kl, "FLAGS ", 6) == 0) {
                    WriteKludge = 0;
                }
                else if (strncmp( kl, "DOMAIN ", 7) == 0) {
                    WriteKludge = 0;
                }
                if (WriteKludge) {
                    BytesWritten = fwrite( Kludge, 1, strlen(Kludge),
                                           tmp );
                    SuccessWrite = (BytesWritten == strlen(Kludge));
                    ZeroWritten = 0;
                    EOLWritten = 0;
                    SaveCondition = SAVE_WRITE_LINE;
                }
                else
                    SaveCondition = SAVE_SKIP_LINE;
            }
            else {
                BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                   f );
                BufPos = 0;
            }
            break;
        case SAVE_WRITE_LINE:
            CurPos = BufPos;
            buf = GlobalMsgBuf + CurPos;
            // while not end of line
            EOLRead = 0;
            while (BufPos < BytesRead && *buf != 0x0D) {
                BufPos++;
                buf++;
            }
            // check CRLF sequence
            if (BufPos < BytesRead) { // 0x0D
                EOLRead = 1;
                BufPos++;
                buf++;
                if (BufPos < BytesRead && *buf == 0x0A) {
                    BufPos++;
                    buf++;
                }
            }
            if (BufPos > CurPos) {
                BytesWritten = fwrite( GlobalMsgBuf + CurPos, 1 ,
                                       BufPos - CurPos, tmp );
                ZeroWritten = (GlobalMsgBuf[BufPos-1] == '\0');
                EOLWritten = EOLRead;
                SuccessWrite = (BytesWritten == BufPos-CurPos);
            }
            SaveCondition = SAVE_READ;
            break;
        case SAVE_SKIP_LINE:
            buf = GlobalMsgBuf + BufPos;
            // while not end of line
            while (BufPos < BytesRead && *buf != 0x0D) {
                BufPos++;
                buf++;
            }
            // check CRLF sequence
            if (BufPos < BytesRead) { // 0x0D
                BufPos++;
                buf++;
                if (BufPos < BytesRead && *buf == 0x0A) {
                    BufPos++;
                    buf++;
                }
            }
            SaveCondition = SAVE_READ;
            break;
        }
    }

    if (SuccessWrite && SaveCondition == SAVE_MIDDLE_OF_KLUDGE) {
        Kludge[i] = '\0';
        // check kludge
        WriteKludge = 1; // unknown writable kludge
        kl = Kludge+1;
        if (strncmp( kl, "FMPT ", 5) == 0) {
            WriteKludge = 0;
        }
        else if (strncmp( kl, "TOPT ", 5) == 0) {
            WriteKludge = 0;
        }
        else if (strncmp( kl, "INTL ", 5) == 0) {
            WriteKludge = 0;
        }
        else if (strncmp( kl, "FLAGS ", 6) == 0) {
            WriteKludge = 0;
        }
        else if (strncmp( kl, "DOMAIN ", 6) == 0) {
            WriteKludge = 0;
        }
        if (WriteKludge) {
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge),
                                   tmp );
            ZeroWritten = 0;
            EOLWritten = 0;
            SuccessWrite = (BytesWritten == strlen(Kludge));
        }
    }

    if (SuccessWrite && !ZeroWritten) {
        BytesWritten = fwrite( "", 1, 1, tmp );
        SuccessWrite = (BytesWritten == 1);
    }
    
    fclose(tmp);

    return (SuccessWrite);
}

int CMsg::Move(char *Dir, long &FreeNumber) {
    static char TempFileName[MAX_PATH];
    int         SuccessWrite;
    size_t      BytesRead, BytesWritten;
    FILE       *tmp;

    Header.Attrs = Attrs;

    // check, if moving in message directory
    if (stricmp( Dir, MsgFileDir ) == 0) {
        strcpy( TempFileName, "Config.Temp" );
        strcat( TempFileName, "Tracker.tmp" );
        tmp = GetTempFile(TempFileName, "wb" );
        if (tmp) {
            SuccessWrite = SaveMsg( tmp );
            if (SuccessWrite) {
                // compare letters of disks
                if (*MsgFileName == *TempFileName) {
                    fclose(f);
                    f = NULL;
                    remove( MsgFileName );
                    rename( TempFileName, MsgFileName );
                }
                else {
                    rewind(f);
                    tmp = fopen( TempFileName, "rb" );
                    BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                       tmp );
                    while (SuccessWrite && BytesRead > 0) {
                        BytesWritten = fwrite( GlobalMsgBuf, 1, BytesRead, f );
                        SuccessWrite = (BytesWritten == BytesRead);
                        BytesRead = fread( GlobalMsgBuf, 1, GlobalMsgBufSize,
                                           tmp );
                    }
                    fclose(tmp);
                    remove( TempFileName );
                }
            }
        }
    }
    else {
        sprintf( TempFileName, "%s%ld.msg", Dir, FreeNumber );
        tmp = fopen( TempFileName, "wb" );
        SuccessWrite = 0;
        if (tmp) {
            SuccessWrite = SaveMsg( tmp );
            if (SuccessWrite) {
                fclose(f);
                f = NULL;
                remove( MsgFileName );
                FreeNumber++;
            }
        }
    }

    
    if (f) {
        fclose(f);
	f = NULL;
    }

    return (SuccessWrite && tmp != NULL);
}

int CMsg::FromUs() {
    if (FromPoint(GlobalVars.MainAddr))
        return (1);
    return (0);
}

int CMsg::ToUs() {
    if (ToPoint(GlobalVars.MainAddr))
        return (1);
    return (0);
}

void CMsg::Reply( uint16 ReplyAttrs, uint16 ReplyFlags, char *Subject,
                  SStringList *BodyLines )
{
    FILE        *r;
    SMsgHeader   ReplyHeader;
    static char  Kludge[MAX_KLUDGE_SIZE];
    static char  ReplyFileName[MAX_PATH];
    size_t       BytesWritten;
    int          SuccessWrite;
    SStringList *CurLine;

    // fill reply header
    memset( &ReplyHeader, 0, sizeof(ReplyHeader) );
    strncpy( ReplyHeader.FromUserName, Header.ToUserName, 35 );
    if (Header.FromUserName)
        strncpy( ReplyHeader.ToUserName, Header.FromUserName, 35 );
    else
        *(ReplyHeader.ToUserName) = '\0';
    strncpy( ReplyHeader.Subject, Subject, 71 );
    GetCurrentDateTime( ReplyHeader.DateTime );
    ReplyHeader.Attrs     = ReplyAttrs;
    ReplyHeader.OrigZone  = To.Zone;
    ReplyHeader.OrigNet   = To.Net;
    ReplyHeader.OrigNode  = To.Node;
    ReplyHeader.OrigPoint = To.Point;
    ReplyHeader.DestZone  = From.Zone;
    ReplyHeader.DestNet   = From.Net;
    ReplyHeader.DestNode  = From.Node;
    ReplyHeader.DestPoint = From.Point;
    
    sprintf( ReplyFileName, "%s%ld.msg", GlobalVars.NetmailPath,
             FreeMsgNumber );
    FreeMsgNumber++;
    r = fopen( ReplyFileName, "wb" );
    if (r) {
        // write header
        BytesWritten = fwrite( &ReplyHeader, 1, sizeof(ReplyHeader), r );
        SuccessWrite = BytesWritten == sizeof(Header);
        // write FMPT kludge, if needed
        if (SuccessWrite) {
            if (To.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\1FMPT %hd%c", To.Point, 0x0D );
                #else
                sprintf( Kludge, "\1FMPT %hu%c", To.Point, 0x0D );
                #endif
                BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
                SuccessWrite = BytesWritten == strlen(Kludge);
            }
        }
        // write TOPT kludge, if needed
        if (SuccessWrite) {
            if (From.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\1TOPT %hd%c", From.Point, 0x0D );
                #else
                sprintf( Kludge, "\1TOPT %hu%c", From.Point, 0x0D );
                #endif
                BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
                SuccessWrite = BytesWritten == strlen(Kludge);
            }
        }
        // write INTL kludge
        if (SuccessWrite) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge, "\1INTL %hd:%hd/%hd %hd:%hd/%hd%c",
                     From.Zone, From.Net, From.Node,
                     To.Zone, To.Net, To.Node,
                     0x0D );
            #else
            sprintf( Kludge, "\1INTL %hu:%hu/%hu %hu:%hu/%hu%c",
                     From.Zone, From.Net, From.Node,
                     To.Zone, To.Net, To.Node,
                     0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write FLAGS kludge
        if (SuccessWrite && ReplyFlags != 0) {
            strcpy( Kludge, "\1FLAGS" );
            if (ReplyFlags & FLAG_DIRECT) {
                strcat( Kludge, " " );
                strcat( Kludge, "DIR" );
            }
            if (ReplyFlags & FLAG_IMMEDIATE) {
                strcat( Kludge, " " );
                strcat( Kludge, "IMM" );
            }
            if (ReplyFlags & FLAG_KILL_FILE_SENT) {
                strcat( Kludge, " " );
                strcat( Kludge, "KFS" );
            }
            if (ReplyFlags & FLAG_TRUNCATE_FILE) {
                strcat( Kludge, " " );
                strcat( Kludge, "TFS" );
            }
            if (ReplyFlags & FLAG_LOCK) {
                strcat( Kludge, " " );
                strcat( Kludge, "LOK" );
            }
            if (ReplyFlags & FLAG_CONFIRM_REQ) {
                strcat( Kludge, " " );
                strcat( Kludge, "CFM" );
            }
            // add 0x0D
            Kludge[strlen(Kludge)+1] = '\0';
            Kludge[strlen(Kludge)] = 0x0D;
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write MSGID: kludge
        if (SuccessWrite) {
            if (To.Point != 0) {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\01MSGID: %hd:%hd/%hd.%hd@%s %0lx%c",
                         To.Zone, To.Net, To.Node,
                         To.Point, To.Domain,
                         MsgIdHandler.Get(), 0x0D
                       );
                #else
                sprintf( Kludge, "\01MSGID: %hu:%hu/%hu.%hu@%s %0lx%c",
                         To.Zone, To.Net, To.Node,
                         To.Point, To.Domain,
                         MsgIdHandler.Get(), 0x0D
                       );
                #endif
            }
            else {
                #ifdef SIGNED_ADDR
                sprintf( Kludge, "\01MSGID: %hd:%hd/%hd@%s %0lx%c",
                         To.Zone, To.Net, To.Node,
                         To.Domain, MsgIdHandler.Get(), 0x0D
                       );
                #else
                sprintf( Kludge, "\01MSGID: %hu:%hu/%hu%s %0lx%c",
                         To.Zone, To.Net, To.Node,
                         To.Domain, MsgIdHandler.Get(), 0x0D
                       );
                #endif
            }
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write REPLY: kludge
        if (SuccessWrite && *MsgIdLine) {
            strcpy( Kludge, "\01REPLY: " );
            strcat( Kludge, MsgIdLine );
            Kludge[strlen(Kludge)+1] = '\0';
            Kludge[strlen(Kludge)] = 0x0D;
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write DOMAIN kludge
        if (SuccessWrite && stricmp(To.Domain, From.Domain) != 0) {
            #ifdef SIGNED_ADDR
            sprintf( Kludge,
                     "\1DOMAIN %s %hd:%hd/%hd.%hd %s %hd:%hd/%hd.%hd%c",
                     From.Domain, From.Zone, From.Net, From.Node, From.Point,
                     To.Domain, To.Zone, To.Net,
                     To.Node, To.Point, 0x0D );
            #else
            sprintf( Kludge,
                     "\1DOMAIN %s %hu:%hu/%hu.%hu %s %hu:%hu/%hu.%hu%c",
                     From.Domain, From.Zone, From.Net, From.Node, From.Point,
                     To.Domain, To.Zone, To.Net,
                     To.Node, To.Point, 0x0D );
            #endif
            BytesWritten = fwrite( Kludge, 1, strlen(Kludge), r );
            SuccessWrite = BytesWritten == strlen(Kludge);
        }
        // write body lines
        CurLine = BodyLines;
        Kludge[0] = 0x0D;
        while (CurLine != NULL && SuccessWrite) {
            if (strlen(CurLine->String) > 0) {
                BytesWritten = fwrite( CurLine->String, 1, strlen(CurLine->String),
                                       r );
                SuccessWrite = BytesWritten == strlen(CurLine->String);
            }
            if (SuccessWrite) {
                BytesWritten = fwrite( Kludge, 1, 1, r );
                SuccessWrite = BytesWritten == 1;
                CurLine = CurLine->Next;
            }
        }
        if (SuccessWrite) {
            Kludge[0] = '\0';
            BytesWritten = fwrite( Kludge, 1, 1, r );
            SuccessWrite = BytesWritten == 1;
        }
        fclose(r);
    }
    else {
	    GlobalVars.Log << "Failed to create " << ReplyFileName << endl;
    }
}

void CMsg::SaveAsReceived() {
    fclose(f);
    
    Header.Attrs |= ATTR_RECEIVED;
    f = fopen( MsgFileName, "r+b" );
    fwrite( &Header, 1, sizeof(Header), f );
    fclose( f );
    f = NULL;
}
