// changes log:
// 01.02.2000 22:18: added info on errors (we keep it in szLastError) in CTic::Read

#include "pch.h"

#include "Common.hpp"
#include "FTNStruc.hpp"
#include "CfgParse.hpp"
#include "Misc.hpp"
#include "BSO.hpp"
#include "Adrenalin.hpp"
#include "crc32.hpp"
#include "Tic.hpp"
#include "ticlexems.hpp"

/*
 * Uncomment the definition below to use strict fsc accordance
 * algorithm
 * 
 */

//#define STRICT_FSC_MAPPING

// ---------
// CTicNamer
// ---------

CTicNamer::CTicNamer() {
    resetFields();
}

CTicNamer::CTicNamer(char *seekPath) {
    resetFields();
    readTICs(seekPath);
}

CTicNamer::~CTicNamer() {
    freeFields();
}

void CTicNamer::resetFields() {
    intervals       = NULL;
    intervalsCount  = 0;
    freeNumber      = -1;
    currentInterval = -1;
}

void CTicNamer::freeFields() {
    if (intervals != NULL && intervalsCount > 0)
        free(intervals);
    resetFields();
}

int compareIntegers(const void *p1, const void *p2) {
    int *i1, *i2;

    i1 = (int*)p1;
    i2 = (int*)p2;
    return ((*i1 < *i2) ? -1 : ((*i1 == *i2) ? 0 : 1));
}

void CTicNamer::readTICs(char *seekPath) {
    WIN32_FIND_DATA  findData;
    HANDLE           fileHandle;
    BOOL             fileFoundFlag;
    char             seekMask[MAX_PATH];
    int             *numbers;
    int             *newNumbers;
    int              numbersCount;
    int              newNumber;
    SNumberInterval *newIntervals;
    int              i, j, len;

    freeFields();
    
    numbers = NULL;
    numbersCount = 0;
    
    strcpy( seekMask, seekPath );
    if (seekMask[strlen(seekMask)-1] == '\\')
        strcat( seekMask, "*.tic" );
    else
        strcat( seekMask, "\\*.tic" );

    fileHandle = FindFirstFile( seekMask, &findData );
    fileFoundFlag = (fileHandle != INVALID_HANDLE_VALUE);
    while (fileFoundFlag) {
        newNumber = parseNumber( findData.cFileName );
        if (newNumber >= 0) {
            newNumbers = (int*)realloc( numbers, sizeof(int)*(numbersCount+1));
            if (newNumbers != NULL) {
                numbers = newNumbers;
                numbers[numbersCount] = newNumber;
                numbersCount++;
            }
        }
        fileFoundFlag = FindNextFile( fileHandle, &findData );
    }
    if (fileHandle != INVALID_HANDLE_VALUE)
        FindClose( fileHandle );

    if (numbersCount > 0) {
        qsort( (void*)numbers, numbersCount, sizeof(int), compareIntegers );
        i = 0;
        while (i < numbersCount) {
            len = 1;
            newNumber = numbers[i];
            j = i+1;
            while (j < numbersCount && numbers[j] == newNumber+1) {
                newNumber = numbers[j];
                len++;
                j++;
            }
            newIntervals = (SNumberInterval*)realloc( intervals,
                                                      (intervalsCount+1)*
                                                      sizeof(SNumberInterval));
            if (newIntervals != NULL) {
                intervals = newIntervals;
                intervals[intervalsCount].number = numbers[i];
                intervals[intervalsCount].length = len;
                intervalsCount++;
            }
            i += len;
        }
        free( numbers );
        
        // prepare address string

        char	addr[5+1+5+1+5+1+5+1+MAX_DOMAIN_SIZE+1];

        sprintf( addr, "%d:%d/%d.%d@%s",
                 GlobalVars.MainAddr.Zone, GlobalVars.MainAddr.Net,
                 GlobalVars.MainAddr.Node, GlobalVars.MainAddr.Point,
                 GlobalVars.MainAddr.Domain );

        UINT	crc = getBufferCRC32( addr, strlen(addr) );

        // cout << "DEBUG\nAddress crc: " << crc << "\nDEBUG\n";

        freeNumber      = (((int)time(NULL) ^ crc)) & 0xFFFFF;
        if (freeNumber > 999999) {
        	freeNumber = 0;
        }

        currentInterval = 0;
        while (currentInterval < intervalsCount && freeNumber >= intervals[currentInterval].number) {
        	if (freeNumber < (intervals[currentInterval].number + intervals[currentInterval].length)) {
        		freeNumber = intervals[currentInterval].number + intervals[currentInterval].length;
        	}
       		currentInterval++;
        }

        if (freeNumber > 999999) {
        	freeNumber = 0;
        	currentInterval = 0;
        	if (intervals[0].number == 0) {
        		freeNumber = intervals[0].length;
        		currentInterval++;
        	}
        }

        // debug
        // cout << "DEBUG\nFree number: " << freeNumber << "\nDEBUG\n";
        
    }
    else {
        freeNumber      = 0;
        currentInterval = 1;
    }
}

int CTicNamer::parseNumber(char *ticName) {
    int number;
    int count;

    // ticName is in format <letter><letter><6 numbers>.tic
    // its length is 12

    if (strlen(ticName) != 12)
        return (-1);

    // skip 2 first letters
    ticName += 2;

    number = 0;
    count  = 0;
    while (count < 6 && isdigit(*ticName)) {
        number *= 10;
        number += *ticName - '0';
        ticName++;
        count++;
    }
    
    if (count != 6)
        return (-1);
    
    if (stricmp( ticName, ".tic" ) != 0)
        return (-1);

    return (number);
}


int CTicNamer::getFreeNumber() {
    int oldFreeNumber = freeNumber;

    if (freeNumber == 999999) {
    	freeNumber = 0;
    	currentInterval = 0;
    }
    else {
    	freeNumber++;
    }

    if (currentInterval < intervalsCount) {
        if (freeNumber == intervals[currentInterval].number) {
            freeNumber += intervals[currentInterval].length;
            if (freeNumber > 999999) {
            	freeNumber = 0;
            	currentInterval = 0;
        	if (intervals[0].number == 0) {
        		freeNumber = intervals[0].length;
        		currentInterval++;
        	}
            }
            else {
            	currentInterval++;
            }
        }
    }

    return (oldFreeNumber);
}

void CTicNamer::printDebugInfo() {
    int i;
    
    cout << "Intervals count: " << intervalsCount << endl;
    for (i = 0; i < intervalsCount; i++) {
        cout << "interval " << i << ": from " << intervals[i].number
            << ", with length " << intervals[i].length << endl;
    }
    cout << "First 10 free numbers:\n";
    for (i = 0; i < 10; i++)
        cout << getFreeNumber() << endl;
}


// uncomment next definition to see TIC information during parsering
// #define TIC_DEBUG_INFO

CTic::CTic() : BSOManager(), ticNamer(GlobalVars.NewTicsPath) {
    LDescList   = NULL;
    ViaList     = NULL;
    PathList    = NULL;
    SeenbyList  = NULL;
    UnknownList = NULL;
    Free();
}

CTic::~CTic() {
    Free();
}

void CTic::Free() {
    SStringList *LastString;
    
    *Area      = '\0';
    *AreaDesc  = '\0';
    *File      = '\0';
    *FullName  = '\0';
    CRC32      = 0;
    *Magic = '\0';
    *Replaces  = '\0';
    *Desc      = '\0';
    while (LDescList != NULL) {
        LastString = LDescList->Next;
        delete [] LDescList->String;
        delete LDescList;
        LDescList = LastString;
    }
    Size             = -1;
    Date             = 0;
    Release          = 0;
    *Author          = '\0';
    *Source          = '\0';
    *App             = '\0';
    memset( &Origin, 0, sizeof(SFTNAddress) );
    memset( &From,   0, sizeof(SFTNAddress) );
    memset( &To,     0, sizeof(SFTNAddress) );
    *CreatedBy       = '\0';
    while (ViaList != NULL) {
        LastString = ViaList->Next;
        delete [] ViaList->String;
        delete ViaList;
        ViaList = LastString;
    }
    LastVia = NULL;
    while (PathList != NULL) {
        LastString = PathList->Next;
        delete [] PathList->String;
        delete PathList;
        PathList = LastString;
    }
    LastPath = NULL;
    while (SeenbyList != NULL) {
        LastSeenby = SeenbyList->Next;
        delete SeenbyList;
        SeenbyList = LastSeenby;
    }
    *PW = '\0';
    while (UnknownList != NULL) {
        LastString = UnknownList->Next;
        delete [] UnknownList->String;
        delete UnknownList;
        UnknownList = LastString;
    }
    *FromPwd = '\0';
}

int CTic::ParseAddr(char *s, SFTNAddress &Addr) {
    char FullAddr[24+MAX_DOMAIN_SIZE];
    char TempStr[24+MAX_DOMAIN_SIZE];
    char Domain[MAX_DOMAIN_SIZE];
    int  ColonPresent, SlashPresent, PointPresent, DogPresent, GridPresent;
    int  DogPos, GridPos;
    int  i;

    if (strlen(s) >= 24+MAX_DOMAIN_SIZE)
        return (0);
    
    GridPresent = ColonPresent = SlashPresent = PointPresent = DogPresent = 0;
    i = 0;
    while (s[i] != '\0') {
        switch (s[i]) {
        case '#':
            if (GridPresent == 0) {
                GridPresent = 1;
                GridPos     = i;
            }
            break;
        case ':':
            ColonPresent = 1;
            break;
        case '/':
            SlashPresent = 1;
            break;
        case '.':
            PointPresent = 1;
            break;
        case '@':
            if (DogPresent == 0) {
                DogPresent = 1;
                DogPos     = i;
            }
            break;
        }
        i++;
    }
    strcpy( TempStr, s );
    if (GridPresent) {
        if (!(ColonPresent & SlashPresent & PointPresent & !DogPresent))
            return (0);
        if (GridPos > MAX_DOMAIN_SIZE)
            return (0);
        strncpy( Domain, s, GridPos );
        strcpy( TempStr, s + GridPos + 1);
    }
    if (DogPresent && !PointPresent) {
        strncpy( Domain, TempStr + DogPos + 1, MAX_DOMAIN_SIZE - 1 );
        TempStr[DogPos] = '\0';
    }
    if (!SlashPresent) {
        #ifdef SIGNED_ADDR
        sprintf( FullAddr, "%ld/%s", GlobalVars.MainAddr.Net, TempStr );
        #else
        sprintf( FullAddr, "%lu/%s", GlobalVars.MainAddr.Net, TempStr );
        #endif
        strcpy( TempStr, FullAddr );
    }
    if (!ColonPresent) {
        #ifdef SIGNED_ADDR
        sprintf( FullAddr, "%ld:%s", GlobalVars.MainAddr.Zone, TempStr );
        #else
        sprintf( FullAddr, "%lu:%s", GlobalVars.MainAddr.Zone, TempStr );
        #endif
    }
    else
        strcpy( FullAddr, TempStr );

    if (!ParseFTNAddress( FullAddr, Addr ))
        return (0);

    if (GridPresent)
        strcpy( Addr.Domain, Domain );
    else if (!DogPresent) {
	    // let's find link with the same zone
	    pAdrenalin->Links.SelectFirst();
	    while (pAdrenalin->Links.Selected != NULL) {
		    if (pAdrenalin->Links.Selected->Address.Zone == Addr.Zone)
			    break;
		    pAdrenalin->Links.SelectNext();
	    }
	    if (pAdrenalin->Links.Selected != NULL)
		    strcpy( Addr.Domain, pAdrenalin->Links.Selected->Address.Domain );
	    else
		    strcpy( Addr.Domain, GlobalVars.MainAddr.Domain );
    }
    else if (!PointPresent) {
        strcpy( Addr.Domain, Domain );
    }

    return (*(Addr.Domain) != '\0');
}

int CTic::Read(char *TicFileName) {
    SStringList *LastUnknown, *LastLDesc;
    int          i;
    char         number;
    SFTNAddress  SeenbyAddr;
    CCfgParser   CfgParser( TicFileName );
    int          CRCDefined;

    Free();
    
    if (CfgParser.Failed())
        return (0);

    CRCDefined = 0;
    
    while (!CfgParser.Eof()) {
        CfgParser.ReadLine();
        if (!CfgParser.Failed() && CfgParser.Count() > 0) {
            switch (getTicLexemID( CfgParser[0] )) {
                /******
                  Area
                 ******/
                case TIC_LEXEM_AREA: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Area\n";
                    #endif
                    if (*Area != '\0') {
			    sprintf( szLastError, "duplicate AREA tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) > AREA_NAME_SIZE-1) {
			    sprintf( szLastError, "too long area name in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( Area, CfgParser[1] );
                	break;
                }
                /**********
                  AreaDesc
                 **********/
                case TIC_LEXEM_AREADESC: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse AreaDesc\n";
                    #endif
                    if (*AreaDesc != '\0') {
			    sprintf( szLastError, "duplicate AREADESC tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() == 1) {
			    sprintf( szLastError, "value absent in line %d", CfgParser.Line );
			    return (0);
		    }
                    i = strlen( CfgParser.String() + 9 );
                    if (i > AREA_DESC_SIZE-1) {
                        // cout << "Warning: area description was shortened\n";
                        i = AREA_DESC_SIZE-1;
                    }
                    memcpy( AreaDesc, CfgParser.String() + 9, i );
                    AreaDesc[i] = '\0';
                    break;
                }
                /******
                  File
                 ******/
                case TIC_LEXEM_FILE: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse File\n";
                    #endif
                    if (*File != '\0') {
			    sprintf( szLastError, "duplicate FILE tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) > MAX_PATH-1) {
			    sprintf( szLastError, "too long filename in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( File, CfgParser[1] );
                    break;
                }
                /**********
                  FullName
                 **********/
                case TIC_LEXEM_FULLNAME: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse FullName\n";
                    #endif
                    if (*FullName != '\0') {
			    sprintf( szLastError, "duplicate FULLNAME tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) > MAX_PATH-1) {
			    sprintf( szLastError, "too long filename in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( FullName, CfgParser[1] );
                    if (FullName[0] == '"')
                        RemoveBoundaryQuotes( FullName );
                    if (FullName[0] == '\0') {
			    sprintf( szLastError, "filename absent in line %d", CfgParser.Line );
			    return (0);
		    }
                    break;
                }
                /*****
                  CRC
                 *****/
                case TIC_LEXEM_CRC: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse CRC\n";
                    #endif
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
    		#ifdef STRICT_FSC_MAPPING
                    if (strlen(CfgParser[1]) != 8) {
			    sprintf( szLastError, "CRC string length differs from 8 symbols in line %d", CfgParser.Line );
			    return (0);
		    }
                    #endif
                    if (CRCDefined) {
			    sprintf( szLastError, "duplicate CRC tag in line %d", CfgParser.Line );
			    return (0);
		    }
    
                    #ifdef STRICT_FSC_MAPPING
    
                    for (CRC32 = 0, i = 0; i < 8; i++) {
                        number = toupper(CfgParser[1][i]);
                        if (!isalnum(number) || number > 'F')
                            return (0);
                        CRC32 <<= 4;
                        if (isalpha(number))
                            CRC32 += number - 'A' + 10;
                        else
                            CRC32 += number - '0';
                    }
    
                    #else
    
                    for (CRC32 = 0, i = 0; i < strlen(CfgParser[1]); i++) {
                        	number = toupper(CfgParser[1][i]);
                        	if (!isalnum(number) || number > 'F')
                            	return (0);
                        	CRC32 <<= 4;
                        	if (isalpha(number))
                            	CRC32 += number - 'A' + 10;
                        	else
                            	CRC32 += number - '0';
                    }
    
                    #endif
    
                    CRCDefined = 1;
                    break;
                }
                /*******
                  Magic
                 *******/
                case TIC_LEXEM_MAGIC: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Magic\n";
                    #endif
                    if (*Magic != '\0') {
			    sprintf( szLastError, "duplicate MAGIC tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) > MAX_PATH-1) {
			    sprintf( szLastError, "too long filename in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( Magic, CfgParser[1] );
                    break;
                }
                /**********
                  Replaces
                 **********/
                case TIC_LEXEM_REPLACES: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Replaces\n";
                    #endif
                    if (*Replaces != '\0') {
			    sprintf( szLastError, "duplicate REPLACES tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) > MAX_PATH-1) {
			    sprintf( szLastError, "too long filename in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( Replaces, CfgParser[1] );
                    break;
                }
                /******
                  Desc
                 ******/
                case TIC_LEXEM_DESC: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Desc\n";
                    #endif
                    if (*Desc != '\0') {
			    break;
		    }
                    if (CfgParser.String()[0] == ' ') { 
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser.String() + 5) > 256) {
			    sprintf( szLastError, "too long description in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( Desc, CfgParser.String() + 5 );
                    break;
                }
                /*******
                  LDesc
                 *******/
                case TIC_LEXEM_LDESC: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse LDesc\n";
                    #endif
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (LDescList == NULL) {
                        LDescList = new SStringList;
                        LastLDesc = LDescList;
                    }
                    else {
                        LastLDesc->Next = new SStringList;
                        LastLDesc = LastLDesc->Next;
                    }
                    LastLDesc->Next = NULL;
                    LastLDesc->String = new char[strlen(CfgParser.String() + 6)
                                                 + 1];
                    strcpy( LastLDesc->String, CfgParser.String() + 6 );
                    break;
                }
                /******
                  Size
                 ******/
                case TIC_LEXEM_SIZE: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Size\n";
                    #endif
                    if (Size != -1) {
			    sprintf( szLastError, "duplicate SIZE tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (sscanf( CfgParser[1], "%ld", &Size ) != 1) {
			    sprintf( szLastError, "error parsering size value in line %d", CfgParser.Line );
			    return (0);
		    }
                    break;
                }
                /******
                  Date
                 ******/
                case TIC_LEXEM_DATE: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Date\n";
                    #endif
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) != 8) {
			    sprintf( szLastError, "date value length differs from 8 in line %d", CfgParser.Line );
			    return (0);
		    }
                    for (Date = 0, i = 0; i < 8; i++) {
                        number = toupper(CfgParser[1][i]);
                        if (!isalnum(number) || number > 'F')
                            return (0);
                        Date <<= 4;
                        if (isalpha(number))
                            Date += number - 'A' + 10;
                        else
                            Date += number - '0';
                    }
                    break;
                }
                /*********
                  Release
                 *********/
                case TIC_LEXEM_RELEASE: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Release\n";
                    #endif
                    if (CfgParser.Count() > 2) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (strlen(CfgParser[1]) != 8) {
			    sprintf( szLastError, "release value length differs from 8 in line %d", CfgParser.Line );
			    return (0);
		    }
                    for (Release = 0, i = 0; i < 8; i++) {
                        number = toupper(CfgParser[1][i]);
                        if (!isalnum(number) || number > 'F')
                            return (0);
                        Release <<= 4;
                        if (isalpha(number))
                            Release += number - 'A' + 10;
                        else
                            Release += number - '0';
                    }
                    break;
                }
                /********
                  Author
                 ********/
                case TIC_LEXEM_AUTHOR: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Author\n";
                    #endif
                    if (*Author != '\0')
			    break;
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (strlen(CfgParser.String() + 7) > 256) {
			    sprintf( szLastError, "too long AUTHOR value in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    strcpy( Author, CfgParser.String() + 7 );
                    break;
                }
                /********
                  Source
                 ********/
                case TIC_LEXEM_SOURCE: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Source\n";
                    #endif
                    if (*Source != '\0')
                        break;
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (strlen(CfgParser.String() + 7) > 256) {
			    sprintf( szLastError, "too long SOURCE value in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    strcpy( Source, CfgParser.String() + 7 );
                    break;
                }
                /*****
                  App
                 *****/
                case TIC_LEXEM_APP: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse App\n";
                    #endif
                    break;
                }
                /********
                  Origin
                 ********/
                case TIC_LEXEM_ORIGIN: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Origin\n";
                    #endif
                    if (Origin.Zone != 0) {
			    sprintf( szLastError, "duplicate ORIGIN tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() != 2) {
			    sprintf( szLastError, "incorrect number of values (not 1) in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (!ParseAddr( CfgParser[1], Origin )) {
			    sprintf( szLastError, "failed to parse ORIGIN address in line %d", CfgParser.Line );
			    return (0);
		    }
                    break;
                }
                /******
                  From
                 ******/
                case TIC_LEXEM_FROM: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse From\n";
                    #endif
                    if (From.Zone != 0) {
			    sprintf( szLastError, "duplicate FROM tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() > 3) {
			    sprintf( szLastError, "extra values in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (!ParseAddr( CfgParser[1], From )) {
			    sprintf( szLastError, "failed to parse FROM address in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() == 3)
                        strcpy( FromPwd, CfgParser[2] );
                    break;
                }
                /****
                  To
                 *****/
                case TIC_LEXEM_TO: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse To\n";
                    #endif
                    if (To.Zone != 0) {
			    sprintf( szLastError, "duplicate TO tag in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (CfgParser.Count() != 2) {
			    sprintf( szLastError, "incorrect number of values (not 1) in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (!ParseAddr( CfgParser[1], To )) {
			    sprintf( szLastError, "failed to parse TO address in line %d", CfgParser.Line );
			    return (0);
		    }
                    break;
                }
                /*********
                  Created
                 *********/
                case TIC_LEXEM_CREATED: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Created\n";
                    #endif
                    if (*CreatedBy != '\0')
                        break;
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (strlen(CfgParser.String() + 8) > 256) {
			    sprintf( szLastError, "too long CREATED value in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (strnicmp( CfgParser.String(), "Created by ", 11 ) == 0)
                        strcpy( CreatedBy, CfgParser.String() + 11 );
                    else
                        strcpy( CreatedBy, CfgParser.String() + 8 );
                    break;
                }
                /*****
                  Via
                 *****/
                case TIC_LEXEM_VIA: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Via\n";
                    #endif
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (ViaList == NULL) {
                        ViaList = new SStringList;
                        LastVia = ViaList;
                    }
                    else {
                        LastVia->Next = new SStringList;
                        LastVia = LastVia->Next;
                    }
                    LastVia->Next = NULL;
                    LastVia->String = new char[strlen(CfgParser.String() + 4)
                                                 + 1];
                    strcpy( LastVia->String, CfgParser.String() + 4 );
                    break;
                }
                /******
                  Path
                 ******/
                case TIC_LEXEM_PATH: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Path\n";
                    #endif
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (PathList == NULL) {
                        PathList = new SStringList;
                        LastPath = PathList;
                    }
                    else {
                        LastPath->Next = new SStringList;
                        LastPath = LastPath->Next;
                    }
                    LastPath->Next = NULL;
                    LastPath->String = new char[strlen(CfgParser.String() + 5)
                                                 + 1];
                    strcpy( LastPath->String, CfgParser.String() + 5 );
                    break;
                }
                /********
                  SeenBy
                 ********/
                case TIC_LEXEM_SEENBY: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse SeenBy\n";
                    #endif
                    if (CfgParser.String()[0] == ' ') {
			    sprintf( szLastError, "extra spaces in line %d", CfgParser.Line );                    
			    return (0);
		    }
                    if (!ParseAddr( CfgParser.String() + 7, SeenbyAddr )) {
			    sprintf( szLastError, "failed to parse SEENBY address in line %d", CfgParser.Line );
			    return (0);
		    }
                    if (SeenbyList == NULL) {
                        SeenbyList = new SAddrList;
                        LastSeenby = SeenbyList;
                    }
                    else {
                        LastSeenby->Next = new SAddrList;
                        LastSeenby = LastSeenby->Next;
                    }
                    LastSeenby->Next = NULL;
                    ParseAddr( CfgParser.String() + 7, LastSeenby->Addr );
                    break;
                }
                /********
                  Pw
                 ********/
                case TIC_LEXEM_PW: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse PW\n";
                    #endif
                    if (*PW != '\0')
                        break;
                    if (CfgParser.Count() != 2) {
			    sprintf( szLastError, "wrong number of values (not 1) in line %d", CfgParser.Line );
			    return (0);
		    }
                    strcpy( PW, CfgParser[1] );
                    break;
                }
                /****************
                  ReceiptRequest
                 ****************/
                case TIC_LEXEM_RECEIPTREQUEST: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse ReceiptRequest\n";
                    #endif
                    break;
                }
                /*************
                  Destination
                 *************/
                case TIC_LEXEM_DESTINATION: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse Destination\n";
                    #endif
                    break;
                }
                /****************
                  Unknown string
                 ****************/
                default: {
                    #ifdef TIC_DEBUG_INFO
                    cout << "Parse unknown\n";
                    #endif
                    if (UnknownList == NULL) {
                        UnknownList = new SStringList;
                        LastUnknown = UnknownList;
                    }
                    else {
                        LastUnknown->Next = new SStringList;
                        LastUnknown = LastUnknown->Next;
                    }
                    LastUnknown->Next = NULL;
                    LastUnknown->String = new char[strlen(CfgParser.String()) + 1];
                    strcpy( LastUnknown->String, CfgParser.String() );
                }
            }
        }
    }
    if (!CfgParser.Eof())
        return (0);
    // check tic
    if (*Area == '\0' || *File == '\0' || Origin.Zone == 0 ||
        From.Zone == 0 || *PW == '\0' || !CRCDefined)
        return (0);
    return (1);
}

void CTic::PrintInfo() {
    cout << "Tic file information\n";
    cout << "Area = " << Area << endl;
    cout << "File = " << File << endl;
    cout << "Desc = " << Desc << endl;
    cout << "Origin = " << Origin.Zone << ":" << Origin.Net << "/"
         << Origin.Node << "." << Origin.Point << "@" << Origin.Domain << endl;
    cout << "From = " << From.Zone << ":" << From.Net << "/" << From.Node
         << "." << From.Point << "@" << From.Domain << endl;
    cout << "To = " << To.Zone << ":" << To.Net << "/" << To.Node
         << "." << To.Point << "@" << To.Domain << endl;
    cout << "PW = " << PW << endl;
}

int CTic::Forward(SLink *Link, char *FilePath, SLinked *ForwardList) {
    time_t       timestamp;
    char         TicFileName[MAX_PATH];
    char         FileName[MAX_PATH];
    char         ShortFileName[MAX_PATH];
    char         Line[256];
    ofstream     out;
    int          i, j;
    SStringList *CurString;
    SLinked     *CurForward;
    SAddrList   *CurSeenby;
    char         BSOTypeChar;
    
    // create tic
    // form tic name
    sprintf( TicFileName, "%sad%06lu.tic", GlobalVars.NewTicsPath, ticNamer.getFreeNumber() );
    out.open( TicFileName );
    if (!out)
        return (0);
    out << "Area " << Area << endl;
    if (*AreaDesc != '\0')
        out << "AreaDesc " << AreaDesc << endl;
    // form full file name
    strcpy( FileName, FilePath );
    if (*FullName != '\0')
        strcat( FileName, FullName );
    else
        strcat( FileName, File );
    ConvertNameToShort( FileName, ShortFileName );
    // get only file name without path
    for (i = 0, j = -1; ShortFileName[i] != '\0'; i++) {
        if (ShortFileName[i] == '\\')
            j = i;
    }
    j++;
    out << "File " << ShortFileName + j << endl;
    // check full file name
    if (*FullName != '\0')
        strcpy( FileName, FullName );
    else
        strcpy( FileName, File );
    if (strcmp( FileName, ShortFileName + j) != 0) {
        // we need to use FullName tag
        // check spaces
        for (i = 0; FileName[i] != '\0'; i++) {
            if (FileName[i] == ' ')
                break;
        }
        out << "FullName ";
        if (FileName[i] == ' ') // contains space(s)
            out << "\"" << FileName << "\"\n";
        else
            out << FileName << endl;
    }
    // output Size
    if (Size != -1)
        out << "Size " << Size << endl;
    // output CRC
    sprintf( Line, "CRC %08lX", CRC32 );
    out << Line << endl;
    // output Origin
    out << "Origin " << Origin.Zone << ":" << Origin.Net << "/" << Origin.Node;
    if (Origin.Point != 0)
        out << "." << Origin.Point;
    if (stricmp( Origin.Domain, Link->Address.Domain) != 0 ||
        stricmp( Origin.Domain, GlobalVars.MainAddr.Domain) != 0)
        out << "@" << Origin.Domain;
    out << endl;
    // output From
    out << "From " << Link->OurAddr->Zone << ":"
        << Link->OurAddr->Net << "/" << Link->OurAddr->Node;
    if (Link->OurAddr->Point != 0)
        out << "." << Link->OurAddr->Point;
    if (stricmp( Link->OurAddr->Domain, Link->Address.Domain) != 0)
        out << "@" << Link->OurAddr->Domain;
    out << endl;
    // output To
    out << "To " << Link->Address.Zone << ":" << Link->Address.Net
        << "/" << Link->Address.Node;
    if (Link->Address.Point != 0)
        out << "." << Link->Address.Point;
    if (stricmp( Link->OurAddr->Domain, Link->Address.Domain) != 0)
        out << "@" << Link->Address.Domain;
    out << endl;
    // output Desc
    if (*Desc != '\0')
        out << "Desc " << Desc << endl;
    // output LDesc
    CurString = LDescList;
    while (CurString != NULL) {
        out << "LDesc " << CurString->String << endl;
        CurString = CurString->Next;
    }
    // output Created by
    sprintf( Line,
             "Created by %s v%lu.%lu Copyright (c) 1999 Andrey Kuprishov (2:5023/27@Fidonet)",
             PROG_NAME, PROG_VER_MAJOR, PROG_VER_MINOR );
    out << Line << endl;
    // output Replaces
    if (*Replaces != '\0')
        out << "Replaces " << Replaces << endl;
    // output Seenby
    CurSeenby = SeenbyList;
    while (CurSeenby != NULL) {
        out << "Seenby " << CurSeenby->Addr.Zone << ":" << CurSeenby->Addr.Net
            << "/" << CurSeenby->Addr.Node;
        if (CurSeenby->Addr.Point != 0)
            out << "." << CurSeenby->Addr.Point;
        if (stricmp( CurSeenby->Addr.Domain, Link->Address.Domain) != 0 ||
            stricmp( CurSeenby->Addr.Domain, Link->OurAddr->Domain) != 0)
            out << "@" << CurSeenby->Addr.Domain;
        out << endl;
        CurSeenby = CurSeenby->Next;
    }
    // output all forwards as Seenbys
    CurForward = ForwardList;
    while (CurForward != NULL) {
        out << "Seenby " << CurForward->LinkPtr->Address.Zone << ":"
            << CurForward->LinkPtr->Address.Net << "/"
            << CurForward->LinkPtr->Address.Node;
        if (CurForward->LinkPtr->Address.Point > 0)
            out << "." << CurForward->LinkPtr->Address.Point;
        if (stricmp( CurForward->LinkPtr->Address.Domain, Link->Address.Domain) != 0 ||
            stricmp( CurForward->LinkPtr->Address.Domain, Link->OurAddr->Domain) != 0)
            out << "@" << CurForward->LinkPtr->Address.Domain;
        out << endl;
        CurForward = CurForward->Next;
    }
    // output AKA for addressie if absent in seenbys
    if (GlobalVars.AllAKASeen == 0) {
        CurSeenby = SeenbyList;
        while (CurSeenby != NULL) {
            if (AddrCmp( CurSeenby->Addr, *(Link->OurAddr) ) == 0)
                break;
            CurSeenby = CurSeenby->Next;
        }
        if (CurSeenby == NULL) {
            out << "Seenby " << Link->OurAddr->Zone << ":"
                << Link->OurAddr->Net << "/" << Link->OurAddr->Node;
            if (Link->OurAddr->Point != 0)
                out << "." << Link->OurAddr->Point;
            if (stricmp( Link->OurAddr->Domain, Link->Address.Domain) != 0)
                out << "@" << Link->OurAddr->Domain;
            out << endl;
        }
    }
    // output Via lines
    CurString = ViaList;
    while (CurString != NULL) {
        out << "Via " << CurString->String << endl;
        CurString = CurString->Next;
    }
    if (*CreatedBy != '\0') {
        sprintf( Line, "Via %ld:%ld/%ld.%ld@%s %s",
                 From.Zone, From.Net, From.Node, From.Point, From.Domain,
                 CreatedBy );
        out << Line << endl;
    }
    // output Path lines
    CurString = PathList;
    while (CurString != NULL) {
        out << "Path " << CurString->String << endl;
        CurString = CurString->Next;
    }
    timestamp = time(NULL);
    sprintf( Line, "Path %lu:%lu/%lu.%lu@%s %lu %s",
             Link->OurAddr->Zone, Link->OurAddr->Net,
             Link->OurAddr->Node, Link->OurAddr->Point,
             Link->OurAddr->Domain, timestamp,
             asctime( gmtime( &timestamp )));
    out << Line;
    // output PW tag
    out << "PW " << Link->Password << endl;
    // output unknown lines
    CurString = UnknownList;
    while (CurString != NULL) {
        out << CurString->String << endl;
        CurString = CurString->Next;
    }
    out.close();
    
    // do BSO
    
    BSOManager.SetAddress( Link->Address );
    if (!BSOManager.SetBusyFlag()) {
        // wait 5 seconds
        Sleep(5000);
        if (!BSOManager.SetBusyFlag()) {
            unlink( TicFileName );
            return (0);
        }
    }

    switch (Link->BSOType) {
    case BSO_NORMAL:
        BSOTypeChar = 'f';
        break;
    case BSO_CRASH:
        BSOTypeChar = 'c';
        break;
    case BSO_DIRECT:
        BSOTypeChar = 'd';
        break;
    case BSO_HOLD:
        BSOTypeChar = 'h';
        break;
    case BSO_IMMEDIATE:
        BSOTypeChar = 'i';
        break;
    }
    
    sprintf( BSOManager.Name + BSOManager.NameLen, "%clo", BSOTypeChar );
    out.open( BSOManager.Name, ios::app );
    if (!out)
        out.open( BSOManager.Name );
    if (!out) {
        BSOManager.ClearBusyFlag();
        unlink( TicFileName );
        return (0);
    }
    out << ShortFileName << endl;
    ConvertNameToShort( TicFileName, ShortFileName );
    out << "^" << ShortFileName << endl;
    BSOManager.ClearBusyFlag();

    return (1);
}
